/*--------------------------------------------------------------------

   Perceptuum renderer, version 2.1
   Copyright (c) 2002-2003,  Harrison Ainsworth.

   www.hxa7241.org

--------------------------------------------------------------------*/

/*--------------------------------------------------------------------

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston,
   MA 02111-1307, USA

--------------------------------------------------------------------*/


#include "hxa7241math.h"
#include "Logger.h"

#include "Filters.h"   /// own header is included last


using namespace hxa7241::graphics;




//#define ENABLE_TEST_CODE


static const float PI = 3.1415926535897932384626433832795f;


float hxa7241::graphics::boxFilter
(
	float kernelRadius,
	float pointDistance
)
{
	if( kernelRadius < 0.0f )
	{
		kernelRadius = -kernelRadius;
	}

	if( pointDistance < 0.0f )
	{
		pointDistance = -pointDistance;
	}


	float result;

	if( pointDistance < kernelRadius )
	{
		result = 1.0f / (PI * kernelRadius * kernelRadius);
	}
	else
	{
		result = 0.0f;
	}

	return result;
}


float hxa7241::graphics::coneFilter
(
	float kernelRadius,
	float pointDistance
)
{
	if( kernelRadius < 0.0f )
	{
		kernelRadius = -kernelRadius;
	}

	if( pointDistance < 0.0f )
	{
		pointDistance = -pointDistance;
	}


	float result;

	if( pointDistance < kernelRadius )
	{
		result = (1.0f - (pointDistance / kernelRadius)) /
		         ((1.0f / 3.0f) * PI * kernelRadius * kernelRadius);
	}
	else
	{
		result = 0.0f;
	}

	return result;
}


float hxa7241::graphics::pavicicFilter
(
	const float kernelRadius,
	const float pointDistance
)
{
	return pavicicFilter2( kernelRadius * kernelRadius, pointDistance * pointDistance );
}


float hxa7241::graphics::pavicicFilter2
(
	float kernelRadius2,
	float pointDistance2
)
{
	if( kernelRadius2 < 0.0f )
	{
		kernelRadius2 = -kernelRadius2;
	}

	if( pointDistance2 < 0.0f )
	{
		pointDistance2 = -pointDistance2;
	}


	float result;

	if( pointDistance2 < kernelRadius2 )
	{
		/// 1.0f - float(::exp( -1.953f ))
		static const float ONE_MINUS_EXP_NEG_1P953 = 0.85815211103467726466203689184461f;

		result = 1.883f * (0.918f / (PI * kernelRadius2)) *
		         (1.0f - ( (1.0f - float(::exp( -1.953f * (pointDistance2 / (2.0f * kernelRadius2)) ))) /
		                   ONE_MINUS_EXP_NEG_1P953 ));
	}
	else
	{
		result = 0.0f;
	}

	return result;
}


float hxa7241::graphics::cookFilter
(
	const float kernelRadius,
	const float pointDistance
)
{
	return cookFilter2( kernelRadius * kernelRadius, pointDistance * pointDistance );
}


float hxa7241::graphics::cookFilter2
(
	float kernelRadius2,
	float pointDistance2
)
{
	if( kernelRadius2 < 0.0f )
	{
		kernelRadius2 = -kernelRadius2;
	}

	if( pointDistance2 < 0.0f )
	{
		pointDistance2 = -pointDistance2;
	}


	float result;

	if( pointDistance2 < kernelRadius2 )
	{
		const float expNegKR = float(::exp( -kernelRadius2 ));
		result = ( float(::exp( -pointDistance2 )) - expNegKR ) /
		         ( PI * (1.0f - (expNegKR * (1.0f + kernelRadius2))) );

//		const float expNegKR = float(::exp( -kernelRadius2 ));
//		result = ( float(::exp( -pointDistance2 )) - expNegKR ) /
//		         ((1.0f - (expNegKR * (1.0f + kernelRadius2))) / kernelRadius2);
	}
	else
	{
		result = 0.0f;
	}

	return result;
}


float hxa7241::graphics::maxFilter
(
	float kernelRadius,
	float pointDistance
)
{
	if( kernelRadius < 0.0f )
	{
		kernelRadius = -kernelRadius;
	}

	if( pointDistance < 0.0f )
	{
		pointDistance = -pointDistance;
	}


	float result;

	if( pointDistance < kernelRadius )
	{
		const float t = kernelRadius;
		const float s = t * 0.350788f;

		if( pointDistance < s )
		{
			result = 1.0f - ((pointDistance * pointDistance) / (s * t));
		}
		else
		{
			const float tMr = t - pointDistance;
			result = (tMr * tMr) / (t * (t - s));
		}

		const float tMs = t - s;
		const float s2  = s * s;
		const float t2  = t * t;
		result /= (2.0f * PI * ( (s2 * 0.5f) -
		                         ((s2 * s2) / (4.0f * s * t)) +
		                         ((t2 * t2) / (12.0f * t * tMs)) -
		                         ((s2 * t2) / (2.0f * t * tMs)) +
		                         ((2.0f * s2 * s * t) / (3.0f * t * tMs)) -
		                         ((s2 * s2) / (4.0f * t * tMs)) ) );
	}
	else
	{
		result = 0.0f;
	}

	return result;
}




float hxa7241::graphics::mitchellNetravaliFilter
(
	float kernelRadius,
	float pointDistance
)
{
	if( kernelRadius < 0.0f )
	{
		kernelRadius = -kernelRadius;
	}

	if( pointDistance < 0.0f )
	{
		pointDistance = -pointDistance;
	}


	float result;

	if( pointDistance < kernelRadius )
	{
		const float p  = 2.0f * pointDistance / kernelRadius;
		const float p2 = p * p;
		const float p3 = p * p2;

		if( p < 1.0f )
		{
			result = (7.0f * p3) + (-12.0f * p2) + (16.0f / 3.0f);
		}
		else
		{
			result = ((-7.0f / 3.0f) * p3) + (12.0f * p2) + (-20.0f * p) + (32.0f / 3.0f);
		}

		result *= 1.357150f / (PI * kernelRadius * kernelRadius * (19.0f / 180.0f) * 6.0f);
	}
	else
	{
		result = 0.0f;
	}

	return result;
}








bool hxa7241::graphics::test_Filters
(
	const bool isLogging
)
{
#ifndef ENABLE_TEST_CODE
	if(isLogging) hxa7241::utility::Logger() << "test_Filters::  [disabled]";

	return false;


#else
	bool isSucceeded = true;


	if(isLogging) hxa7241::utility::Logger() << "test_Filters::  === filters test ===";


	static const char* filterNames[]   = { "boxFilter", "coneFilter", "pavicicFilter", "cookFilter", "maxFilter" };//, "mitchellNetravaliFilter" };
	float (*filters[])( float, float ) = { boxFilter, coneFilter, pavicicFilter, cookFilter, maxFilter };//, mitchellNetravaliFilter };
	const dword n = sizeof(filters) / sizeof(void*);

	for( dword fi = 0;  fi < n;  ++fi )
	{
		float (*filter)( float, float ) = filters[fi];
		const char* filterName          = filterNames[fi];


		/// check for positivity, symmetry, and finite support
		{
			bool isPositive  = true;
			bool isSymmetric = true;
			bool isFinite    = true;

			const float radius = 3.7f;
			const float inc    = radius / 32.0f;
			const float end    = radius * 1.5f;;
			for( float p = 0.0f;  p < end;  p += inc )
			{
				const float fp = filter( radius, +p );
				const float fn = filter( radius, -p );

				isPositive  &= (fp >= 0.0f);
				isSymmetric &= (fp == fn);
				isFinite    &= ((p < radius) | (fp == 0.0f));
			}

			if(isLogging) hxa7241::utility::Logger() << "test_Filters::  " << filterName << " is "  << (isPositive  ? "" : "*** not ") << "positive" <<
			                                                                               ", is "  << (isSymmetric ? "" : "*** not ") << "symmetric" <<
			                                                                               ", has " << (isFinite ?    "" : "*** not ") << "finite support";

			isSucceeded &= (isPositive & isSymmetric & isFinite);
		}


		/// check for normalisation
		{
			float sum   = 0.0f;
			//dword count = 0;

			static const dword HALF_WIDTH = 512 >> 1;   /// < 2^16
			for( dword y = -HALF_WIDTH;  y <= HALF_WIDTH;  ++y )
			{
				for( dword x = -HALF_WIDTH;  x <= HALF_WIDTH;  ++x )
				{
					const float distance = float(::sqrt((x * x) + (y * y)));
					if( distance < float(HALF_WIDTH) )
					{
						sum += filter( float(HALF_WIDTH), distance );
						//++count;
					}
				}
			}

			const float dif    = sum - 1.0f;
			const float difAbs = dif >= 0.0f ? dif : -dif;

			bool isNormalised = difAbs < 0.005f;

			if(isLogging) hxa7241::utility::Logger() << "test_Filters::  " << filterName << " is " << (isNormalised  ? "" : "*** not ") << "normalised:  " << sum;

			isSucceeded &= isNormalised;
		}
	}


	if(isLogging) hxa7241::utility::Logger() << "test_Filters::  " << (isSucceeded  ? "--- succeeded" : "*** failed");




//	{
//		const float radius = 1.0f;
//		const float inc    = radius / 16.0f;
//
//		for( float p = -radius;  p < radius;  p += inc )
//		{
//			const float box  = boxFilter(radius, p);
//			const float cone = coneFilter(radius, p);
//			const float pav  = pavicicFilter(radius, p);
//			const float cook = cookFilter(radius, p);
//			const float max  = maxFilter(radius, p);
//			const float mitc = mitchellNetravaliFilter(radius, p);
//
//			if(isLogging) hxa7241::utility::Logger() << "test_Filters::  " << p << "  boxFilter: " << box << "  coneFilter: " << cone << "  pavicic: "  << pav << "  cookFilter: " << cook << "  maxFilter: " << max << "  mitcFilter: " << mitc;
//		}
//	}


	return isSucceeded;
#endif
}
