/*--------------------------------------------------------------------

   Perceptuum renderer, version 2.1
   Copyright (c) 2002-2003,  Harrison Ainsworth.

   www.hxa7241.org

--------------------------------------------------------------------*/

/*--------------------------------------------------------------------

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston,
   MA 02111-1307, USA

--------------------------------------------------------------------*/


#ifndef OctreeAuxiliary_h
#define OctreeAuxiliary_h


#include "Primitives.h"
#include "Vector3f.h"




#include "hxa7241_graphics.h"
namespace hxa7241 { namespace graphics {
	using namespace hxa7241::primitives;
	using hxa7241::graphics::Vector3f;


/**
 * geometric data for the bound of an octree cell.<br/><br/>
 *
 * position is that of the lower corner, radius is that of the circumsphere.
 * <br/><br/>
 *
 * subcell numbering:
 * <pre>
 * 		y z           110 111
 * 		|/   010 011  100 101
 * 		 -x  000 001
 * </pre>
 */
class OctreeBound
{
/// standard object services -----------------------------------------------------------------------
public:
	         OctreeBound( const Vector3f& positionOfLowerCorner,
	                      float           size );
	         OctreeBound( const OctreeBound& parentCellBound,
	                      dword              subCellIndex );

	        ~OctreeBound();
	         OctreeBound( const OctreeBound& );
	OctreeBound& operator=( const OctreeBound& );


/// queries ----------------------------------------------------------------------------------------
	        const Vector3f& getPosition()                                                     const;
	        float           getSize()                                                         const;
	        const Vector3f& getCenter()                                                       const;
	        float           getRadius()                                                       const;


/// fields -----------------------------------------------------------------------------------------
private:
	Vector3f positionOfLowerCorner_m;
	float    size_m;
	Vector3f center_m;
	float    circumSphereRadius_m;
};




/**
 * global and local octree cell data.<br/><br/>
 *
 * to be made dynamically during each level of tree descent, so storage is
 * avoided, except to hold one at the root.<br/><br/>
 *
 * subcell numbering:
 * <pre>
 * 		y z           110 111
 * 		|/   010 011  100 101
 * 		 -x  000 001
 * </pre>
 */
class OctreeData
{
/// standard object services -----------------------------------------------------------------------
public:
	         OctreeData( const Vector3f& positionOfLowerCorner,
	                     float           size,
	                     dword           maxElementCountPerCell,
	                     dword           maxLevels );
	         OctreeData( const OctreeData& parentCellData,
	                     dword             subCellIndex );

	        ~OctreeData();
	         OctreeData( const OctreeData& );
	OctreeData& operator=( const OctreeData& );


/// queries ----------------------------------------------------------------------------------------
	        const OctreeBound& getBound()                                                     const;
	        dword              getMaxElementCountPerCell()                                    const;

	        dword              getLevel()                                                     const;
	        bool               isSubdivide( dword itemCount )                                 const;


/// fields -----------------------------------------------------------------------------------------
private:
	/// local to cell
	OctreeBound bound_m;
	dword       level_m;

	/// global for octree
	dword       maxElementsPerCell_m;
	dword       maxLevel_m;
};




/**
 * visitor abstract base, for Octree implementation use.<br/><br/>
 *
 * return value of getSubcellOverlapsV is 8 bits, each bit is a bool
 * corresponding to a subcell, the high bit for subcell 7, the low bit for
 * subcell 0.<br/><br/>
 *
 * return value of getSubcellSelectionsV is 16 bits, comprising eight bit-pairs,
 * each containing an ESelection value corresponding to a subcell, the high pair
 * for subcell 7, the low pair for subcell 0.<br/><br/>
 *
 * subcell numbering:
 * <pre>
 * 		y z           110 111
 * 		|/   010 011  100 101
 * 		 -x  000 001
 * </pre>
 *
 * @see OctreeCell
 * @see OctreeBranch
 * @see OctreeLeaf
 */
class OctreeVisitor
{
/// standard object services -----------------------------------------------------------------------
protected:
	         OctreeVisitor() {}
public:
	virtual ~OctreeVisitor() {}
private:
	         OctreeVisitor( const OctreeVisitor& );
	OctreeVisitor& operator=( const OctreeVisitor& );


/// commands ---------------------------------------------------------------------------------------
public:
	virtual void  visitItemV( const void* pItem,
	                          bool        isInside )                                             =0;


/// queries ----------------------------------------------------------------------------------------
	virtual bool  isOverlappingCellV( const void*     pItem,
	                                  const Vector3f& position,
	                                  float           size )                               const =0;
	virtual dword getSubcellOverlapsV( const void*     pItem,
	                                   const Vector3f& planePositions )                    const =0;

	/**
	 * different selection states.<br/><br/>
	 *
	 * OUTSIDE: octree cell bound is entirely outside object.<br/>
	 * PARTIAL: octree cell bound is probably partly inside object, but maybe
	 * 	  entirely inside or entirely outside object.<br/>
	 * INSIDE:  octree cell bound is entirely inside object.<br/>
	 */
	enum ESelection { OUTSIDE = 0, PARTIAL = 1, INSIDE = 3 };
	static const dword ALL_INSIDE  = 0x0000FFFF;
	static const dword ALL_OUTSIDE = 0x00000000;

	virtual ESelection isSelected( const OctreeBound&,
	                               bool isInside )                                         const =0;
	virtual dword      getSubcellSelections( const OctreeBound&,
	                                         bool isInside )                               const =0;
};








/// inlines ///

/// OctreeBound ------------------------------------------------------------------------------------
inline
const Vector3f& OctreeBound::getPosition() const
{
	return positionOfLowerCorner_m;
}


inline
float OctreeBound::getSize() const
{
	return size_m;
}


inline
const Vector3f& OctreeBound::getCenter() const
{
	return center_m;
}


inline
float OctreeBound::getRadius() const
{
	return circumSphereRadius_m;
}




/// OctreeData -------------------------------------------------------------------------------------
inline
const OctreeBound& OctreeData::getBound() const
{
	return bound_m;
}


inline
dword OctreeData::getMaxElementCountPerCell() const
{
	return maxElementsPerCell_m;
}


inline
dword OctreeData::getLevel() const
{
	return level_m;
}


}}




#endif//OctreeAuxiliary_h
