/*--------------------------------------------------------------------

   Perceptuum renderer, version 2.1
   Copyright (c) 2002-2003,  Harrison Ainsworth.

   www.hxa7241.org

--------------------------------------------------------------------*/

/*--------------------------------------------------------------------

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston,
   MA 02111-1307, USA

--------------------------------------------------------------------*/




//#include "Serialization.h"
#include "UnitVector3f.h"
#include "Matrix3f.h"
#include "Polar.h"
#include "hxa7241math.h"
#include "Logger.h"
#include <iostream>

#include "Quaternion.h"   /// own header is included last


using namespace hxa7241::graphics;




//#define ENABLE_TEST_CODE


/// standard object services -----------------------------------------------------------------------
Quaternion::Quaternion()
 :	s_m( 1.0f ),
	v_m()
{
}


Quaternion::Quaternion
(
	const float         angle,
	const UnitVector3f& axis
)
{
	Quaternion::setAngleAndAxis( angle, axis );
}


Quaternion::Quaternion
(
	const Matrix3f& m
)
{
	Quaternion::setMatrix( m );
}


Quaternion::Quaternion
(
	const float rotationX,
	const float rotationY,
	const float rotationZ
)
{
	Quaternion::setRotations( rotationX, rotationY, rotationZ );
}


Quaternion::Quaternion
(
	const float     s,
	const Vector3f& v,
	const bool      //dummy
)
{
	Quaternion::setElements( s, v );
}


Quaternion::Quaternion
(
	const float s,
	const float vx,
	const float vy,
	const float vz
)
{
	Quaternion::setElements( s, vx, vy, vz );
}


Quaternion::~Quaternion()
{
}


Quaternion::Quaternion
(
	const Quaternion& other
)
 :	s_m( other.s_m ),
	v_m( other.v_m )
{
}


Quaternion& Quaternion::operator=
(
	const Quaternion& other
)
{
	if( &other != this )
	{
		s_m = other.s_m;
		v_m = other.v_m;
	}

	return *this;
}




/// streaming --------------------------------------------------------------------------------------
std::ostream& hxa7241::graphics::operator<<( std::ostream& out, const Quaternion& obj )
{
	return out << "(" << obj.s_m << ", " << obj.v_m << ")";
}


std::istream& hxa7241::graphics::operator>>( std::istream& in, Quaternion& obj )
{
	char c;
	return in >> c >> obj.s_m >> c >> obj.v_m >> c;
}




/// commands ---------------------------------------------------------------------------------------
void Quaternion::setAngleAndAxis
(
	const float         angle,
	const UnitVector3f& axis
)
{
	float sine;
	float cosine;
	hxa7241::graphics::sinAndcos( angle * 0.5f, sine, cosine );

	s_m = cosine;
	(v_m = axis.getVector()) *= sine;
}


void Quaternion::setMatrix
(
	const Matrix3f& m
)
{
	const Vector3f& row0 = m.getRow0();
	const Vector3f& row1 = m.getRow1();
	const Vector3f& row2 = m.getRow2();
	const float me[3][3] = { { row0.getX(), row1.getX(), row2.getX() },
	                         { row0.getY(), row1.getY(), row2.getY() },
	                         { row0.getZ(), row1.getZ(), row2.getZ() }  };
	/// alternate form
	//const float me[3][3] = { { row0.getX(), row0.getY(), row0.getZ() },
	//						 { row1.getX(), row1.getY(), row1.getZ() },
	//						 { row2.getX(), row2.getY(), row2.getZ() }  };


	const float trace = me[0][0] + me[1][1] + me[2][2] + 1.0f;

	if( trace > 0.00000001f )
	{
		const float a = ::sqrt(trace) * 0.5;
		const float b = 0.25f / a;

		setElements( a,
		             (me[2][1] - me[1][2]) * b,
		             (me[0][2] - me[2][0]) * b,
		             (me[1][0] - me[0][1]) * b );
	}
	else
	{
		if( (me[0][0] > me[1][1]) & (me[0][0] > me[2][2]) )
		{
			const float a = ::sqrt((me[0][0] - (me[1][1] + me[2][2])) + 1.0f) * 0.5;
			const float b = 0.25f / a;

			setElements( (me[2][1] - me[1][2]) * b,
			             a,
			             (me[0][1] + me[1][0]) * b,
			             (me[2][0] + me[0][2]) * b );
		}
		else if( me[1][1] > me[2][2] )
		{
			const float a = ::sqrt((me[1][1] - (me[2][2] + me[0][0])) + 1.0f) * 0.5;
			const float b = 0.25f / a;

			setElements( (me[0][2] - me[2][0]) * b,
			             (me[0][1] + me[1][0]) * b,
			             a,
			             (me[1][2] + me[2][1]) * b );
		}
		else
		{
			const float a = ::sqrt((me[2][2] - (me[0][0] + me[1][1])) + 1.0f) * 0.5;
			const float b = 0.25f / a;

			setElements( (me[1][0] - me[0][1]) * b,
			             (me[2][0] + me[0][2]) * b,
			             (me[1][2] + me[2][1]) * b,
			             a );
		}
	}
}


void Quaternion::setRotations
(
	const float rotationX,
	const float rotationY,
	const float rotationZ
)
{
	const Quaternion qx( rotationX, UnitVector3f::X() );
	const Quaternion qy( rotationY, UnitVector3f::Y() );
	const Quaternion qz( rotationZ, UnitVector3f::Z() );

	((*this = qz) *= qy) *= qx;
}


Quaternion& Quaternion::invert()
{
	v_m.negate();

	return *this;
}


Quaternion& Quaternion::operator*=
(
	const Quaternion& other
)
{
	const float w = (s_m * other.s_m) - v_m.dot( other.v_m );
	v_m = (((other.v_m * s_m) += (v_m * other.s_m)) += v_m.cross( other.v_m ));
	s_m = w;

//	const float w = (s_m * other.s_m) - ((v_m.getX() * other.v_m.getX()) + (v_m.getY() * other.v_m.getY()) + (v_m.getZ() * other.v_m.getZ()));
//	const float x = (s_m * other.v_m.getX()) + (other.s_m * v_m.getX()) + (v_m.getY() * other.v_m.getZ()) - (v_m.getZ() * other.v_m.getY());
//	const float y = (s_m * other.v_m.getY()) + (other.s_m * v_m.getY()) + (v_m.getZ() * other.v_m.getX()) - (v_m.getX() * other.v_m.getZ());
//	const float z = (s_m * other.v_m.getZ()) + (other.s_m * v_m.getZ()) + (v_m.getX() * other.v_m.getY()) - (v_m.getY() * other.v_m.getX());
//
//	s_m = w;
//	v_m.setXYZ( x, y, z );

	return *this;
}


Quaternion& Quaternion::operator*=
(
	const float scaling
)
{
	float        angle;
	UnitVector3f axis;

	getAngleAndAxis( angle, axis );

	angle *= scaling;

	setAngleAndAxis( angle, axis );

	return *this;
}




/// queries ----------------------------------------------------------------------------------------
void Quaternion::getAngleAndAxis
(
	float&        angle,
	UnitVector3f& axis
) const
{
	const float cosine = s_m;

	float sine = float(::sqrt( 1.0f - (cosine * cosine) ));
	if( float(::fabs(sine)) < 0.0005f )
	{
		sine = 1.0f;
	}
	const float oneOverSine = 1.0f / sine;


	angle = float(::acos( cosine )) * 2.0f;
	axis.setXYZ( v_m.getX() * oneOverSine,
	             v_m.getY() * oneOverSine,
	             v_m.getZ() * oneOverSine );
}


void Quaternion::getMatrix
(
	Matrix3f& m
) const
{
	const float xx = v_m.getX() * v_m.getX();
	const float yy = v_m.getY() * v_m.getY();
	const float zz = v_m.getZ() * v_m.getZ();
	const float ss = s_m        * s_m;
	const float xy = v_m.getX() * v_m.getY();
	const float sz = s_m        * v_m.getZ();
	const float xz = v_m.getX() * v_m.getZ();
	const float sy = s_m        * v_m.getY();
	const float yz = v_m.getY() * v_m.getZ();
	const float sx = s_m        * v_m.getX();

	m.setElements( ss + xx - yy - zz,  2.0f * (xy + sz),  2.0f * (xz - sy),  0,
	               2.0f * (xy - sz),  ss - xx + yy - zz,  2.0f * (yz + sx),  0,
	               2.0f * (xz + sy),  2.0f * (yz - sx),  ss - xx - yy + zz,  0 );
	/// alternate form
	//m.setElements( ss + xx - yy - zz,  2.0f * (xy - sz),  2.0f * (xz + sy),  0,
	//			   2.0f * (xy + sz),  ss - xx + yy - zz,  2.0f * (yz - sx),  0,
	//			   2.0f * (xz - sy),  2.0f * (yz + sx),  ss - xx - yy + zz,  0 );
}


Quaternion Quaternion::getInverse() const
{
	return Quaternion( *this ).invert();
}


Quaternion Quaternion::operator*
(
	const Quaternion& other
) const
{
	return Quaternion( *this ) *= other;
}


Quaternion Quaternion::operator*
(
	const float scaling
) const
{
	return Quaternion( *this ) *= scaling;
}


void Quaternion::rotate
(
	Vector3f& point
) const
{
	/// make non-unitized quaternion
	Quaternion qp;
	qp.s_m = 0.0f;
	qp.v_m = point;

	point = ((getInverse() *= qp) *= *this).v_m;
	/// alternate form
	//point = ((*this * qp) *= getInverse()).v_m;
}


bool Quaternion::operator==
(
	const Quaternion& other
) const
{
	return (s_m == other.s_m) & (v_m == other.v_m);
}


bool Quaternion::operator!=
(
	const Quaternion& other
) const
{
	return (s_m != other.s_m) | (v_m != other.v_m);
}




/// constants --------------------------------------------------------------------------------------
const Quaternion& Quaternion::QUARTER_X()
{
	static const Quaternion k( 0.70710678118654752440084436210485f, 0.70710678118654752440084436210485f, 0.0f, 0.0f );
	return k;
}


const Quaternion& Quaternion::QUARTER_Y()
{
	static const Quaternion k( 0.70710678118654752440084436210485f, 0.0f, 0.70710678118654752440084436210485f, 0.0f );
	return k;
}


const Quaternion& Quaternion::QUARTER_Z()
{
	static const Quaternion k( 0.70710678118654752440084436210485f, 0.0f, 0.0f, 0.70710678118654752440084436210485f );
	return k;
}




/// implementation ---------------------------------------------------------------------------------
void Quaternion::setElements
(
	const float     s,
	const Vector3f& v
)
{
	s_m = s;
	v_m = v;

	unitize();
}


void Quaternion::setElements
(
	const float s,
	const float vx,
	const float vy,
	const float vz
)
{
	s_m = s;
	v_m.setXYZ( vx, vy, vz );

	unitize();
}


void Quaternion::unitize()
{
	const float dotSelf	= (s_m * s_m) + v_m.dot( v_m );

	if( dotSelf != 1.0f )
	{
		const float length = float(::sqrt( dotSelf ));

		if( length == 0.0f )
		{
			s_m = 1.0f;
			v_m = Vector3f::ZERO();
		}
		else
		{
			const float oneOverLength = 1.0f / length;

			s_m *= oneOverLength;
			v_m *= oneOverLength;
		}
	}
}




/*/// serialization ----------------------------------------------------------------------------------
SerializerOut& hxa7241::graphics::operator<<( SerializerOut& out, const Quaternion& obj )
{
	out.write( "s", obj.s_m );
	out.write( "v", obj.v_m );

	return out;
}


SerializerIn& hxa7241::graphics::operator>>( SerializerIn& in, Quaternion& obj )
{
	in.read( "s", obj.s_m );
	in.read( "v", obj.v_m );

	return in;
}


const char* Vector3f::CLASS_NAME()
{
	static const char CLASS_NAME[] = "hxa7241::graphics::Quaternion";
	return CLASS_NAME;
}


const char* Vector3f::CLASS_DESCRIPTION()
{
	static const char CLASS_DESCRIPTION[] = "a unit quaternion";
	return CLASS_DESCRIPTION;
}*/




/// test -------------------------------------------------------------------------------------------
bool Quaternion::test
(
	const bool isLogging
)
{
	static const char PREFIX[] = "Quaternion::test  ";


#ifndef ENABLE_TEST_CODE
	if(isLogging) hxa7241::utility::Logger() << PREFIX << "[disabled]";

	return false;


#else
	bool isSucceeded = true;

	if(isLogging) hxa7241::utility::Logger() << PREFIX << "=== general test ===";


	/// to and from angle,axis are opposites
	{
		Quaternion q;

		/// to
		const float        angle0 = 3.590392f;
		const UnitVector3f axis0( 1.0f, -0.5f, 0.3f );
        q.setAngleAndAxis( angle0, axis0 );

		/// from
		float        angle1;
		UnitVector3f axis1;
        q.getAngleAndAxis( angle1, axis1 );

		const bool isOk = (float(::fabs(angle0 - angle1)) < 0.001f) &
		                  (axis0.dot(axis1) > 0.999f);
		isSucceeded &= isOk;

		if(isLogging) hxa7241::utility::Logger() << PREFIX << "   angles dif: " << (angle1 - angle0) << "  axiss dot: " << axis0.dot(axis1);
		if(isLogging) hxa7241::utility::Logger() << PREFIX << "to/from angle,axis test " << (isOk ? "succeeded" : "failed");
	}


	/// to and from matrix are opposites
	{
		Quaternion q;

		/// to
		Matrix3f m0;
		{
			Matrix3f mx;
			Matrix3f my;
			Matrix3f mz;
			mx.setToRotateX( 1.0f );
			my.setToRotateY( 2.0f );
			mz.setToRotateZ( 4.0f );

			((m0 = mz).multiplyEq(my)).multiplyEq(mx);
		}
        q.setMatrix( m0 );

		/// from
		Matrix3f m1;
		q.getMatrix( m1 );

		bool isOk = true;
		{
			Vector3f d[3] = { Vector3f(m1.getRow0() - m0.getRow0()),
							  Vector3f(m1.getRow1() - m0.getRow1()),
							  Vector3f(m1.getRow2() - m0.getRow2()) };
			for( dword i = 3;  i-- > 0; )
			{
				isOk &= float(::fabs(d[i].getX())) < 0.001f;
				isOk &= float(::fabs(d[i].getY())) < 0.001f;
				isOk &= float(::fabs(d[i].getZ())) < 0.001f;
			}
			isOk &= m1.getCol3().isZero();
		}

		isSucceeded &= isOk;

		if(isLogging) hxa7241::utility::Logger() << PREFIX << "   matr row0 dif: " << (m1.getRow0() - m0.getRow0());
		if(isLogging) hxa7241::utility::Logger() << PREFIX << "   matr row1 dif: " << (m1.getRow1() - m0.getRow1());
		if(isLogging) hxa7241::utility::Logger() << PREFIX << "   matr row2 dif: " << (m1.getRow2() - m0.getRow2());
		if(isLogging) hxa7241::utility::Logger() << PREFIX << "   matr col3 dif: " << (m1.getCol3() - m0.getCol3());
		if(isLogging) hxa7241::utility::Logger() << PREFIX << "to/from matrix test " << (isOk ? "succeeded" : "failed");
	}


	/// invert is self opposite
	{
		const Quaternion q0( 1.0f, 2.0f, 4.0f );

		Quaternion q1( q0 );
		q1.invert();
		q1.invert();

		const bool isOk = (float(::fabs(q1.s_m - q0.s_m)) < 0.001f) &
		                  (float(::fabs(q1.v_m.getX() - q0.v_m.getX())) < 0.001f) &
		                  (float(::fabs(q1.v_m.getY() - q0.v_m.getY())) < 0.001f) &
		                  (float(::fabs(q1.v_m.getZ() - q0.v_m.getZ())) < 0.001f);
		isSucceeded &= isOk;

		if(isLogging) hxa7241::utility::Logger() << PREFIX << "   quat dif: " << (q1.s_m - q0.s_m) << ", " << (q1.v_m - q0.v_m);
		if(isLogging) hxa7241::utility::Logger() << PREFIX << "invert self-opposite test " << (isOk ? "succeeded" : "failed");
	}


	/// all commands maintain invariants - unitization
	{
		bool isOk = true;

		Quaternion q;

		q.setAngleAndAxis( 3.590392f, UnitVector3f( 1.0f, -0.5f, 0.3f ) );
		isOk &= float(::fabs( float(::sqrt( (q.s_m * q.s_m) + q.v_m.dot( q.v_m ) )) - 1.0f )) < 0.001f;
		if(isLogging) hxa7241::utility::Logger() << PREFIX << "   " << float(::sqrt( (q.s_m * q.s_m) + q.v_m.dot( q.v_m ) ));

		{
			Matrix3f m0;
			{
				Matrix3f mx;
				Matrix3f my;
				Matrix3f mz;
				mx.setToRotateX( 1.0f );
				mx.setToRotateY( 2.0f );
				mx.setToRotateZ( 4.0f );

				((m0 = mx).multiplyEq(my)).multiplyEq(mz);
			}
			q.setMatrix( m0 );
			isOk &= float(::fabs( float(::sqrt( (q.s_m * q.s_m) + q.v_m.dot( q.v_m ) )) - 1.0f )) < 0.001f;
			if(isLogging) hxa7241::utility::Logger() << PREFIX << "   " << float(::sqrt( (q.s_m * q.s_m) + q.v_m.dot( q.v_m ) ));
		}

		q.setRotations( 1.0f, 2.0f, 4.0f );
		isOk &= float(::fabs( float(::sqrt( (q.s_m * q.s_m) + q.v_m.dot( q.v_m ) )) - 1.0f )) < 0.001f;
		if(isLogging) hxa7241::utility::Logger() << PREFIX << "   " << float(::sqrt( (q.s_m * q.s_m) + q.v_m.dot( q.v_m ) ));

		q.invert();
		isOk &= float(::fabs( float(::sqrt( (q.s_m * q.s_m) + q.v_m.dot( q.v_m ) )) - 1.0f )) < 0.001f;
		if(isLogging) hxa7241::utility::Logger() << PREFIX << "   " << float(::sqrt( (q.s_m * q.s_m) + q.v_m.dot( q.v_m ) ));

		const Quaternion q1( 1.0f, UnitVector3f( -0.25f, 10.0f, 0.3f ) );
		q.operator*=( q1 );
		isOk &= float(::fabs( float(::sqrt( (q.s_m * q.s_m) + q.v_m.dot( q.v_m ) )) - 1.0f )) < 0.001f;
		if(isLogging) hxa7241::utility::Logger() << PREFIX << "   " << float(::sqrt( (q.s_m * q.s_m) + q.v_m.dot( q.v_m ) ));

		q.setRotations( 1.0f, 2.0f, 4.0f );
		q.operator*=( 0.837405f );
		isOk &= float(::fabs( float(::sqrt( (q.s_m * q.s_m) + q.v_m.dot( q.v_m ) )) - 1.0f )) < 0.001f;
		if(isLogging) hxa7241::utility::Logger() << PREFIX << "   " << float(::sqrt( (q.s_m * q.s_m) + q.v_m.dot( q.v_m ) ));

		q.setElements( 1.0f, -2.2f, 3.3f, 11.0f );
		isOk &= float(::fabs( float(::sqrt( (q.s_m * q.s_m) + q.v_m.dot( q.v_m ) )) - 1.0f )) < 0.001f;
		if(isLogging) hxa7241::utility::Logger() << PREFIX << "   " << float(::sqrt( (q.s_m * q.s_m) + q.v_m.dot( q.v_m ) ));

		q.unitize();
		isOk &= float(::fabs( float(::sqrt( (q.s_m * q.s_m) + q.v_m.dot( q.v_m ) )) - 1.0f )) < 0.001f;
		if(isLogging) hxa7241::utility::Logger() << PREFIX << "   " << float(::sqrt( (q.s_m * q.s_m) + q.v_m.dot( q.v_m ) ));

		isSucceeded &= isOk;

		if(isLogging) hxa7241::utility::Logger() << PREFIX << "invariant test " << (isOk ? "succeeded" : "failed");
	}


	/// invert is invert
	{
		const Vector3f p0( 10.0f, 22.0f, -33.0f );

		/// set quaternion and point
		Quaternion q( 3.590392f, UnitVector3f( 1.0f, -0.5f, 0.3f ) );
		Vector3f   p1( p0 );

		/// rotate point
		q.rotate( p1 );

		/// invert quaternion
		q.invert();

		/// rotate point
		q.rotate( p1 );

		/// test equality with original point
		const bool isOk = (float(::fabs(p1.getX() - p0.getX())) < 0.001f) &
		                  (float(::fabs(p1.getY() - p0.getY())) < 0.001f) &
		                  (float(::fabs(p1.getZ() - p0.getZ())) < 0.001f);
		isSucceeded &= isOk;

		if(isLogging) hxa7241::utility::Logger() << PREFIX << "   point dif: " << (p1 - p0);
		if(isLogging) hxa7241::utility::Logger() << PREFIX << "invert test " << (isOk ? "succeeded" : "failed");
	}


	/// rotations are same
	{
		const Vector3f p0( 10.0f, 22.0f, -33.0f );

		/// set matrix from x,y,z
		Matrix3f m0;
		{
			Matrix3f mx;
			Matrix3f my;
			Matrix3f mz;
			mx.setToRotateX( 1.0f );
			my.setToRotateY( 2.0f );
			mz.setToRotateZ( 4.0f );

			((m0 = mz).multiplyEq(my)).multiplyEq(mx);
		}

		/// set quaternion a from matrix
		Quaternion qm( m0 );

		/// set quaternion b from x,y,z
		Quaternion qr( 1.0f, 2.0f, 4.0f );

		Vector3f pm0( p0 );
		Vector3f pqm( p0 );
		Vector3f pqr( p0 );

		/// rotate point
		m0.multiply( pm0 );
		qm.rotate( pqm );
		qr.rotate( pqr );

		/// test equality
		bool isOk = (float(::fabs(pqm.getX() - pm0.getX())) < 0.001f) &
		            (float(::fabs(pqm.getY() - pm0.getY())) < 0.001f) &
		            (float(::fabs(pqm.getZ() - pm0.getZ())) < 0.001f);
		isOk &= (float(::fabs(pqr.getX() - pm0.getX())) < 0.001f) &
		        (float(::fabs(pqr.getY() - pm0.getY())) < 0.001f) &
		        (float(::fabs(pqr.getZ() - pm0.getZ())) < 0.001f);
		isSucceeded &= isOk;

		if(isLogging) hxa7241::utility::Logger() << PREFIX << "   quat from matr dif: " << (pqm - pm0);
		if(isLogging) hxa7241::utility::Logger() << PREFIX << "   quat from rots dif: " << (pqr - pm0);
		if(isLogging) hxa7241::utility::Logger() << PREFIX << "rotations test " << (isOk ? "succeeded" : "failed");
	}


	/// rotation scaling correct
	{
		const Vector3f p0( 10.0f, 22.0f, -33.0f );

		Quaternion qr( 0.0f, 1.0f, 0.0f );
		Quaternion qs( 0.0f, 0.5f, 0.0f );
		qs *= 2.0f;

		Vector3f pqr( p0 );
		Vector3f pqs( p0 );

		qr.rotate( pqr );
		qs.rotate( pqs );

		/// test equality
		bool isOk = (float(::fabs(pqs.getX() - pqr.getX())) < 0.001f) &
		            (float(::fabs(pqs.getY() - pqr.getY())) < 0.001f) &
		            (float(::fabs(pqs.getZ() - pqr.getZ())) < 0.001f);
		isSucceeded &= isOk;

		if(isLogging) hxa7241::utility::Logger() << PREFIX << "   quat from scal dif: " << (pqs - pqr);
		if(isLogging) hxa7241::utility::Logger() << PREFIX << "rotations scaling test " << (isOk ? "succeeded" : "failed");
	}


	/// constants correct
	{
		const Vector3f p0( 10.0f, 22.0f, -33.0f );

		/// set matrixs for quarter turns
		Matrix3f mx;
		Matrix3f my;
		Matrix3f mz;
		mx.setToRotateX( 1.5707963267948966192313216916398f );
		my.setToRotateY( 1.5707963267948966192313216916398f );
		mz.setToRotateZ( 1.5707963267948966192313216916398f );

		Vector3f pxm( p0 );
		Vector3f pym( p0 );
		Vector3f pzm( p0 );
		Vector3f pxq( p0 );
		Vector3f pyq( p0 );
		Vector3f pzq( p0 );

		/// rotate point with matrixs
		mx.multiply( pxm );
		my.multiply( pym );
		mz.multiply( pzm );

		/// rotate point with quaternion constants
		Quaternion::QUARTER_X().rotate( pxq );
		Quaternion::QUARTER_Y().rotate( pyq );
		Quaternion::QUARTER_Z().rotate( pzq );

		/// test equality
		bool isOk = (float(::fabs(pxq.getX() - pxm.getX())) < 0.001f) &
		            (float(::fabs(pxq.getY() - pxm.getY())) < 0.001f) &
		            (float(::fabs(pxq.getZ() - pxm.getZ())) < 0.001f);
		isOk &= (float(::fabs(pyq.getX() - pym.getX())) < 0.001f) &
		        (float(::fabs(pyq.getY() - pym.getY())) < 0.001f) &
		        (float(::fabs(pyq.getZ() - pym.getZ())) < 0.001f);
		isOk &= (float(::fabs(pzq.getX() - pzm.getX())) < 0.001f) &
		        (float(::fabs(pzq.getY() - pzm.getY())) < 0.001f) &
		        (float(::fabs(pzq.getZ() - pzm.getZ())) < 0.001f);

		isSucceeded &= isOk;

		if(isLogging) hxa7241::utility::Logger() << PREFIX << "   x point dif: " << (pxq - pxm);
		if(isLogging) hxa7241::utility::Logger() << PREFIX << "   y point dif: " << (pyq - pym);
		if(isLogging) hxa7241::utility::Logger() << PREFIX << "   z point dif: " << (pzq - pzm);
		if(isLogging) hxa7241::utility::Logger() << PREFIX << "constants test " << (isOk ? "succeeded" : "failed");
	}


	if(isLogging) hxa7241::utility::Logger() << PREFIX << (isSucceeded ? "--- succeeded" : "*** failed");

	return isSucceeded;
#endif
}
