/*--------------------------------------------------------------------

   Perceptuum renderer, version 2.1
   Copyright (c) 2002-2003,  Harrison Ainsworth.

   www.hxa7241.org

--------------------------------------------------------------------*/

/*--------------------------------------------------------------------

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston,
   MA 02111-1307, USA

--------------------------------------------------------------------*/


#include "Matrix3f.h"
#include "Vector3f.h"
#include "UnitVector3f.h"
#include "ClipFlags.h"

#include "Transclip.h"   /// own header is included last


using namespace hxa7241::graphics;




/// standard object services -----------------------------------------------------------------------
Transclip::Transclip()
 :	Transform(),
	matrix_m ( Matrix4f::IDENTITY() )
{
}


Transclip::Transclip
(
	const Matrix4f& worldToClip
)
 :	Transform(),
	matrix_m ( worldToClip )
{
}


Transclip::~Transclip()
{
}


Transclip::Transclip( const Transclip& other )
 :	Transform( other ),
	matrix_m ( other.matrix_m )
{
}


Transclip& Transclip::operator=( const Transclip& other )
{
	assign( other );

	return *this;
}




/// commands ---------------------------------------------------------------------------------------
Transclip& Transclip::operator+=( const Transclip& other )
{
	matrix_m.multiplyEqPre( other.matrix_m );

	return *this;
}


Transclip& Transclip::operator+=( const Matrix3f& otherM3 )
{
	const Matrix4f m4( otherM3 );
	matrix_m.multiplyEqPre( m4 );

	return *this;
}


Transclip& Transclip::invert()
{
	matrix_m.invert();

	return *this;
}




/// queries ----------------------------------------------------------------------------------------
void Transclip::transform
(
	Vector3f& point
) const
{
   matrix_m.multiply( point, point );
}


void Transclip::transform
(
	UnitVector3f& uvect
) const
{
   matrix_m.multiply( uvect, uvect );
}


bool Transclip::getClipping
(
	const ClipFlags& enclosingClipFlags,
	const Vector3f   boundCorners[8],
	ClipFlags&       clipFlags
) const
{
	bool isOutsideEntirely = false;

	/// do clipping only if enclosing bound not entirely inside
	if( enclosingClipFlags.isClipped() )
	{
		ClipFlags or;
		ClipFlags and;

		/// step through points
		for( int i = 8;  i-- > 0; )
		{
			/// transform point to clip space
			Vector4f cornerClipSpace;
			matrix_m.multiply( boundCorners[i], cornerClipSpace );

			/// clip point, and aggregate flags
			const ClipFlags cornerFlags( cornerClipSpace );
			or  |= cornerFlags;
			and &= cornerFlags;
		}

		/// set output
		clipFlags         = or;
		isOutsideEntirely = and.isClipped();
	}

	return !isOutsideEntirely;
}


const Matrix4f& Transclip::getMatrix() const
{
	return matrix_m;
}




/// implementation ---------------------------------------------------------------------------------
void Transclip::assign
(
	const Transclip& other
)
{
	if( &other != this )
	{
		Transform::operator=( other );

		matrix_m = other.matrix_m;
	}
}
