/*--------------------------------------------------------------------

   Perceptuum renderer, version 2.1
   Copyright (c) 2002-2003,  Harrison Ainsworth.

   www.hxa7241.org

--------------------------------------------------------------------*/

/*--------------------------------------------------------------------

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston,
   MA 02111-1307, USA

--------------------------------------------------------------------*/


#ifndef Vector4f_h
#define Vector4f_h


#include "Primitives.h"
#include "Vector3f.h"
#include <iosfwd>

#include "hxa7241_utility.h"




#include "hxa7241_graphics.h"
namespace hxa7241 { namespace graphics {
	using namespace hxa7241::primitives;
	using hxa7241::utility::SerializerOut;
	using hxa7241::utility::SerializerIn;


/**
 * a 4d vector of floats.<br/><br/>
 *
 * provides explicit conversion to and from 3d vector.
 *
 * @see Vector3f
 */
class Vector4f
{
/// standard object services -----------------------------------------------------------------------
public:
	                  Vector4f();
	                  Vector4f( float x, float y, float z, float w );
	explicit          Vector4f( const float xyzw[4] );
	                  Vector4f( const Vector3f&, float w );
	explicit          Vector4f( const Vector3f& );

	                 ~Vector4f();
	                  Vector4f( const Vector4f& );
	        Vector4f& operator=( const Vector4f& );


/// streaming --------------------------------------------------------------------------------------
	friend  std::ostream& operator<<( std::ostream&, const Vector4f& );
	friend  std::istream& operator>>( std::istream&,       Vector4f& );


/// access -----------------------------------------------------------------------------------------
	        Vector4f& setXYZW( float x, float y, float z, float w );
	        Vector4f& setXYZW( const float xyzw[4] );
	        Vector4f& setXYZW( const Vector3f&, float w );
	        Vector4f& setXYZ( const Vector3f& );
	        Vector4f& set( int, float );

	        void      getXYZW( float& x, float& y, float& z, float& w )   const;
	        void      getXYZW( float xyzw[4] )                            const;
	        void      getXYZ( Vector3f& )                                 const;
	        Vector3f  getXYZ()                                            const;
	        void      getXYZdivW( Vector3f& )                             const;
	        Vector3f  getXYZdivW()                                        const;

	        float     getX()                                              const;
	        float     getY()                                              const;
	        float     getZ()                                              const;
	        float     getW()                                              const;
	        float     get( int )                                          const;


/// arithmetic -------------------------------------------------------------------------------------
	        float     dot( const Vector4f& )                              const;
	        float     distance( const Vector4f& )                         const;
	        float     length()                                            const;

	        float     total()                                             const;
	        float     average()                                           const;
	        float     smallest()                                          const;
	        float     largest()                                           const;

	        Vector4f  unitized()                                          const;
	        Vector4f& unitizeEq();

	        //Vector4f  cross( const Vector4f& )                            const;
	        //Vector4f& crossEq( const Vector4f& );

	        Vector4f  abs()                                               const;
	        Vector4f& absEq();

	        Vector4f  operator-()                                         const;
	        Vector4f& negate();

	        Vector4f  operator+ ( const Vector4f& )                       const;
	        Vector4f& operator+=( const Vector4f& );
	        Vector4f  operator- ( const Vector4f& )                       const;
	        Vector4f& operator-=( const Vector4f& );

	        Vector4f  operator* ( const Vector4f& )                       const;
	        Vector4f& operator*=( const Vector4f& );
	        Vector4f  operator/ ( const Vector4f& )                       const;
	        Vector4f& operator/=( const Vector4f& );

	        Vector4f  operator* ( float )                                 const;
	        Vector4f& operator*=( float );
	        Vector4f  operator/ ( float )                                 const;
	        Vector4f& operator/=( float );

	        Vector4f  operator^ ( float )                                 const;
	        Vector4f& operator^=( float );
	        Vector4f  operator^ ( const Vector4f& )                       const;
	        Vector4f& operator^=( const Vector4f& );

	friend  Vector4f  operator*( float, const Vector4f& );
	friend  Vector4f  operator/( float, const Vector4f& );


/// logical ---------------------------------------------------------------------------------------
	        bool      operator==( const Vector4f& )                       const;
	        bool      operator!=( const Vector4f& )                       const;
	        bool      isZero()                                            const;


/// clamps -----------------------------------------------------------------------------------------
	        Vector4f& clampToMinOf( const Vector4f& );
	        Vector4f& clampToMaxOf( const Vector4f& );
	        Vector4f& clampBetween( const Vector4f& min, const Vector4f& max );
	        /**
	         * 0 to almost 1, ie: [0,1).
	         */
	        Vector4f& clamp01();
	        /**
	         * 0 to almost 1, ie: [0,1).
	         */
	        void      clamp01( Vector4f& result )                         const;


/// constants --------------------------------------------------------------------------------------
	static const Vector4f& ZERO();
	static const Vector4f& HALF();
	static const Vector4f& ONE();
	static const Vector4f& EPSILON();
	static const Vector4f& ALMOST_ONE();
	static const Vector4f& MIN();
	static const Vector4f& MAX();
	static const Vector4f& X();
	static const Vector4f& Y();
	static const Vector4f& Z();
	static const Vector4f& W();


/// serialization ----------------------------------------------------------------------------------
	friend  SerializerOut& operator<<( SerializerOut&, const Vector4f& );
	friend  SerializerIn&  operator>>( SerializerIn&,        Vector4f& );
	static const char* CLASS_NAME();
	static const char* CLASS_DESCRIPTION();


/// fields -----------------------------------------------------------------------------------------
private:
	float elements_m[4];
};








/// friends
Vector4f operator*( float, const Vector4f& );
Vector4f operator/( float, const Vector4f& );




/// serialization ----------------------------------------------------------------------------------
SerializerOut& operator<<( SerializerOut&, const Vector4f& );
SerializerIn&  operator>>( SerializerIn&,        Vector4f& );

std::ostream& operator<<( std::ostream&, const Vector4f& );
std::istream& operator>>( std::istream&,       Vector4f& );




/// INLINES ///


/// standard object services -----------------------------------------------------------------------
inline
Vector4f::~Vector4f()
{
}


inline
Vector4f::Vector4f( const Vector4f& other )
{
	operator=( other );
}




/// access -----------------------------------------------------------------------------------------
inline
Vector4f& Vector4f::set( const int i, const float f )
{
	elements_m[ i & 0x03 ] = f;

	return *this;
}


inline
void Vector4f::getXYZ( Vector3f& v3 ) const
{
	v3.setXYZ( elements_m );
}


inline
Vector3f Vector4f::getXYZ() const
{
	return Vector3f( elements_m );
}


inline
float Vector4f::getX() const
{
	return elements_m[0];
}


inline
float Vector4f::getY() const
{
	return elements_m[1];
}


inline
float Vector4f::getZ() const
{
	return elements_m[2];
}


inline
float Vector4f::getW() const
{
	return elements_m[3];
}


inline
float Vector4f::get( const int i ) const
{
	return elements_m[ i & 0x03 ];
}


}}




#endif//Vector4f_h
