/*--------------------------------------------------------------------

   Perceptuum renderer, version 2.1
   Copyright (c) 2002-2003,  Harrison Ainsworth.

   www.hxa7241.org

--------------------------------------------------------------------*/

/*--------------------------------------------------------------------

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston,
   MA 02111-1307, USA

--------------------------------------------------------------------*/


#include "FpToInt.h"
#include "Serialization.h"
#include "hxa7241math.h"

#include "XyzE.h"


using namespace hxa7241::graphics;




/// statics ----------------------------------------------------------------------------------------
dword XyzE::toXyzE( const float x, const float y, const float z )
{
	const float xyz[3] = { x, y, z };
	return toXyzE( xyz );
}


dword XyzE::toXyzE( const float xyz[3] )
{
	const float x = xyz[0] > 0.0f ? xyz[0] : 0.0f;
	const float y = xyz[1] > 0.0f ? xyz[1] : 0.0f;
	const float z = xyz[2] > 0.0f ? xyz[2] : 0.0f;

	const float largestOfxyz  =  (x >= y) ? ( x >= z ? x : z ) : ( y >= z ? y : z );

	dword xyze = 0;

	if( largestOfxyz != 0.0f )
	{
		dword exponentOfLargest = 0;
		float mantissaOfLargest = float(::frexp( largestOfxyz, &exponentOfLargest ));

		if( mantissaOfLargest == 1.0f )   /// shouldnt be needed, but...
		{
			mantissaOfLargest = 0.5f;
			exponentOfLargest++;
		}

		const float amount = mantissaOfLargest * 256.0f / largestOfxyz;

		using hxa7241::utility::fpToIntTrunc;
		xyze = ( fpToIntTrunc( x * amount ) << 24 ) |
			   ( fpToIntTrunc( y * amount ) << 16 ) |
			   ( fpToIntTrunc( z * amount ) <<  8 ) |
			   ( exponentOfLargest + 128 );
	}

	return xyze;
}




void XyzE::fromXyzE( const dword xyze, float& x, float& y, float& z )
{
	float xyz[3];
	fromXyzE( xyze, xyz );

	x = xyz[0];
	y = xyz[1];
	z = xyz[2];
}


float* XyzE::fromXyzE( const dword xyze, float xyz[3] )
{
	const dword exponentOfLargest = xyze & 0x000000FF;

	if( exponentOfLargest == 0 )
	{
		xyz[0] = 0.0f;
		xyz[1] = 0.0f;
		xyz[2] = 0.0f;
	}
	else
	{
		const float oneOver256 = 1.0f / 256.0f;

		const float amount = float(::ldexp( oneOver256, exponentOfLargest - 128 ));

		xyz[0] = (float((xyze >> 24) & 0xFF) + 0.5f) * amount;
		xyz[1] = (float((xyze >> 16) & 0xFF) + 0.5f) * amount;
		xyz[2] = (float((xyze >>  8) & 0xFF) + 0.5f) * amount;
	}

	return xyz;
}




/// serialization ----------------------------------------------------------------------------------
SerializerOut& hxa7241::graphics::operator<<( SerializerOut& out, const XyzE& obj )
{
	out.write( "xyzE", obj.xyze_m );

	return out;
}


SerializerIn& hxa7241::graphics::operator>>( SerializerIn& in, XyzE& obj )
{
	in.read( "xyzE", obj.xyze_m );

	return in;
}


const char* XyzE::CLASS_NAME()
{
	static const char CLASS_NAME[] = "hxa7241::graphics::XyzE";
	return CLASS_NAME;
}


const char* XyzE::CLASS_DESCRIPTION()
{
	static const char CLASS_DESCRIPTION[] = "a greg ward realpixel value";
	return CLASS_DESCRIPTION;
}




/*VOID  GW32i::Compress ( const POINT3i& Triple,   DWORD& Value )   // static
{

   FP     x = Triple.GetX(),   y = Triple.GetY(),   z = Triple.GetZ();

   FP     LargestOfxyz;
   INT    ExponentOfLargest;
   FP     MantissaOfLargest,
          Amount;



   if( x  >=  y ) {
      if( x  >=  z )      LargestOfxyz  =  x;
      else                LargestOfxyz  =  z;
   }else
      if( y  >=  z )      LargestOfxyz  =  y;
      else                LargestOfxyz  =  z;
   //LargestOfxyz   =   x >= y  ?  ( x >= z  ?  x : z ) : ( y >= z  ?  y : z );


   if( LargestOfxyz  ==  (FP)0.0 )//< 1e-32 )
   {
      Value  =  0;
   }
   else
   {
      MantissaOfLargest  =  ::frexpf( LargestOfxyz,  &ExponentOfLargest );

      if( MantissaOfLargest  ==  (FP)1.0 )
          MantissaOfLargest   =  (FP)0.5,
          ExponentOfLargest  +=  1;

      Amount  =  MantissaOfLargest * (FP)256.0 / LargestOfxyz;

      Value   =  (  ( FPk::FPtoDWORD( x * Amount ) ) << 24  )           //  xCmpr    yCmpr    zCmpr    ExpLrg
              |  (  ( FPk::FPtoDWORD( y * Amount ) ) << 16  )           // 00000000,00000000,00000000,00000000
              |  (  ( FPk::FPtoDWORD( z * Amount ) ) <<  8  )
              |  (    (DWORD)( ExponentOfLargest + 128 )   );
      //Value   =  (  ( (DWORD)( x * Amount ) ) << 24  )           //  xCmpr    yCmpr    zCmpr    ExpLrg
      //        |  (  ( (DWORD)( y * Amount ) ) << 16  )           // 00000000,00000000,00000000,00000000
      //        |  (  ( (DWORD)( z * Amount ) ) <<  8  )
      //        |  (    (DWORD)( ExponentOfLargest + 128 )   );
   }

}*/


/*VOID  GW32i::Expand ( DWORD Value,   POINT3i& Triple )   // static
{

   static const FLT   OneOver256_k  =  (FP)1.0 / (FP)256.0;

   INT  ExponentOfLargest;
   FP   Amount;


                                                             //  xCmpr    yCmpr    zCmpr    ExpLrg
   ExponentOfLargest  =  (INT)( Value & 0x000000FF );        // 00000000,00000000,00000000,11111111

   if( ExponentOfLargest  ==  0 )
   {
      Triple  =  POINT3i::Zero_sk;
   }
   else
   {
      Amount  =  ::ldexpf( OneOver256_k,  ExponentOfLargest - 128 );

      Triple.SetXYZ( ( (FP)( (BYTE)(( Value >> 24 )              ) ) + (FP)0.5 ) * Amount,
                     ( (FP)( (BYTE)(( Value >> 16 ) & 0x000000FF ) ) + (FP)0.5 ) * Amount,
                     ( (FP)( (BYTE)(( Value >>  8 ) & 0x000000FF ) ) + (FP)0.5 ) * Amount  );
   }

}*/
