/*--------------------------------------------------------------------

   Perceptuum renderer, version 2.1
   Copyright (c) 2002-2003,  Harrison Ainsworth.

   www.hxa7241.org

--------------------------------------------------------------------*/

/*--------------------------------------------------------------------

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston,
   MA 02111-1307, USA

--------------------------------------------------------------------*/


#include "RenderingState.h"
#include "RenderingStatus.h"
#include "RenderKernel.h"
#include "ImageAccumulation.h"
#include "TimePoint.h"
#include "Exception.h"
#include "Logger.h"

/// test ///
//#include "BssdfConstant.h"
//#include "BssdfSpike.h"
//#include "BssdfPhong.h"
//#include "BssdfSchlick.h"
//#include "BssdfWard.h"
//#include "HistogramSmall.h"
//#include "SobolSequence.h"
//#include "HaltonSequence.h"
//#include "RandomFast.h"
//#include "RandomGood.h"
//#include "ImageCieXyz.h"
//#include "SamplingControl.h"


#include "Render.h"


using namespace perceptuum2::control;




/// statics
const char Render::RENDER_STARTED[]     = "= render started ";
const char Render::RENDER_DONE_FRAME[]  = "= rendered frame ";
const char Render::RENDER_COMPLETED[]   = "= render completed ";
const char Render::RENDER_INTERRUPTED[] = "= render interrupted ";
const char Render::AT_FRAME_END[]       = "at frame end ";
const char Render::AT_LINE_END[]        = "at line end ";
const char Render::AT[]                 = "at ";
const char Render::DURATION[]           = "total work duration ";




/// standard object services -----------------------------------------------------------------------
Render::Render()
{
}


Render::~Render()
{
}




/// main command -----------------------------------------------------------------------------------
void Render::render
(
	const Scene&            scene,
	const RenderingOptions& options,
	RenderingStatus&        status,
	ImageCieXyz&            ,//image,
	RenderingState&         state
)
{
	try
	{
		if( !state.getImageAccumulation().isCompleted() )
		{
			RenderKernel engine( scene, options, state );

			hxa7241::utility::Logger() << RENDER_STARTED << AT << hxa7241::utility::TimePoint().getStringRFC1123();
			status.start();

			attendToFrameDone( true );

			do
			{
				engine.renderFrame();


				status.setCompletedFrames( engine.getFrameCount() );
				status.setCompletedProportion( engine.getCompletedRatio() );

				attendToFrameDone( true );

				hxa7241::utility::Logger() << RENDER_DONE_FRAME << engine.getFrameCount() << ' ' << AT << hxa7241::utility::TimePoint().getStringRFC1123() <<
					                          "; " << DURATION << hxa7241::utility::TimePoint::getStringDuration( status.getDuration() );
			}
			while( shouldGo() & !engine.isCompleted() );

			status.stop();

			if( engine.isCompleted() )
			{
				hxa7241::utility::Logger() << RENDER_COMPLETED << AT << hxa7241::utility::TimePoint().getStringRFC1123() <<
				                              "; " << DURATION << hxa7241::utility::TimePoint::getStringDuration( status.getDuration() );
				status.setFinished();
			}
			else
			{
				hxa7241::utility::Logger() << RENDER_INTERRUPTED << AT_FRAME_END << AT << hxa7241::utility::TimePoint().getStringRFC1123() <<
				                              "; " << DURATION << hxa7241::utility::TimePoint::getStringDuration( status.getDuration() );
			}
		}

		attendToRenderDone( true );
	}
	catch( const std::exception& e )
	{
		hxa7241::utility::Logger() << hxa7241::utility::Logger::EXC2 << e.what();
	}
	catch( ... )
	{
		hxa7241::utility::Exception e;
		hxa7241::utility::Logger() << hxa7241::utility::Logger::EXC2 << e.what();
	}
}


/*void Render::render
(
	const Scene&            scene,
	const RenderingOptions& options,
	RenderingStatus&        status,
	ImageCieXyz&            image,
	RenderingState&         state
)
{
	try
	{
		//perceptuum2::transport::SamplingControl::test( true );


		//hxa7241::utility::HistogramSmall::test( true );


//		{
//			hxa7241::utility::HaltonSequence hhh[15];
//			for( int i = 0;  i < 15;  ++i )
//			{
//				hhh[i].setBase( i );
//				hhh[i].setIndex( 0 );
//			}
//
//			for( int j = 0;  j < 100;  ++j )
//			{
//				hxa7241::utility::Logger lll;
//				lll << "halton " << j << ":  ";
//
//				for( int i = 0;  i < 15;  ++i )
//				{
//					lll << hhh[i].getItem() << "  ";
//					hhh[i].next();
//				}
//			}
//		}


		//hxa7241::utility::SobolSequence::test( true );


//		const hxa7241::utility::Sheet<byte> byteSheet( image.getWidth(), image.getHeight() );
//		hxa7241::utility::SobolSequence::testView( 17, 18, 100, byteSheet );
//		for( dword i = image.getLength();  i-- > 0; )
//		{
//			image[i] = CieXyz::ONE() * float(ubyte(byteSheet[i]));
//		}

		/// iter 100
		/// yes: 2,3  3,4  4,5  5,6  6,7  9,10  11,12  12,13  13,14  14,15  16,17  17,18  26,27  27,28  31,32
		///	     32,35, 35,36  36,37  37,38  38,39
		/// umm: 0,1    8,9  15,16  21,22  23,24  24,25
		/// no:  7,8  10,11  28,29  29,30  >=40

		// 24 29 30 32
		// 35 36 37 38 39



//		const hxa7241::utility::Sheet<byte> byteSheet( image.getWidth(), image.getHeight() );
//		{
//			const dword dim0 = 6;
//			const dword dim1 = 7;
//			const dword iterations = 100;
//
//			byteSheet.zeroMemory();
//
//			hxa7241::utility::HaltonSequence hs0( dim0 );
//			hxa7241::utility::HaltonSequence hs1( dim1 );
//
//			const dword width  = byteSheet.getWidth();
//			const dword height = byteSheet.getHeight();
//
//			for( dword i = iterations;  i-- > 0; )
//			{
//				dword x = dword(hs0.getItem() * float(width));
//				dword y = dword(hs1.getItem() * float(height));
//
//				x = x < width  ? x : width-1;
//				y = y < height ? y : height-1;
//
//				byteSheet.getElement( x, y ) = 0xFF;
//
//				hs0.next();
//				hs1.next();
//			}
//		}
//		for( dword i = image.getLength();  i-- > 0; )
//		{
//			image[i] = CieXyz::ONE() * float(ubyte(byteSheet[i]));
//		}

		/// iter 100
		/// yes: 0,1  1,2  2,3, 3,4  1,5  8,3  2,6
		/// no:  4,5  5,6  6,7  7,8  8,9  9,10


//		{
//			const dword iterations = 1000;
//
//			image.zeroMemory();
//
//			hxa7241::utility::RandomGood rf0( 0x0380E038 );
//
//			const dword width  = image.getWidth();
//			const dword height = image.getHeight();
//
//			for( dword i = iterations;  i-- > 0; )
//			{
//				dword x = dword(rf0.getItem() * float(width));
//				rf0.next();
//				dword y = dword(rf0.getItem() * float(height));
//				rf0.next();
//
//				x = x < width  ? x : width-1;
//				y = y < height ? y : height-1;
//
//				image.getElement( x, y ) = CieXyz::ONE() * 100.0f;
//			}
//		}


		using perceptuum2::interaction::Light;
		using hxa7241::graphics::UnitVector3f;
		using hxa7241::graphics::CieXyz;

//		perceptuum2::interaction::BssdfConstant bl( UnitVector3f::Y(), UnitVector3f::X(),
//		                                            1.0f, 1.0f, 0.0f,
//		                                            Light( CieXyz(0.75, 0.5, 0.25) ) );
//		bl.test( true );
//		/// isBrdfNotSdf, partitionsBits, isCos, isTransmission
//		bl.testView( true, 0xFF, false, false, image );
//		perceptuum2::interaction::BssdfSpike bm( UnitVector3f::Y(), UnitVector3f::X(),
//		                                         1.0f, 1.0f, 0.0f,
//		                                         Light( CieXyz(0.75, 0.5, 0.25) ) );
//		bm.test( true );
//		bm.testView( false, 0xFF, false, false, image );


//		perceptuum2::interaction::BssdfPhong bp( UnitVector3f::Y(), UnitVector3f::X(),
//		                                         1.0f, 1.0f, 0.0f,
//		                                         0.5f, Light(CieXyz(10,5,-2)), Light(CieXyz(-2,1,0.5)), 1.0f );
//		bp.test( true );
//		perceptuum2::interaction::BssdfPhong bp2( UnitVector3f::Y(), UnitVector3f::X(),
//		                                          1.0f, 1.0f, 0.0f,
//		                                          0.5f, Light::ONE(), Light::ONE(), 0.0f );
//		bp2.test( true );
//		perceptuum2::interaction::BssdfPhong bp3( UnitVector3f::Y(), UnitVector3f::X(),
//		                                          1.0f, 1.0f, 0.0f,
//		                                          0.5f, Light::ONE(), Light::ONE(), 0.75f );
//		bp3.test( true );
//		bp3.testView( true, 0xFF, true, false, image );
//		perceptuum2::interaction::BssdfPhong bp4( UnitVector3f::Y(), UnitVector3f::X(),
//		                                          1.0f, 1.0f, 0.5f,
//		                                          0.5f, Light(CieXyz(1,1,1)), Light(CieXyz(1,1,1)), 0.5f );
//		//bp4.test( true );
//		bp4.testView( false, 0xFF, true, true, image );
//		perceptuum2::interaction::BssdfPhong bp5( UnitVector3f::Y(), UnitVector3f::X(),
//		                                          1.0f, 1.0f, 0.0f,
//		                                          0.5f, Light::ONE(), Light::ONE(), 0.25f );
//		bp5.test( true );
//		perceptuum2::interaction::BssdfPhong bp6( UnitVector3f::Y(), UnitVector3f::X(),
//		                                          1.0f, 1.0f, 0.0f,
//		                                          0.5f, Light::ONE(), Light::ONE(), 0.01f );
//		bp6.test( true );
//		bp6.testView( false, 0xFF, true, false, image );


//		perceptuum2::interaction::BssdfSchlick bs1( UnitVector3f::Y(), UnitVector3f::X(),
//		                                            1.0f, 1.0f, 0.0f,
//		                                            Light::ONE() / 2.0f, 1.0f, 0.0f );
//		bs1.test( true );
//		perceptuum2::interaction::BssdfSchlick bs2( UnitVector3f::Y(), UnitVector3f::X(),
//		                                            1.0f, 1.0f, 0.0f,
//		                                            Light::ONE() / 2.0f, 0.0f, 0.0f );
//		bs2.test( true );
//		perceptuum2::interaction::BssdfSchlick bs3( UnitVector3f::Y(), UnitVector3f::X(),
//		                                            1.0f, 1.0f, 0.0f,
//		                                            Light::ONE() / 2.0f, 0.5f, 0.0f );
//		bs3.test( true );


//		perceptuum2::interaction::BssdfWard bw( UnitVector3f::Y(), UnitVector3f::X(),
//		                                        1.0f, 1.0f, 0.0f,
//		                                        Light::ONE() / 4.0f, Light::ONE() / 4.0f, 0.25f, 0.75f );
//		bw.test( true );
//		bw.testView( false, 0xFF, true, false, image );
//		bw.setBssdfWard( Light::ZERO(), Light::ONE() / 2.0f, 0.001f, 0.001f );
//		bw.test( true );
//		bw.setBssdfWard( Light::ZERO(), Light::ONE() / 2.0f, 0.025f, 0.025f );
//		bw.test( true );
//		bw.setBssdfWard( Light::ZERO(), Light::ONE() / 2.0f, 0.5f, 0.5f );
//		bw.test( true );
//		bw.setBssdfWard( Light::ZERO(), Light::ONE() / 2.0f, 1.0f, 1.0f );
//		bw.test( true );


		image.setAlphaOpaque();

		attendToRenderDone( true );
	}
	catch( const std::exception& e )
	{
		hxa7241::utility::Logger() << hxa7241::utility::Logger::EXC2 << e.what();
	}
	catch( ... )
	{
		hxa7241::utility::Exception e;
		hxa7241::utility::Logger() << hxa7241::utility::Logger::EXC2 << e.what();
	}
}*/




/*void Render::render
(
	const Scene&            ,//scene,
	const RenderingOptions& ,//options,
	RenderingStatus&        status,
	ImageCieXyz&            image,
	RenderingState&         //state
)
{
//	const dword time = dword(std::time(0));
//	hxa7241::utility::Logger() << "duration " << -(time & 0xFFFF) << "s = " << hxa7241::utility::TimePoint::getStringDuration( -(time & 0xFFFF) );
//
//	using hxa7241::utility::Float01c;
//	using hxa7241::utility::Float01;
//	hxa7241::utility::Logger() << "Float01c " << Float01c(-9.0f) << "  " << Float01c(9.0f) << "  " << Float01c(1.0f) << "  " << ((1.0f - Float01c(1.0f)) == 0.0f);
//	hxa7241::utility::Logger() << "Float01 " << Float01(-9.0f) << "  " << Float01(9.0f) << "  " << Float01(1.0f) << "  " << ((1.0f - Float01(1.0f)) == 0.0f);
//	hxa7241::utility::Logger() << "Float01 Float01c " << (Float01(-9.0f) == Float01c(-9.0f)) << "  " << (Float01(9.0f) == Float01c(9.0f));

//	hxa7241::utility::RandomGood randtest( dword(std::time(0)) );
//	for( int i = 100;  i-- > 0; ) randtest.next();
//
//	hxa7241::utility::ByteSerializerOut::serializeOut( randtest, "randomgoodtest.byte" );
//	hxa7241::utility::RandomGood randtestin( 1 );
//	hxa7241::utility::ByteSerializerIn::serializeIn( randtestin, "randomgoodtest.byte" );
//
//	for( int i = 0;  i < 32;  ++i )
//	{
//		hxa7241::utility::Logger() << "rand " << i << "  " << randtest.getDword() << "  " << randtestin.getDword() << (randtest.getDword() != randtestin.getDword() ? "  WRONG" : "  ok");
//		randtest.next();
//		randtestin.next();
//	}

//	{
//		using hxa7241::utility::WeightedComposition;
//		std::vector<float> s1;
//		std::vector<float> s2;
//		WeightedComposition wc1( s1, 20 );
//		WeightedComposition wc2( s2, 20 );
//
//		static const float f1[] = { 1, 2, 3, 4, 0, 0, 7, 8, 9, 0 };
//		static const float f2[] = { 0, 0, 2, 3, 4, 5, 6, 7, 0, 0 };
//		for( int i = 0;  i < 10;  ++i )
//		{
//			wc1.addElement( f1[i] );
//			wc2.addElement( f2[i] );
//		}
//		for( int i = 0;  i < 10;  ++i )
//		{
//			hxa7241::utility::Logger() << "wc " << i << "  " << s1[i] << "  " << s2[i];
//		}
//
//		dword ei1[10] = { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 };
//		float ef1[10] = { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 };
//		dword ei2[10] = { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 };
//		float ef2[10] = { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 };
//		hxa7241::utility::RandomGood rrr( dword(std::time(0)) );
//		const sss = 100000;
//		for( int i = sss;  i-- > 0; )
//		{
//			rrr.next();
//
//			dword ii1;
//			float ff1;
//			wc1.findElement( rrr.getFloat(), ii1, ff1 );
//			ei1[ii1] += 1;
//
//			dword ii2;
//			float ff2;
//			wc2.findElement( rrr.getFloat(), ii2, ff2 );
//			ei2[ii2] += 1;
//
//			if( i < 10 )
//			{
//				hxa7241::utility::Logger() << "finding " << i << "  " << rrr.getFloat() << "  (" << ff1 << " " << ii1 << " " << ei1[ii1] << ")  (" << ff2 << " " << ii2 << " " << ei2[ii2] << ")";
//			}
//		}
//
//		for( int i = 0;  i < 10;  ++i )
//		{
//			hxa7241::utility::Logger() << "found " << i << "  [" << (ei1[i]/float(sss)) << " " << (f1[i]/34.0f) << "]  [" << (ei2[i]/float(sss)) << " " << (f2[i]/27.0f) << "]";
//		}
//
//
//		dword ii;
//		float ff;
//		wc2.findElement( 0, ii, ff );
//		hxa7241::utility::Logger() << "found " << 0 << "  " << ii << " " << ff;
//		wc2.findElement( 0.074, ii, ff );
//		hxa7241::utility::Logger() << "found " << 0.074 << "  " << ii << " " << ff;
//		wc2.findElement( 0.0741, ii, ff );
//		hxa7241::utility::Logger() << "found " << 0.0741 << "  " << ii << " " << ff;
//		wc2.findElement( FLOAT_ALMOST_ONE, ii, ff );
//		hxa7241::utility::Logger() << "found " << FLOAT_ALMOST_ONE << "  " << ii << " " << ff;
//	}

//	{
//		using hxa7241::graphics::Transplace;
//		using hxa7241::graphics::Vector3f;
//		using hxa7241::graphics::UnitVector3f;
//		Transplace trp1;
//		//Transplace trp2;
//
//		//hxa7241::utility::Logger() << "degreesToRadians " << 89 << " = " << hxa7241::graphics::degreesToRadians( 89 ) << "  " << 449 << " = " << hxa7241::graphics::degreesToRadians( 449 );
//
//		trp1.setTransplace( Vector3f( 123, -234, 345 ), Vector3f( -3, 4, -5 ) );
//		Transplace trp2( trp1, true );
//		//trp2 = trp1;
//		//trp2.invert();
//
//		hxa7241::utility::Logger() << "trp1 " << trp1.matrix_m.getRow0() << " " << trp1.matrix_m.getRow1() << " " << trp1.matrix_m.getRow2() << " " << trp1.matrix_m.getCol3();
//		hxa7241::utility::Logger() << "trp2 " << trp2.matrix_m.getRow0() << " " << trp2.matrix_m.getRow1() << " " << trp2.matrix_m.getRow2() << " " << trp2.matrix_m.getCol3();
//
//		UnitVector3f v1( 1, 2, 3 );
//		hxa7241::utility::Logger() << "v1 " << v1;
//		trp1.transform( v1 );
//		hxa7241::utility::Logger() << "v1 " << v1;
//		trp2.transform( v1 );
//		hxa7241::utility::Logger() << "v1 " << v1;
//
//		trp2 += trp1;
//		hxa7241::utility::Logger() << "trp2 " << trp2.matrix_m.getRow0() << " " << trp2.matrix_m.getRow1() << " " << trp2.matrix_m.getRow2() << " " << trp2.matrix_m.getCol3();
//
//
//		trp1.setTransplace( Vector3f( 0, 0, 345 ), Vector3f( 0, 0, 0 ) );
//		trp2.setTransplace( Vector3f( 0, 0, -345 ), Vector3f( 0, 0, 0 ) );
//
//		hxa7241::utility::Logger() << "trp1 " << trp1.matrix_m.getRow0() << " " << trp1.matrix_m.getRow1() << " " << trp1.matrix_m.getRow2() << " " << trp1.matrix_m.getCol3();
//		hxa7241::utility::Logger() << "trp2 " << trp2.matrix_m.getRow0() << " " << trp2.matrix_m.getRow1() << " " << trp2.matrix_m.getRow2() << " " << trp2.matrix_m.getCol3();
//
//		trp2 += trp1;
//		hxa7241::utility::Logger() << "trp2 " << trp2.matrix_m.getRow0() << " " << trp2.matrix_m.getRow1() << " " << trp2.matrix_m.getRow2() << " " << trp2.matrix_m.getCol3();
//	}



	//perceptuum2::imaging::ImageAccumulation& ia = state.getImageAccumulation();

//	{
//		using perceptuum2::transport::SamplingControl;
//		//SamplingControl sc;
//		SamplingControl& sc = state.getSamplingControl();
//		hxa7241::utility::Logger() << "scr: " << sc.getFrameIndex();
//
//		//sc.switchIllumination( sc.TWO );
//
//		for( int j = 10;  j-- > 0; )
//		{
//			sc.advanceSequence( sc.FRAME );
//			sc.advanceSequence( sc.LIGHT );
//
//			sc.switchIllumination( sc.ONE );
//			sc.advanceSequence( sc.ILLUM );
//
//			sc.switchIllumination( sc.TWO );
//			sc.advanceSequence( sc.ILLUM );
//		}
//
//		//hxa7241::utility::XmlSerializerOut::serializeOut( sc, "samplingcontrol.xml" );
//		hxa7241::utility::ByteSerializerOut::serializeOut( sc, "samplingcontrol" );
//
//		SamplingControl sc1;
//
//		hxa7241::utility::ByteSerializerIn::serializeIn( sc1, "samplingcontrol" );
//		//sc1 = sc;
//
//		bool bb = true;
//		for( int j = 10;  j-- > 0; )
//		{
//			sc.advanceSequence( sc.FRAME );
//			sc1.advanceSequence( sc.FRAME );
//			sc.advanceSequence( sc.LIGHT );
//			sc1.advanceSequence( sc.LIGHT );
//
//			sc.switchIllumination( sc.ONE );
//			sc1.switchIllumination( sc.ONE );
//			sc.advanceSequence( sc.ILLUM );
//			sc1.advanceSequence( sc.ILLUM );
//			bb &= sc1.getSamplingPoint( sc.TIME ) == sc.getSamplingPoint( sc.TIME );
//			bb &= sc1.getSamplingPoint( sc.SCREEN_X ) == sc.getSamplingPoint( sc.SCREEN_X );
//			bb &= sc1.getSamplingPoint( sc.SCREEN_Y ) == sc.getSamplingPoint( sc.SCREEN_Y );
//			bb &= sc1.getSamplingPoint( sc.LENS_X ) == sc.getSamplingPoint( sc.LENS_X );
//			bb &= sc1.getSamplingPoint( sc.LENS_Y ) == sc.getSamplingPoint( sc.LENS_Y );
//			bb &= sc1.getSamplingPoint( sc.SPECULARNESS ) == sc.getSamplingPoint( sc.SPECULARNESS );
//			bb &= sc1.getSamplingPoint( sc.LIGHTSOURCE_POS_X ) == sc.getSamplingPoint( sc.LIGHTSOURCE_POS_X );
//			bb &= sc1.getSamplingPoint( sc.LIGHTSOURCE_POS_Y ) == sc.getSamplingPoint( sc.LIGHTSOURCE_POS_Y );
//			bb &= sc1.getSamplingPoint( sc.LIGHTSOURCE_POS_T ) == sc.getSamplingPoint( sc.LIGHTSOURCE_POS_T );
//			bb &= sc1.getSamplingPoint( sc.LIGHTSOURCE_DIR_X ) == sc.getSamplingPoint( sc.LIGHTSOURCE_DIR_X );
//			bb &= sc1.getSamplingPoint( sc.LIGHTSOURCE_DIR_Y ) == sc.getSamplingPoint( sc.LIGHTSOURCE_DIR_Y );
//			bb &= sc1.getSamplingPoint( sc.LIGHT_INTERACTION_X ) == sc.getSamplingPoint( sc.LIGHT_INTERACTION_X );
//			bb &= sc1.getSamplingPoint( sc.LIGHT_INTERACTION_Y ) == sc.getSamplingPoint( sc.LIGHT_INTERACTION_Y );
//			bb &= sc1.getSamplingPoint( sc.LIGHT_INTERACTION_T ) == sc.getSamplingPoint( sc.LIGHT_INTERACTION_T );
//			bb &= sc1.getSamplingPoint( sc.EYE_INTERACTION_X ) == sc.getSamplingPoint( sc.EYE_INTERACTION_X );
//			bb &= sc1.getSamplingPoint( sc.EYE_INTERACTION_Y ) == sc.getSamplingPoint( sc.EYE_INTERACTION_Y );
//			bb &= sc1.getSamplingPoint( sc.EYE_INTERACTION_T ) == sc.getSamplingPoint( sc.EYE_INTERACTION_T );
//		hxa7241::utility::Logger() << "== " << sc1.getSamplingPoint( sc.LIGHTSOURCE_DIR_Y ) << " " << sc.getSamplingPoint( sc.LIGHTSOURCE_DIR_Y );
//
//			sc.switchIllumination( sc.TWO );
//			sc1.switchIllumination( sc.TWO );
//			sc.advanceSequence( sc.ILLUM );
//			sc1.advanceSequence( sc.ILLUM );
//			bb &= sc1.getSamplingPoint( sc.LIGHTSOURCE_POS_X ) == sc.getSamplingPoint( sc.LIGHTSOURCE_POS_X );
//			bb &= sc1.getSamplingPoint( sc.LIGHTSOURCE_POS_Y ) == sc.getSamplingPoint( sc.LIGHTSOURCE_POS_Y );
//			bb &= sc1.getSamplingPoint( sc.LIGHTSOURCE_POS_T ) == sc.getSamplingPoint( sc.LIGHTSOURCE_POS_T );
//			bb &= sc1.getSamplingPoint( sc.LIGHTSOURCE_DIR_X ) == sc.getSamplingPoint( sc.LIGHTSOURCE_DIR_X );
//			bb &= sc1.getSamplingPoint( sc.LIGHTSOURCE_DIR_Y ) == sc.getSamplingPoint( sc.LIGHTSOURCE_DIR_Y );
//		}
//
//		hxa7241::utility::Logger() << "copy check: " << bb;
//	}


	hxa7241::utility::RandomGood rand( dword(std::time(0)) );

	try
	{
		hxa7241::utility::Logger() << RENDER_STARTED << AT << hxa7241::utility::TimePoint().getStringRFC1123();
		status.start();

		for( int frameNumber = 0;  frameNumber < 100;  ++frameNumber )
		{
			status.setCurrentFrame( frameNumber +1 );

			const dword w = image.getWidth();
			const dword h = image.getHeight();
			{
				for( dword y = 0;  y < h;  ++y )
				{
					for( dword x = 0;  x < w;  ++x )
					{
						rand.next();
						const float r = rand.getFloat() * 30.0f;
						rand.next();
						const float g = rand.getFloat() * 30.0f;
						rand.next();
						const float b = rand.getFloat() * 30.0f;

						image.getElement( x, y )      = CieXyz( r, g, b );
						image.getAlphaElement( x, y ) = 0xFF;
						//ia.addSample( x, y, 0.5f, 0.5f, hxa7241::graphics::CieXyz(r,g,b), 0xFF );
					}
				}
			}
//			{
//				sc.advanceSequence( sc.FRAME );
//				sc.advanceSequence( sc.LIGHT );
//				sc.advanceSequence( sc.ILLUM );
//
//				const float tPt = sc.getSamplingPoint( sc.TIME );
//				const float sxPt = sc.getSamplingPoint( sc.SCREEN_X );
//				const float syPt = sc.getSamplingPoint( sc.SCREEN_Y );
//				const float lxPt = sc.getSamplingPoint( sc.LENS_X );
//				const float lyPt = sc.getSamplingPoint( sc.LENS_Y );
//				const float specPt = sc.getSamplingPoint( sc.SPECULARNESS );
//				const float lspxPt = sc.getSamplingPoint( sc.LIGHTSOURCE_POS_X );
//				const float lspyPt = sc.getSamplingPoint( sc.LIGHTSOURCE_POS_Y );
//				const float lsptPt = sc.getSamplingPoint( sc.LIGHTSOURCE_POS_T );
//				const float lsdxPt = sc.getSamplingPoint( sc.LIGHTSOURCE_DIR_X );
//				const float lsdyPt = sc.getSamplingPoint( sc.LIGHTSOURCE_DIR_Y );
//				const float lixPt = sc.getSamplingPoint( sc.LIGHT_INTERACTION_X );
//				const float liyPt = sc.getSamplingPoint( sc.LIGHT_INTERACTION_Y );
//				const float litPt = sc.getSamplingPoint( sc.LIGHT_INTERACTION_T );
//				const float eixPt = sc.getSamplingPoint( sc.EYE_INTERACTION_X );
//				const float eiyPt = sc.getSamplingPoint( sc.EYE_INTERACTION_Y );
//				const float eitPt = sc.getSamplingPoint( sc.EYE_INTERACTION_T );
//
//				//hxa7241::utility::Logger() << "FrameIndex: " << sc.getFrameIndex();
//				hxa7241::utility::Logger() << "" << tPt << " " << sxPt << " " << syPt << " " << lxPt << " " << lyPt << " " << specPt << " " << lspxPt << " " << lspyPt << " " << lsptPt << " " << lsdxPt << " " << lsdyPt << " " << lixPt << " " << liyPt << " " << litPt << " " << eixPt << " " << eiyPt << " " << eitPt;
//
//				const dword x = lxPt * w;
//				const dword y = lyPt * h;
//
//				image.getElement( x, y ) = CieXyz( 30, 30, 30 );
//				image.getAlphaElement( x, y ) = 0xFF;
//			}
//			//if( frameNumber == 5 )
//				ia.addSample( 2, 2, 0.5f, 0.5f, hxa7241::graphics::CieXyz(20,20,20), 0.48f );//(frameNumber / 10.0f) );
//
//			if( ((h * w) <= 25) )//&& (frameNumber == 999) )
//			{
//				hxa7241::utility::Logger lll;
//				lll << "\nluminance:\n";
//				for( dword y = 0;  y < h;  ++y )
//				{
//					for( dword x = 0;  x < w;  ++x )
//					{
//						lll << ia.pImage_m->getElement( x, y ).getY() << " ";
//					}
//					lll << "\n";
//				}
//				lll << "\nalpha:\n";
//				for( dword y = 0;  y < h;  ++y )
//				{
//					for( dword x = 0;  x < w;  ++x )
//					{
//						lll << dword(ubyte(ia.pImage_m->getAlphaElement( x, y ))) << " ";
//					}
//					lll << "\n";
//				}
//				lll << "\nvariance:\n";
//				for( dword y = 0;  y < h;  ++y )
//				{
//					for( dword x = 0;  x < w;  ++x )
//					{
//						lll << ia.pixelStats_m.getElement( x, y ).getVariance() << " ";
//					}
//					lll << "\n";
//				}
//				ia.analyse();
//				lll << "variance:\n";
//				for( dword y = 0;  y < h;  ++y )
//				{
//					for( dword x = 0;  x < w;  ++x )
//					{
//						lll << ia.pixelStats_m.getElement( x, y ).getVariance() << " ";
//					}
//					lll << "\n";
//				}
//			}
//			else
//			{
//				ia.analyse();
//			}
//			hxa7241::utility::Logger() << "getCompletedRatio:" << ia.getCompletedRatio() << " " << (ia.getCompletedRatio()>=1.0f) << "  samplesCount:" << ia.getSamplesCount() << "  samplesRequiredCount:" << ia.getSamplesRequiredCount();


			status.setCompletedFrames( frameNumber +1 );
			attendToFrameDone( true );
			hxa7241::utility::Logger() << RENDER_DONE_FRAME << (frameNumber +1) << ' ' << AT << hxa7241::utility::TimePoint().getStringRFC1123() <<
										  ";  " << DURATION << hxa7241::utility::TimePoint::getStringDuration( status.getDuration() );
			if( !shouldGo() )//|| ia.isCompleted() )
			{
				break;
			}
		}

		status.stop();
		hxa7241::utility::Logger() << RENDER_COMPLETED << AT << hxa7241::utility::TimePoint().getStringRFC1123() <<
									  ";  " << DURATION << hxa7241::utility::TimePoint::getStringDuration( status.getDuration() );
		status.setFinished();
		attendToRenderDone( true );
	}
	catch( const std::exception& e )
	{
		hxa7241::utility::Logger() << hxa7241::utility::Logger::EXC2 << e.what();
	}
	catch( ... )
	{
		hxa7241::utility::Exception e;
		hxa7241::utility::Logger() << hxa7241::utility::Logger::EXC2 << e.what();
	}
}*/




/// other commands ---------------------------------------------------------------------------------
void Render::processImage
(
	const void*  ,//options,
	ImageCieXyz& //image
) const
{
}




/// overrideables ----------------------------------------------------------------------------------
bool Render::shouldGo() const
{
	return true;
}


bool Render::shouldStop() const
{
	return false;
}


bool Render::shouldStopAtFrame() const
{
	return false;
}


void Render::attendToLineDone( dword )
{
}


void Render::attendToFrameDone( bool )
{
}


void Render::attendToRenderDone( bool )
{
}
