/*--------------------------------------------------------------------

   Perceptuum renderer, version 2.1
   Copyright (c) 2002-2003,  Harrison Ainsworth.

   www.hxa7241.org

--------------------------------------------------------------------*/

/*--------------------------------------------------------------------

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston,
   MA 02111-1307, USA

--------------------------------------------------------------------*/


#include "Scene.h"
#include "RenderingOptions.h"
#include "RenderingStatus.h"
#include "RenderingState.h"
#include "ImageCieXyz.h"
#include "ImageRgba.h"
#include "ImageWriter.h"
#include "Thread.h"
#include "ThreadLock.h"
#include "Exception.h"
#include "DisplayMeasures.h"
#include "Logger.h"
#include <time>

#include "RenderDesktop.h"


using namespace perceptuum2::control;




/// statics
const char  RenderDesktop::CLASS_NAME[]         = "perceptuum2::control::RenderDesktop";
const char* RenderDesktop::SUMMARY_NAMES[]      = { "options summary", "scene summary", "status summary", "state summary", "image summary" };
const char  RenderDesktop::EXCEPTION_MESSAGE1[] = "rgb image size doesnt match render image size";
const char  RenderDesktop::EXCEPTION_MESSAGE2[] = "rgb image size hasnt been initialized";
const char  RenderDesktop::EXCEPTION_MESSAGE3[] = "render thread already started";




/// standard object services -----------------------------------------------------------------------
RenderDesktop::RenderDesktop
(
	const ThreadFactory& threadFactory,
	const string&        filesPathname
)
 :	Render(),
	Runnable(),
	renderData_m( filesPathname, threadFactory ),
	pThread_m( threadFactory.createThread( this ) ),
	pThreadLock_m( pThread_m->createLock() ),
	dataFilePathname_m(),
	optionsSummary_m(),
	sceneSummary_m(),
	statusSummary_m(),
	stateSummary_m(),
	imageSummary_m(),
	statusLine_m(),
	tonemapAdaptation_m(),
	lastPersistTime_m( 0 ),
	timeBetweenPersists_m(),
	imageRgba_m(),
	imageConverter_m()
{
	/// get time between persists
	timeBetweenPersists_m = renderData_m.getOptions().getPersistPeriod();

	/// set image conversion
	setImageConversion( renderData_m, imageConverter_m );

	copyInfo();
}


RenderDesktop::~RenderDesktop()
{
	delete pThreadLock_m;
	delete pThread_m;
}




/// commands ---------------------------------------------------------------------------------------
void RenderDesktop::setRgbImage
(
	dword*const pPixels,
	const dword width,
	const dword height,
	const ERgbByteFormat rgbByteFormat
)
{
	static const char METHOD_NAME[] = "setRgbImage";

	dword widthCurrent;
	dword heightCurrent;
	getImageSize( widthCurrent, heightCurrent );

	if( (width != widthCurrent) || (height != heightCurrent) )
	{
		throw hxa7241::utility::Exception( EXCEPTION_MESSAGE1, CLASS_NAME, METHOD_NAME );
	}

	imageRgba_m.set( pPixels, width, height, static_cast<ImageRgba::ERgbByteFormat>(rgbByteFormat) );

	/// get the image displayed
	attendToFrameDone( true );
}


void RenderDesktop::start()
{
	static const char METHOD_NAME[] = "start";

	if( !isThreadRunning() )
	{
		if( imageRgba_m.getMemory() != 0 )
		{
			setPermission( GO );

			pThread_m->start();
		}
		else
		{
			throw hxa7241::utility::Exception( EXCEPTION_MESSAGE2, CLASS_NAME, METHOD_NAME );
		}
	}
	else
	{
		throw hxa7241::utility::Exception( EXCEPTION_MESSAGE3, CLASS_NAME, METHOD_NAME );
	}
}


void RenderDesktop::stop()
{
	//if( isThreadRunning() )   /// is it really necessary to know ?
	{
		setPermission( STOP );
	}
}


void RenderDesktop::stopWhenFrameDone()
{
	//if( isThreadRunning() )   /// is it really necessary to know ?
	{
		setPermission( PERMISSION_STOP_AT_FRAME );
	}
}


void RenderDesktop::stopForcibly()
{
	setPermission( STOP );
	pThread_m->stopForcibly();
}


void RenderDesktop::unStop()
{
	//if( isThreadRunning() )   /// is it really necessary to know ?
	{
		setPermission( GO );
	}
}


void RenderDesktop::setTonemapAdaptationLuminance
(
	const float fp
)   /// synchronized
{
	hxa7241::utility::ThreadSync ts( pThreadLock_m );

	tonemapAdaptation_m = fp;

	/// re-convert image
	{
		hxa7241::utility::ThreadSuspension suspension( pThread_m );

		convertImage();
	}
}


void RenderDesktop::persist()   /// synchronized
{
	hxa7241::utility::ThreadSync ts( pThreadLock_m );

	renderData_m.save();
}




/// queries ----------------------------------------------------------------------------------------
bool  RenderDesktop::isThreadRunning() const
{
	return pThread_m->isAlive();
}


void  RenderDesktop::getProjectPathname
(
	string& pathname
) const   /// synchronized
{
	hxa7241::utility::ThreadSync ts( pThreadLock_m );

	pathname = dataFilePathname_m;
}


void RenderDesktop::getStartFileNames
(
	vector<string>& names
)
{
	RenderDataFiling::getStartFileNames( names );
}


void  RenderDesktop::getSummaryNames
(
	vector<string>& names
) const
{
	const dword noOfNames = sizeof(SUMMARY_NAMES) / sizeof(char*);
	if( names.size() != noOfNames )
	{
		names.resize( noOfNames );
	}

	for( dword i = 0;  i < noOfNames;  ++i )
	{
		names[i] = SUMMARY_NAMES[i];
	}
}


void  RenderDesktop::getSummaries
(
	vector<string>& summaries
) const   /// synchronized
{
	hxa7241::utility::ThreadSync ts( pThreadLock_m );

	const dword noOfSummaries = 5;
	if( summaries.size() != noOfSummaries )
	{
		summaries.resize( noOfSummaries );
	}

	summaries[0] = optionsSummary_m;
	summaries[1] = sceneSummary_m;
	summaries[2] = statusSummary_m;
	summaries[3] = stateSummary_m;
	summaries[4] = imageSummary_m;
}


void  RenderDesktop::getStatusLine
(
	string& s
) const   /// synchronized
{
	hxa7241::utility::ThreadSync ts( pThreadLock_m );

	s = statusLine_m;
}


void  RenderDesktop::getFormattingNames
(
	vector<string>& names
) const
{
	perceptuum2::imaging::ImageWriter::getFormatNames( names );
}


void  RenderDesktop::writeImage
(
	const dword format,
	ostream& imageBytes
) const   /// synchronized
{
	hxa7241::utility::ThreadSync ts( pThreadLock_m );

	{
		hxa7241::utility::ThreadSuspension suspension( pThread_m );

		perceptuum2::imaging::ImageWriter::write( renderData_m.getImage(), imageConverter_m, format, imageBytes );
	}
}


float RenderDesktop::getTonemapAdaptationLuminance() const   /// synchronized
{
	hxa7241::utility::ThreadSync ts( pThreadLock_m );

	return tonemapAdaptation_m;
}


float RenderDesktop::getTonemapAveLuminance() const   /// synchronized
{
	hxa7241::utility::ThreadSync ts( pThreadLock_m );

	return renderData_m.getImage().getLogMeanLuminance();
}


void RenderDesktop::getImageSize
(
	dword& width,
	dword& height
) const
{
	using perceptuum2::imaging::ImageCieXyz;
	const ImageCieXyz& image = renderData_m.getImage();

	width  = image.getWidth();
	height = image.getHeight();
}




/// overrides --------------------------------------------------------------------------------------
void RenderDesktop::run()
{
	using perceptuum2::modelling::Scene;

	render( renderData_m.getScene(), renderData_m.getOptions(),
	        renderData_m.getStatus(), renderData_m.getImage(), renderData_m.getState() );
}


bool RenderDesktop::shouldGo() const
{
	return  getPermission() == GO;
}


bool RenderDesktop::shouldStop() const
{
	return  getPermission() == STOP;
}


bool RenderDesktop::shouldStopAtFrame() const
{
	return  getPermission() == PERMISSION_STOP_AT_FRAME;
}


void RenderDesktop::attendToLineDone
(
	const dword lineNumber
)   /// synchronized
{
	hxa7241::utility::ThreadSync ts( pThreadLock_m );

	if( lineNumber >= 0 )
	{
		/// get latest status line
		renderData_m.getStatus().getSummaryLine( statusLine_m );

		/// convert line
		convertImageLine( lineNumber );
	}
}


void RenderDesktop::attendToFrameDone
(
	const bool shouldRepaint
)   /// synchronized
{
	hxa7241::utility::ThreadSync ts( pThreadLock_m );

	/// auto adapt, if last auto-adapt acted on a blank image
	if( tonemapAdaptation_m <= 0.0001f )
	{
		tonemapAdaptation_m = renderData_m.getImage().getLogMeanLuminance();
	}

	if( shouldRepaint )
	{
		/// get latest status line
		renderData_m.getStatus().getSummaryLine( statusLine_m );

		convertImage();
	}

	/// persist if timely
	const dword currentTime = dword(std::time(0));
	if( (currentTime - lastPersistTime_m) > timeBetweenPersists_m )
	{
		if( lastPersistTime_m != 0 )
		{
			persist();
		}
		lastPersistTime_m = currentTime;
	}

	/// copy synched info
	copyInfo();


	//hxa7241::utility::Logger() << statusLine_m << "  - done";
}


void RenderDesktop::attendToRenderDone
(
	const bool shouldRepaint
)   /// synchronized
{
	hxa7241::utility::ThreadSync ts( pThreadLock_m );

	if( shouldRepaint )
	{
		/// auto adapt, and convert image
		tonemapAdaptation_m = renderData_m.getImage().getLogMeanLuminance();
		convertImage();
	}

	/// persist, and copy synched info
	persist();
	copyInfo();
}




/// implementation ---------------------------------------------------------------------------------
void RenderDesktop::copyInfo()
{
	renderData_m.getFilesPathname( dataFilePathname_m );

	renderData_m.getOptions().getSummary( optionsSummary_m );
	renderData_m.getScene().getSummary( sceneSummary_m );
	renderData_m.getStatus().getSummary( statusSummary_m );
	renderData_m.getState().getSummary( stateSummary_m );
	renderData_m.getImage().getSummary( imageSummary_m );

	renderData_m.getStatus().getSummaryLine( statusLine_m );
}


void RenderDesktop::convertImage()
{
	imageConverter_m.set( tonemapAdaptation_m );
	imageConverter_m.convertXYZtoRGB( renderData_m.getImage(), true, imageRgba_m );
}


void RenderDesktop::convertImageLine
(
	const dword lineNumber
)
{
	imageConverter_m.set( tonemapAdaptation_m );
	imageConverter_m.convertXYZtoRGB( renderData_m.getImage(), lineNumber, true, imageRgba_m );
}


void RenderDesktop::setImageConversion
(
	const RenderData& renderData,
	ImageConverter&   imageConverter
)
{
	DisplayMeasures dm;
	const RenderingOptions& ro = renderData.getOptions();
	ro.getDisplayCharacteristics( dm );

	using perceptuum2::imaging::CieXyz;
	using hxa7241::graphics::Vector3f;

	CieXyz   blackPoint;
	CieXyz   whitePoint;
	Vector3f chromaticitiesXyz[3];
	Vector3f gammaXyz;
	dm.getBlackPoint( blackPoint );
	dm.getWhitePoint( whitePoint );
	dm.getChromaticitiesXyz( chromaticitiesXyz );
	dm.getGammaXyz( gammaXyz );

	imageConverter.set( blackPoint, whitePoint, chromaticitiesXyz );
	imageConverter.set( gammaXyz );
}
