/*--------------------------------------------------------------------

   Perceptuum renderer, version 2.1
   Copyright (c) 2002-2003,  Harrison Ainsworth.

   www.hxa7241.org

--------------------------------------------------------------------*/

/*--------------------------------------------------------------------

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston,
   MA 02111-1307, USA

--------------------------------------------------------------------*/


#include "ColorTransformXyzToRgbFp.h"


using namespace perceptuum2::imaging;




/// statics
/// ITU-R BT.709 reference primaries (sRGB conformant)
const Vector3f ColorTransformXyzToRgbFp::sRGB_CHROMATICITIES[ 3 ] =
	{ Vector3f( 0.64f, 0.30f, 0.150f ),     /// x
	  Vector3f( 0.33f, 0.60f, 0.060f ),     /// y
	  Vector3f( 0.03f, 0.10f, 0.790f ) };   /// z
/// CIE standard illuminant D65 (sRGB conformant)
const Vector3f ColorTransformXyzToRgbFp::sRGB_WHITEPOINT = Vector3f( 0.951f, 1.00f, 1.089f );




/// standard object services -----------------------------------------------------------------------
ColorTransformXyzToRgbFp::ColorTransformXyzToRgbFp()
 :	XYZtoRGBmatrix_m(),
	RGBtoXYZmatrix_m()
{
	set( sRGB_CHROMATICITIES, Vector3f::ONE() );
}


ColorTransformXyzToRgbFp::ColorTransformXyzToRgbFp
(
	const Vector3f  chromaticities[3],
	const Vector3f& whitePoint
)
 :	XYZtoRGBmatrix_m(),
	RGBtoXYZmatrix_m()
{
	set( chromaticities, whitePoint );
}


ColorTransformXyzToRgbFp::~ColorTransformXyzToRgbFp()
{
}


ColorTransformXyzToRgbFp::ColorTransformXyzToRgbFp( const ColorTransformXyzToRgbFp& other )
 :	XYZtoRGBmatrix_m( other.XYZtoRGBmatrix_m ),
	RGBtoXYZmatrix_m( other.RGBtoXYZmatrix_m )
{
}


ColorTransformXyzToRgbFp& ColorTransformXyzToRgbFp::operator=( const ColorTransformXyzToRgbFp& other )
{
	if( &other != this )
	{
		XYZtoRGBmatrix_m = other.XYZtoRGBmatrix_m;
		RGBtoXYZmatrix_m = other.RGBtoXYZmatrix_m;
	}

	return *this;
}




/// commands ---------------------------------------------------------------------------------------
void ColorTransformXyzToRgbFp::setSRGB()
{
	set( sRGB_CHROMATICITIES, sRGB_WHITEPOINT );
}


void ColorTransformXyzToRgbFp::set
(
	const Vector3f  chromaticities[3],
	const Vector3f& whitePoint
)
{
	Matrix3f chrm( chromaticities[0],
	               chromaticities[1],
	               chromaticities[2], Vector3f::ZERO() );

	/// start matrix with chromaticities
	RGBtoXYZmatrix_m = chrm;

	/// extract white color
	Vector3f whiteColor;
	if( whitePoint != Vector3f::ZERO() )
	{
		(whiteColor = whitePoint) /= whitePoint.getY();
	}
	else
	{
		whiteColor = sRGB_WHITEPOINT;
	}

	/// inverted chromaticities * white color to calculate the unknown
	Vector3f C;
	chrm.invert();
	chrm.multiply( whiteColor, C );

	/// scaled chrms makes the conversion matrix
	RGBtoXYZmatrix_m.scale( C );

	/// inverse conversion is the same, but inverted
	XYZtoRGBmatrix_m = RGBtoXYZmatrix_m;
	XYZtoRGBmatrix_m.invert();
}




/// queries ----------------------------------------------------------------------------------------
void ColorTransformXyzToRgbFp::transform
(
	Vector3f& point
) const
{
	XYZtoRGBmatrix_m.multiply( point );
}


void ColorTransformXyzToRgbFp::transform
(
	const Array<Vector3f>& points
) const
{
	for( dword i = points.getLength();  i-- > 0; )
	{
		XYZtoRGBmatrix_m.multiply( points[i] );
	}
}


void ColorTransformXyzToRgbFp::invTransform
(
	Vector3f& point
) const
{
	RGBtoXYZmatrix_m.multiply( point );
}


void ColorTransformXyzToRgbFp::invTransform
(
	const Array<Vector3f>& points
) const
{
	for( dword i = points.getLength();  i-- > 0; )
	{
		RGBtoXYZmatrix_m.multiply( points[i] );
	}
}
