/*--------------------------------------------------------------------

   Perceptuum renderer, version 2.1
   Copyright (c) 2002-2003,  Harrison Ainsworth.

   www.hxa7241.org

--------------------------------------------------------------------*/

/*--------------------------------------------------------------------

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston,
   MA 02111-1307, USA

--------------------------------------------------------------------*/


#include "FpToInt.h"
#include "hxa7241math.h"

#include "GammaTransform.h"   /// own header is included last


using namespace perceptuum2::imaging;




/// statics
const Vector3f GammaTransform::sRGB_GAMMA( 2.2f, 2.2f, 2.2f );




/// standard object services -----------------------------------------------------------------------
GammaTransform::GammaTransform()
 :	gamma_m(),
	invGamma_m(),
	gammaTable_m(),
	invGammaTable_m()
{
	GammaTransform::set( Vector3f::ONE() );
}


GammaTransform::GammaTransform
(
	const Vector3f& gamma
)
 :	gamma_m(),
	invGamma_m(),
	gammaTable_m(),
	invGammaTable_m()
{
	GammaTransform::set( gamma );
}


GammaTransform::~GammaTransform()
{
}


GammaTransform::GammaTransform( const GammaTransform& other )
 :	gamma_m        ( other.gamma_m ),
	invGamma_m     ( other.invGamma_m ),
	gammaTable_m   ( other.gammaTable_m ),
	invGammaTable_m( other.invGammaTable_m )
{
}


GammaTransform& GammaTransform::operator=( const GammaTransform& other )
{
	if( &other != this )
	{
		gamma_m         = other.gamma_m;
		invGamma_m      = other.invGamma_m;

		gammaTable_m    = other.gammaTable_m;
		invGammaTable_m = other.invGammaTable_m;
	}

	return *this;
}




/// commands ---------------------------------------------------------------------------------------
void GammaTransform::setSRGB()
{
	setGamma( sRGB_GAMMA );
}


void GammaTransform::set
(
	const Vector3f& gamma
)
{
	setGamma( gamma );
}




/// queries ----------------------------------------------------------------------------------------
float GammaTransform::gammaEncode
(
	const float fp01
) const
{
	float invGammaMean = invGamma_m.average();
	if( gamma_m == sRGB_GAMMA )
	{
		invGammaMean = 0.0f;
	}

	return gammaEncode( fp01, invGammaMean );
}


float GammaTransform::gammaEncode
(
	      float fp01,
	const float invGamma
)
{
	/// make sure fp01 really is [0-1]
	if( fp01 < 0.0f )
	{
		fp01 = 0.0f;
	}
	else
	if( fp01 > 1.0f )
	{
		fp01 = 1.0f;
	}

	/// gamma encode
	float fp01g;
	if( invGamma <= 0.0f )
	{
		/// sRGB gamma encoding
		if( fp01 <= 0.00304f )
		{
			fp01g = 12.92f * fp01;
		}
		else
		{
			fp01g = (1.055f * float(::pow( fp01, 1.0f / 2.4f ))) - 0.055f;
		}
	}
	else
	{
		/// general/non-standard gamma encoding
		fp01g = float(::pow( fp01, invGamma ));
	}

	return fp01g;
}


void GammaTransform::gammaEncodeQuantize255
(
	const Vector3f& colorFp,
	dword           color24[3]
) const
{
	getFromTable( invGammaTable_m, colorFp, color24 );
}




float GammaTransform::gammaDecode
(
	const float fp01
) const
{
	float gammaMean = gamma_m.average();
	if( gamma_m == sRGB_GAMMA )
	{
		gammaMean = 0.0f;
	}

	return gammaDecode( fp01, gammaMean );
}


float GammaTransform::gammaDecode
(
	      float fp01,
	const float gamma
)
{
	/// make sure fp01 really is [0-1]
	if( fp01 < 0.0f )
	{
		fp01 = 0.0f;
	}
	else
	if( fp01 > 1.0f )
	{
		fp01 = 1.0f;
	}

	/// gamma decode
	float fp01g;
	if( gamma <= 0.0f )
	{
		/// sRGB gamma decoding
		if( fp01 <= 0.03928f )
		{
			fp01g = fp01 * (1.0f / 12.92f);
		}
		else
		{
			fp01g = float( ::pow( (fp01 + 0.055f) * (1.0f / 1.055f), 2.4f ) );
		}
	}
	else
	{
		/// general/non-standard gamma encoding
		fp01g = float(::pow( fp01, gamma ));
	}

	return fp01g;
}


void GammaTransform::gammaDecode
(
	const Vector3f& in,
	Vector3f&       out
) const
{
	dword out24[3];
	getFromTable( gammaTable_m, in, out24 );

	/// fractionalize
	using hxa7241::utility::ubyteToFp;
	out.setXYZ( ubyteToFp( out24[0] ),
	            ubyteToFp( out24[1] ),
	            ubyteToFp( out24[2] ) );
}




void GammaTransform::getGamma
(
	Vector3f& gamma
) const
{
	gamma = gamma_m;
}


void GammaTransform::getInvGamma
(
	Vector3f& invGamma
) const
{
	invGamma = invGamma_m;
}




/// implementation ---------------------------------------------------------------------------------
void GammaTransform::setGamma
(
	const Vector3f& gamma
)
{
	gamma_m = gamma;
	gamma_m.clampToMinOf( Vector3f( 0.1f, 0.1f, 0.1f ) );

	invGamma_m = 1.0f / gamma_m;


	/// use special gamma values to flag use of special sRGB encoding
	Vector3f invGammaF( invGamma_m );
	Vector3f gammaF( gamma_m );
	if( gamma_m == sRGB_GAMMA )
	{
		invGammaF = Vector3f::ZERO();
		gammaF    = Vector3f::ZERO();
	}

	/// make inv gamma lookup table
	{
		invGammaTable_m.setLength( TABLE_LENGTH * 3 );
		gammaTable_m.setLength( TABLE_LENGTH * 3 );

		float _0to1Fraction;
		byte* tablePtr  = invGammaTable_m.getMemory();
		byte* table2Ptr = gammaTable_m.getMemory();

		for( dword i = TABLE_LENGTH;  i-- > 0; )
		{
			_0to1Fraction = (float(i) + 0.5f) / float(TABLE_LENGTH);

			using hxa7241::utility::fpToIntTrunc;
			tablePtr[ i                       ] = byte(hxa7241::utility::fpToInt(255.0f * gammaEncode( _0to1Fraction, invGammaF.getX() )));
			tablePtr[ i +  TABLE_LENGTH       ] = byte(hxa7241::utility::fpToInt(255.0f * gammaEncode( _0to1Fraction, invGammaF.getY() )));
			tablePtr[ i + (TABLE_LENGTH << 1) ] = byte(hxa7241::utility::fpToInt(255.0f * gammaEncode( _0to1Fraction, invGammaF.getZ() )));
//			tablePtr[ i                       ] = byte(fpToIntTrunc( float(::pow( _0to1Fraction, invGamma_m.getX() )) * 256.0f ));
//			tablePtr[ i +  TABLE_LENGTH       ] = byte(fpToIntTrunc( float(::pow( _0to1Fraction, invGamma_m.getY() )) * 256.0f ));
//			tablePtr[ i + (TABLE_LENGTH << 1) ] = byte(fpToIntTrunc( float(::pow( _0to1Fraction, invGamma_m.getZ() )) * 256.0f ));

			table2Ptr[ i                       ] = byte(hxa7241::utility::fpToInt(255.0f * gammaDecode( _0to1Fraction, gammaF.getX() )));
			table2Ptr[ i +  TABLE_LENGTH       ] = byte(hxa7241::utility::fpToInt(255.0f * gammaDecode( _0to1Fraction, gammaF.getY() )));
			table2Ptr[ i + (TABLE_LENGTH << 1) ] = byte(hxa7241::utility::fpToInt(255.0f * gammaDecode( _0to1Fraction, gammaF.getZ() )));
		}
	}
}


void GammaTransform::getFromTable
(
	const Array<hxa7241::primitives::byte>& table,
	const Vector3f& colorFp,
	dword           color24[3]
) const
{
	Vector3f color01( colorFp );
	color01.clamp01();

	const dword length   = TABLE_LENGTH;
	const float lengthFp = float( length );
	byte*       tablePtr = table.getMemory();

	using hxa7241::utility::fpToIntTrunc;
	color24[0] = dword(ubyte( tablePtr[ fpToIntTrunc( color01.getX() * lengthFp )                   ] ));
	color24[1] = dword(ubyte( tablePtr[ fpToIntTrunc( color01.getY() * lengthFp ) + length          ] ));
	color24[2] = dword(ubyte( tablePtr[ fpToIntTrunc( color01.getZ() * lengthFp ) + length + length ] ));
}
