/*--------------------------------------------------------------------

   Perceptuum renderer, version 2.1
   Copyright (c) 2002-2003,  Harrison Ainsworth.

   www.hxa7241.org

--------------------------------------------------------------------*/

/*--------------------------------------------------------------------

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston,
   MA 02111-1307, USA

--------------------------------------------------------------------*/


#include "Serialization.h"
#include "hxa7241math.h"
#include <sstream>

#include "ImageCieXyz.h"


using namespace perceptuum2::imaging;




//// standard object services ----------------------------------------------------------------------
ImageCieXyz::ImageCieXyz()
 :	Sheet<CieXyz>(),
	alpha_m      ()
{
}


ImageCieXyz::ImageCieXyz( const dword width, const dword height )
 :	Sheet<CieXyz>( width, height ),
	alpha_m      ( width, height )
{
	alpha_m.zeroMemory();
}


ImageCieXyz::~ImageCieXyz()
{
}


ImageCieXyz::ImageCieXyz( const ImageCieXyz& other )
 :	Sheet<CieXyz>( other ),
	alpha_m      ( other.alpha_m )
{
}


ImageCieXyz& ImageCieXyz::operator=( const ImageCieXyz& other )
{
	Sheet<CieXyz>::operator=( other );
	alpha_m = other.alpha_m;

	return *this;
}




/// commands ---------------------------------------------------------------------------------------
void ImageCieXyz::setSize
(
	const dword width,
	const dword height
)
{
	Sheet<CieXyz>::setSize( width, height );
	alpha_m.setSize( width, height );

	alpha_m.zeroMemory();
}


void ImageCieXyz::zeroMemory() const
{
	Sheet<CieXyz>::zeroMemory();
	alpha_m.zeroMemory();
}


void ImageCieXyz::setAlphaOpaque() const
{
	byte* pAlphas = getAlphaMemory();
	for( const byte* pEnd = pAlphas + getLength();  pAlphas < pEnd;  ++pAlphas )
	{
		*pAlphas = byte(0xFF);
	}
}




/// queries ----------------------------------------------------------------------------------------
hxa7241::utility::byte* ImageCieXyz::getAlphaMemory() const
{
	return alpha_m.getMemory();
}


hxa7241::utility::byte* ImageCieXyz::getAlphaRow( const dword y ) const
{
	return alpha_m.getRow( y );
}


hxa7241::utility::byte& ImageCieXyz::getAlphaElement( const dword x, const dword y ) const
{
	return alpha_m.getElement( x, y );
}


void ImageCieXyz::getSummary( string& s ) const
{
	std::ostringstream oss;

	oss << '\n' << "image CIE XYZ" << '\n' << '\n' << '\n';
	oss << "width:   " << getWidth()  << '\n';
	oss << "height:  " << getHeight() << '\n';

	const dword bytes = (getLength() * (sizeof(*getMemory()) + sizeof(*(alpha_m.getMemory())))) + (2 * sizeof(dword));
	oss << "size:  " << (float(bytes) / 1048576.0f) << " MB" << '\n';

	{
		float mean;
		float logMean;
		getLuminanceMeans( mean, logMean );
		oss << '\n';
		oss << "mean luminance:  " << mean << " cd/m^2" << '\n';
		oss << "log mean luminance:  " << logMean << " cd/m^2" << '\n';

		float min;
		float max;
		getLuminanceRange( min, max );
		oss << '\n';
		oss << "min luminance:  " << min << " cd/m^2" << '\n';
		oss << "max luminance:  " << max << " cd/m^2" << '\n';
	}

	oss << '\n';

	s = oss.str();
}


void ImageCieXyz::getLuminanceMeans
(
	float& mean,
	float& logMean
) const
{
	float sumOfLuminances    = 0.0f;
	float sumOfLogLuminances = 0.0f;
	dword noOfPixelsCounted  = 0;
	{
		const CieXyz* pPixels = getMemory();
		const dword   length  = getLength();

		for( int i = length;  i-- > 0;  ++pPixels )
		{
			const float Y = pPixels->getY();
			if( float(::fabs( Y - 0.0f )) > FLOAT_EPSILON )//( Y != 0.0f )//
			{
				sumOfLuminances    += Y;
				sumOfLogLuminances += float(::log10( Y ));

				++noOfPixelsCounted;
			}
		}
	}

	float meanLuminance    = 0.0f;
	float logMeanLuminance = 1.0e-4f;
	if( noOfPixelsCounted > 0 )
	{
		meanLuminance    = sumOfLuminances / float(noOfPixelsCounted);
		logMeanLuminance = float(::pow( 10.0f, sumOfLogLuminances / float(noOfPixelsCounted) ));
	}

	mean    = meanLuminance;
	logMean = logMeanLuminance;
}


void ImageCieXyz::getLuminanceRange
(
	float& min,
	float& max
) const
{
	float minLuminance = FLOAT_MAX;
	float maxLuminance = FLOAT_MIN_NEG;
	{
		const CieXyz* pPixels = getMemory();
		const dword   length  = getLength();

		if( length > 0 )
		{
			for( int i = length;  i-- > 0;  ++pPixels )
			{
				const float Y = pPixels->getY();
				if( float(::fabs( Y - 0.0f )) > FLOAT_EPSILON )//( Y != 0.0f )//
				{
					minLuminance = minLuminance <= Y ? minLuminance : Y;
					maxLuminance = maxLuminance >= Y ? maxLuminance : Y;
				}
			}
		}
		else
		{
			minLuminance = 0.0f;
			maxLuminance = 0.0f;
		}
	}

	min = minLuminance;
	max = maxLuminance;
}


float ImageCieXyz::getLogMeanLuminance() const
{
	float mean;
	float logMean;
	getLuminanceMeans( mean, logMean );

	return logMean;


//	float sumOfLogLuminances = 0.0f;
//	//float sumOfLuminances    = 0.0f;
//	dword noOfPixelsCounted  = 0;
//	{
//		const CieXyz* pPixels = getMemory();
//		const dword   length  = getLength();
//
//		for( int i = length;  i-- > 0;  ++pPixels )
//		{
//			const float Y = pPixels->getY();
//			if( float(::fabs( Y - 0.0f )) > FLOAT_EPSILON )//( Y != 0.0f )//
//			{
//				sumOfLogLuminances += float(::log10( Y ));
//				//sumOfLuminances += Y;
//				++noOfPixelsCounted;
//			}
//		}
//	}
//
//	float logAveLuminance = 1.0e-4f;
//	//float aveLuminance = 0.0f;
//	if( noOfPixelsCounted > 0 )
//	{
//		logAveLuminance = float(::pow( 10.0f, sumOfLogLuminances / float(noOfPixelsCounted) ));
//		//aveLuminance = sumOfLuminances / float(noOfPixelsCounted);
//	}
//
//	return logAveLuminance;
//	//return aveLuminance;
}




/// serialization ----------------------------------------------------------------------------------
SerializerOut& perceptuum2::imaging::operator<<( SerializerOut& out, const ImageCieXyz& obj )
{
	/// write size
	out.write( "width" , obj.getWidth()  );
	out.write( "height", obj.getHeight() );

	/// write pixel array
	{
		const CieXyz* pPixels = obj.getMemory();
		const CieXyz* pEnd    = pPixels + obj.getLength();
		for( ;  pPixels < pEnd;  ++pPixels )
		{
			out.write( "pixel", *pPixels );
		}
	}

	/// write alpha array
	{
		const byte* pAlphas = obj.alpha_m.getMemory();
		const byte* pEnd    = pAlphas + obj.alpha_m.getLength();
		for( ;  pAlphas < pEnd;  ++pAlphas )
		{
			out.write( "alpha", *pAlphas );
		}
	}

	return out;
}


SerializerIn& perceptuum2::imaging::operator>>( SerializerIn& in, ImageCieXyz& obj )
{
	/// read size
	dword width;
	dword height;
	in.read( "width" , width  );
	in.read( "height", height );

	/// read pixel array
	{
		/// allocate storage
		obj.setSize( width, height );

			  CieXyz* pPixels = obj.getMemory();
		const CieXyz* pEnd    = pPixels + obj.getLength();
		for( ;  pPixels < pEnd;  ++pPixels )
		{
			in.read( "pixel", *pPixels );
		}
	}

	/// read alpha array
	{
		/// allocate storage
		obj.alpha_m.setSize( width, height );

			  byte* pAlphas = obj.alpha_m.getMemory();
		const byte* pEnd    = pAlphas + obj.alpha_m.getLength();
		for( ;  pAlphas < pEnd;  ++pAlphas )
		{
			in.read( "alpha", *pAlphas );
		}
	}

	return in;
}


const char* ImageCieXyz::CLASS_NAME()
{
	static const char CLASS_NAME[] = "perceptuum2::imaging::ImageCieXyz";
	return CLASS_NAME;
}


const char* ImageCieXyz::CLASS_DESCRIPTION()
{
	static const char CLASS_DESCRIPTION[] = "IMAGE-CIEXYZ-96BIT-ALPHA-8BIT"
	                                        "\ndata starts at the first dword alignment after this description text, "
	                                        "\nfirst dword is width in pixels, next dword is height, "
	                                        "\nfollowing dwords are IEEE single floats, each pixel being an adjacent triple, "
	                                        "first X, then Y, then Z. "
	                                        "\nfirst pixel is top left in the image, first (width) pixels is top line in the image. "
	                                        "immediately following all the xyz pixels are alpha values, each 1 byte, non-premultiplied. "
	                                        "\nmulti byte values are in network byte order: higher bytes first."
	                                        "\n\n";
//	                                        "IMAGE-CIEXYZ-REALPIXELS-ALPHA-8BIT"
//	                                        "\ndata starts at the first dword alignment after this description text, "
//	                                        "\nfirst dword is width in pixels, next dword is height, "
//	                                        "\nfollowing dwords are pixels of CIE XYZ color, in ward real-pixel format, "
//	                                        "r in high byte, then g, b, then exponent. "
//	                                        "\nfirst pixel is top left in the image, first (width) pixels is top line in the image. "
//	                                        "immediately following all the xyz pixels are alpha values, each 1 byte, non-premultiplied. "
//	                                        "\nmulti byte values are in network byte order: higher bytes first."
//	                                        "\n\n";
	return CLASS_DESCRIPTION;
}
