/*--------------------------------------------------------------------

   Perceptuum renderer, version 2.1
   Copyright (c) 2002-2003,  Harrison Ainsworth.

   www.hxa7241.org

--------------------------------------------------------------------*/

/*--------------------------------------------------------------------

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston,
   MA 02111-1307, USA

--------------------------------------------------------------------*/


#include "ImageCieXyz.h"
#include "ImageRgba.h"
#include "CieXyz.h"
#include "FpToInt.h"
#include "Exception.h"

#include "ImageConverter.h"


using namespace perceptuum2::imaging;




/// statics
const char ImageConverter::CLASS_NAME[]         = "perceptuum2::imaging::ImageConverter";
const char ImageConverter::EXCEPTION_MESSAGE1[] = "images not the same size";




/// standard object services -----------------------------------------------------------------------
ImageConverter::ImageConverter()
 :	toneMapper_m(),
	colorConverter_m(),
	gammaTransformer_m()
{
}


ImageConverter::ImageConverter
(
	const CieXyz&   whiteColor,
	const Vector3f  chromaticities[3],
	const Vector3f& gamma
)
 :	toneMapper_m      (),
	colorConverter_m  ( chromaticities, whiteColor ),
	gammaTransformer_m( gamma )
{
}


ImageConverter::ImageConverter
(
	const CieXyz&   blackPoint,
	const CieXyz&   whitePoint,
	const Vector3f  chromaticities[3],
	const Vector3f& gamma
)
 :	toneMapper_m      ( blackPoint.getY(), whitePoint.getY() ),
	colorConverter_m  ( chromaticities, whitePoint ),
	gammaTransformer_m( gamma )
{
}


ImageConverter::~ImageConverter()
{
}




/// commands ---------------------------------------------------------------------------------------
void ImageConverter::setSRGB()
{
	colorConverter_m.setSRGB();
	gammaTransformer_m.setSRGB();
}


void ImageConverter::set
(
	const CieXyz&  whiteColor,
	const Vector3f chromaticities[3]
)
{
	colorConverter_m.set( chromaticities, whiteColor );
}


void ImageConverter::set
(
	const CieXyz&  blackPoint,
	const CieXyz&  whitePoint,
	const Vector3f chromaticities[3]
)
{
	toneMapper_m.set( blackPoint.getY(), whitePoint.getY() );

	colorConverter_m.set( chromaticities, whitePoint );
}


void ImageConverter::set
(
	const Vector3f& gamma
)
{
	gammaTransformer_m.set( gamma );
}


void ImageConverter::set
(
	const float adaptLuminance
)
{
	toneMapper_m.set( adaptLuminance );
}




/// queries ----------------------------------------------------------------------------------------
void ImageConverter::convertXYZtoRGB
(
	const ImageCieXyz& xyz,
	const bool         reAssociateAlpha,
	ImageRgba&         rgba
) const
{
	checkImageDimensionMatch( xyz, rgba );

	for( dword row = 0;  row < rgba.getHeight();  ++row )
	{
		convertXYZtoRGB( xyz, row, reAssociateAlpha, rgba );
	}
}


void ImageConverter::convertXYZtoRGB
(
	const ImageCieXyz& xyz,
	const dword        row,
	const bool         reAssociateAlpha,
	ImageRgba&         rgba
) const
{
	checkImageDimensionMatch( xyz, rgba );

	/// setup simple cache
	CieXyz lastXyz( 0.0f, 0.0f, 0.0f );
	byte   lastAlpha  = 0xFF;
	const  dword a[4] = { 0, 0, 0, 0xFF };
	dword  lastRgba   = rgba.toDword( a );

	const CieXyz* pXyz     = xyz.getRow( row );
	const byte*   pAlpha   = xyz.getAlphaRow( row );
	      dword*  pRgba    = rgba.getRow( row );
	      dword*  pRgbaEnd = pRgba + rgba.getWidth();

	/// step through row of pixels
	for( ;  pRgba < pRgbaEnd;  ++pXyz, ++pAlpha, ++pRgba )
	{
		/// check for cache hit
		if( bool((*pXyz == lastXyz) & (*pAlpha == lastAlpha)) )
		{
			*pRgba = lastRgba;
		}
		else
		{
			/// set new cache values
			lastXyz   = *pXyz;
			lastAlpha = *pAlpha;

			/// unassociate alpha from xyz
			CieXyz c( *pXyz );
			{
				static const float ONE_OVER_255 = 1.0f / 255.0f;
				if( *pAlpha != 0 )
				{
					c /= (float(ubyte(*pAlpha)) * ONE_OVER_255);
				}
				else
				{
					c = CieXyz::ZERO();//c.setXYZ( 0.0f, 0.0f, 0.0f );
				}
			}

			/// convert color, copy alpha
			dword r[4];
			convertXYZtoRGB( c, r );
			r[3] = *pAlpha;

			/// set rgba, and re-associate alpha
			*pRgba = rgba.toDword( r );
			if( reAssociateAlpha )
			{
				*pRgba = rgba.associateAlphaP( *pRgba );
			}

			/// set new cache value
			lastRgba = *pRgba;
		}
	}
}


void ImageConverter::convertRGBtoXYZ
(
	const ImageRgba& rgba,
	const bool       isRgbAlphaAssociated,
	const bool       isTonemapable,
	ImageCieXyz&     xyz
) const
{
	checkImageDimensionMatch( xyz, rgba );

	/// setup simple cache
	CieXyz lastXyz( 0.0f, 0.0f, 0.0f );
	byte   lastAlpha  = 0xFF;
	const  dword a[4] = { 0, 0, 0, 0xFF };
	dword  lastRgba   = rgba.toDword( a );

	const dword*  pRgba    = rgba.getMemory();
	const dword*  pRgbaEnd = pRgba + rgba.getLength();
	      CieXyz* pXyz     = xyz.getMemory();
	      byte*   pAlpha   = xyz.getAlphaMemory();

	for( ;  pRgba < pRgbaEnd;  ++pXyz, ++pAlpha, ++pRgba )
	{
		/// check for cache hit
		if( *pRgba == lastRgba )
		{
			*pXyz   = lastXyz;
			*pAlpha = lastAlpha;
		}
		else
		{
			dword r[4];
			rgba.fromDword( *pRgba, r );

			convertRGBtoXYZ( r, isRgbAlphaAssociated, isTonemapable, *pXyz );
			*pAlpha = byte(r[3]);

			/// set new cache values
			lastRgba  = *pRgba;
			lastXyz   = *pXyz;
			lastAlpha = *pAlpha;
		}
	}
}


void ImageConverter::convertRGBtoXYZ
(
	const Vector3f&                 rgb,
	const hxa7241::primitives::byte alpha,
	const bool                      isRgbAlphaAssociated,
	const bool                      isTonemapable,
	CieXyz&                         xyz
) const
{
	if( !rgb.isZero() )
	{
		Vector3f color( rgb );

		if( isRgbAlphaAssociated )
		{
			/// unassociate rgb alpha
			static const float ONE_OVER_255 = 1.0f / 255.0f;
			if( alpha != 0 )
			{
				color /= (float(ubyte(alpha)) * ONE_OVER_255);
			}
			else
			{
				color = Vector3f::ZERO();//color.setXYZ( 0.0f, 0.0f, 0.0f );
			}
		}

		/// gamma decode
		gammaTransformer_m.gammaDecode( color, color );

		/// unmap color space
		colorConverter_m.invTransform( color );

		if( isTonemapable )
		{
			/// clamp
			color.clampToMinOf( Vector3f::ZERO() );

			/// unmap tone
			toneMapper_m.unMap( color );
			///// unmap tone (achromatically)
			//if( color.getY() != 0.0f )
			//{
			//	color *= toneMapper_m.unMap( color.getY() ) / color.getY();
			//}
		}
		else
		{
			/// clamp
			color.clamp01();
		}

		/// associate alpha
		static const float ONE_OVER_255 = 1.0f / 255.0f;
		color *= (float(ubyte(alpha)) * ONE_OVER_255);

		xyz = color;
	}
	else
	{
		xyz = CieXyz::ZERO();
	}
}


float ImageConverter::getAdaptLuminance() const
{
	return toneMapper_m.getAdaptLuminance();
}




/// implementation ---------------------------------------------------------------------------------
void ImageConverter::checkImageDimensionMatch
(
	const ImageCieXyz& xyz,
	const ImageRgba&   rgba
) const
{
	static const char METHOD_NAME[] = "convertXYZtoRGB";

	if( (xyz.getWidth() != rgba.getWidth()) || (xyz.getHeight() != rgba.getHeight()) )
	{
		throw hxa7241::utility::Exception( EXCEPTION_MESSAGE1, CLASS_NAME, METHOD_NAME );
	}
}


void ImageConverter::convertXYZtoRGB
(
	const CieXyz& xyz,
	      dword   rgb[3]
) const
{
	if( xyz.getY() != 0.0f )
	{
		Vector3f color( xyz );

		/// map tone
		toneMapper_m.map( color );
		///// map tone (achromatically)
		//color *= toneMapper_m.map( color.getY() ) / color.getY();

		/// map color space
		colorConverter_m.transform( color );

		/// gamma correct and quantize
		gammaTransformer_m.gammaEncodeQuantize255( color, rgb );
	}
	else
	{
		rgb[0] = 0;
		rgb[1] = 0;
		rgb[2] = 0;
	}
}


void ImageConverter::convertRGBtoXYZ
(
	const dword   rgba[4],
	const bool    isRgbAlphaAssociated,
	const bool    isTonemapable,
	      CieXyz& xyz
) const
{
	const Vector3f rgbFp( hxa7241::utility::ubyteToFp( rgba[0] ),
	                      hxa7241::utility::ubyteToFp( rgba[1] ),
	                      hxa7241::utility::ubyteToFp( rgba[2] ) );

	convertRGBtoXYZ( rgbFp, byte(rgba[3]), isRgbAlphaAssociated, isTonemapable, xyz );
}
