/*--------------------------------------------------------------------

   Perceptuum renderer, version 2.1
   Copyright (c) 2002-2003,  Harrison Ainsworth.

   www.hxa7241.org

--------------------------------------------------------------------*/

/*--------------------------------------------------------------------

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston,
   MA 02111-1307, USA

--------------------------------------------------------------------*/


#include "ImageCieXyz.h"
#include "XyzE.h"
#include "Sheet.h"
#include "Logger.h"
#include "TimePoint.h"
#include "Exception.h"
#include "png.h"
#include <iostream>
#include <setjmp>

#include "RealPngWriter.h"


using namespace perceptuum2::imaging;




/// statics
const char RealPngWriter::FORMAT_NAME[]         = "rpng";
const char RealPngWriter::IMAGE_TEXT_COMMENT[]  = "real-png format: the 'xyZe' chunk indicates that the rgb channels contain mantissas of CIE XYZ values, and the alpha channel contains the shared exponent, as described in 'real pixels' - ward, 'graphics gems 2' (AP 91)";
const char RealPngWriter::IMAGE_TEXT_SOFTWARE[] = "Perceptuum2 renderer {harrison artifex, london, 2002 - www.hxa7241.org}";
const char RealPngWriter::XYZE_CHUNK_NAME[]     = "xyZe";

const char RealPngWriter::EXCEPTION_MESSAGE_1[] = "failed to initialize libpng";
const char RealPngWriter::EXCEPTION_MESSAGE_2[] = "stream write failure";
const char RealPngWriter::EXCEPTION_MESSAGE_3[] = "stream flush failure";


/// libpng callbacks
void __cdecl writePngData( png_structp pPngObj, png_bytep pData, png_size_t length );
void __cdecl flushPngData( png_structp pPngObj );
//void __cdecl reOrderBytes( png_structp pPngObj, png_row_infop pRowInfo, png_bytep pData );
void __cdecl attendToPngError( png_structp pPngObj, png_const_charp pErrorMsg );
void __cdecl attendToPngWarning( png_structp pPngObj, png_const_charp pWarningMsg );




/// standard object services -----------------------------------------------------------------------
RealPngWriter::RealPngWriter()
{
}




/// queries ----------------------------------------------------------------------------------------
void RealPngWriter::getName( string& s ) const
{
	s = FORMAT_NAME;
}


void RealPngWriter::write
(
	const ImageCieXyz&    imageXyz,
	const ImageConverter& ,//imageConverter,
	ostream&              out
) const
{
	static const char METHOD_NAME[] = "write";


	/// for info see: www.libpng.org/pub/png

	png_structp pPngObj  = 0;
	png_infop   pPngInfo = 0;
	string      pngErrorMsg;

	try
	{
		/// create basic png objects
		pPngObj = ::png_create_write_struct( PNG_LIBPNG_VER_STRING, &pngErrorMsg, attendToPngError, attendToPngWarning );
		if( pPngObj == 0 )
		{
			throw hxa7241::utility::Exception( EXCEPTION_MESSAGE_1, CLASS_NAME(), METHOD_NAME );
		}

		pPngInfo = ::png_create_info_struct( pPngObj );
		if( pPngInfo == 0 )
		{
			throw hxa7241::utility::Exception( EXCEPTION_MESSAGE_1, CLASS_NAME(), METHOD_NAME );
		}

		/// do non c++ exception handling
		if( std::setjmp(pPngObj->jmpbuf) )
		{
			throw hxa7241::utility::Exception( pngErrorMsg.c_str(), CLASS_NAME(), METHOD_NAME );
		}

		/// set some options and callbacks
		{
			::png_set_write_fn( pPngObj, &out, writePngData, flushPngData );
			//::png_set_write_status_fn( pPngObj, attendToPngRowWritten );   /// void attendToPngRowWritten( png_ptr, png_uint_32 row, int pass );
			::png_set_filter( pPngObj, 0, PNG_ALL_FILTERS );
			::png_set_compression_level( pPngObj, 9 );//Z_BEST_COMPRESSION );
		}

		/// set chunks
		{
			::png_set_IHDR( pPngObj, pPngInfo,
							imageXyz.getWidth(), imageXyz.getHeight(), 8, PNG_COLOR_TYPE_RGB_ALPHA,
							PNG_INTERLACE_NONE, PNG_COMPRESSION_TYPE_DEFAULT, PNG_FILTER_TYPE_DEFAULT );

			png_unknown_chunk xyZeChunk = { { XYZE_CHUNK_NAME[0], XYZE_CHUNK_NAME[1], XYZE_CHUNK_NAME[2], XYZE_CHUNK_NAME[3], XYZE_CHUNK_NAME[4] }, 0, 0, 0 };
			::png_set_unknown_chunks( pPngObj, pPngInfo, &xyZeChunk, 1 );
			::png_set_unknown_chunk_location( pPngObj, pPngInfo, 0, 0x02 );   /// 0x02 == PNG_HAVE_PLTE

			const std::string timeNow( hxa7241::utility::TimePoint().getStringRFC1123() );
			png_text texts[] = { { PNG_TEXT_COMPRESSION_NONE, const_cast<char*>("Comment"),       const_cast<char*>(IMAGE_TEXT_COMMENT)  },
			                     { PNG_TEXT_COMPRESSION_NONE, const_cast<char*>("Software"),      const_cast<char*>(IMAGE_TEXT_SOFTWARE) },
			                     { PNG_TEXT_COMPRESSION_NONE, const_cast<char*>("Creation Time"), const_cast<char*>(timeNow.c_str())     } };
			::png_set_text( pPngObj, pPngInfo, texts, 3 );
		}


		::png_write_info( pPngObj, pPngInfo );


		/// convert to xyze real pixels
		hxa7241::utility::Sheet<dword> imageEzyx( imageXyz.getWidth(), imageXyz.getHeight() );
		{
			const CieXyz* pXyz = imageXyz.getMemory();
			dword* pEzyx       = imageEzyx.getMemory();
			dword* pEzyxEnd    = pEzyx + imageEzyx.getLength();
			for( ;  pEzyx < pEzyxEnd; )
			{
				const dword xyze = (pXyz++)->toXyzE().get();
				*(pEzyx++) = ((xyze & 0x000000FF) << 24) |
				             ((xyze & 0x0000FF00) <<  8) |
				             ((xyze & 0x00FF0000) >>  8) |
				             ((xyze & 0xFF000000) >> 24);
			}
		}


		/// set pixels
		std::vector<const dword*> rowPtrs( imageEzyx.getHeight() );
		for( dword i = imageEzyx.getHeight();  i-- > 0; )
		{
			rowPtrs[i] = imageEzyx.getRow( i );
		}
		::png_write_image( pPngObj, reinterpret_cast<png_bytepp>(&(rowPtrs[0])) );


		/// finish writing
		::png_write_end( pPngObj, 0 );


		/// delete basic png object
		::png_destroy_write_struct( &pPngObj, &pPngInfo );
	}
	catch( ... )
	{
		/// clear up
		if( pPngObj != 0 )
		{
			::png_destroy_write_struct( &pPngObj, &pPngInfo );
		}
		throw;
	}
}


const char* RealPngWriter::CLASS_NAME()
{
	static const char CLASS_NAME[] = "RealPngWriter";
	return CLASS_NAME;
}




/// libpng callbacks -------------------------------------------------------------------------------
void __cdecl writePngData
(
	png_structp pPngObj,
	png_bytep   pData,
	png_size_t  length
)
{
	ostream* pOstream = reinterpret_cast<ostream*>( ::png_get_io_ptr( pPngObj ) );

	pOstream->write( reinterpret_cast<const char*>(pData), length );
	if( pOstream->fail() )
	{
		throw hxa7241::utility::Exception( RealPngWriter::EXCEPTION_MESSAGE_2, RealPngWriter::CLASS_NAME(), "writePngData" );
	}
}


void __cdecl flushPngData
(
	png_structp pPngObj
)
{
	ostream* pOstream = reinterpret_cast<ostream*>( ::png_get_io_ptr( pPngObj ) );

	pOstream->flush();
	if( pOstream->fail() )
	{
		throw hxa7241::utility::Exception( RealPngWriter::EXCEPTION_MESSAGE_3, RealPngWriter::CLASS_NAME(), "flushPngData" );
	}
}


//void __cdecl reOrderBytes
//(
//	png_structp   pPngObj,
//	png_row_infop pRowInfo,
//	png_bytep     pData
//)
//{
//}


void __cdecl attendToPngError
(
	png_structp     pPngObj,
	png_const_charp pErrorMsg
)
{
	string* pString = reinterpret_cast<string*>( ::png_get_error_ptr(pPngObj) );
	*pString = pErrorMsg;

	std::longjmp( pPngObj->jmpbuf, 1 );
}


void __cdecl attendToPngWarning
(
	png_structp,
	png_const_charp pWarningMsg
)
{
	using hxa7241::utility::Logger;
	Logger() << Logger::EXC1 << RealPngWriter::CLASS_NAME() << " : " << pWarningMsg;
}
