/*--------------------------------------------------------------------

   Perceptuum renderer, version 2.1
   Copyright (c) 2002-2003,  Harrison Ainsworth.

   www.hxa7241.org

--------------------------------------------------------------------*/

/*--------------------------------------------------------------------

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston,
   MA 02111-1307, USA

--------------------------------------------------------------------*/


#ifndef Brdf_h
#define Brdf_h


#include "Primitives.h"
#include "UnitVector3f.h"
#include "Vector2.h"
#include "Float01.h"
#include "CieXyz.h"
#include "Sheet.h"
#include <vector>

#include "hxa7241_utility.h"




#include "perceptuum2_interaction.h"
namespace perceptuum2 { namespace interaction {
	using namespace hxa7241::primitives;
	using hxa7241::graphics::Vector3f;
	using hxa7241::graphics::UnitVector3f;
	using hxa7241::graphics::Vector2;
	using hxa7241::graphics::CieXyz;
	using hxa7241::utility::Float01;
	using hxa7241::utility::Sheet;
	using std::vector;
	typedef Vector2<Float01> Vector2f01;

/**
 * abstract brdf class.<br/><br/>
 *
 * to make concrete, implement the getReflectivity_, getPartitions_, getCos_,
 * getScaling_, getSample_ and clone methods.<br/><br/>
 *
 * in the public interface, directions are physical, that is: in points towards
 * the surface, out away from the surface.<br/>
 * in the protected interface, directions are 'geometrical', that is: in and out
 * both point away from the surface.<br/><br/>
 *
 * includes comprehensive tests, and simple 2d viewer.
 */
class Brdf
{
/// standard object services -----------------------------------------------------------------------
protected:
	         Brdf();
	         Brdf( const UnitVector3f& normal,
	               const UnitVector3f& tangent );
public:

	virtual ~Brdf();
protected:
	         Brdf( const Brdf& );
	Brdf& operator=( const Brdf& );


/// commands ---------------------------------------------------------------------------------------
public:
	virtual void  setBrdf( const UnitVector3f& normal,
	                       const UnitVector3f& tangent );


/// queries ----------------------------------------------------------------------------------------
public:
	/// brdf
	static const dword PARTITIONS_MAX = 8;

	virtual Float01 getReflectivity()                                                      const;
	virtual void    getPartitions( float   proportions[PARTITIONS_MAX],
	                               Float01 peakness[PARTITIONS_MAX],
	                               dword&  length )                                        const;
	virtual Float01 getCos( dword               part,
	                        const UnitVector3f& outDirection )                             const;
	virtual void    getScaling( dword               part,
	                            const UnitVector3f& inDirection,
	                            const UnitVector3f& outDirection,
	                            Light&              scaling )                              const;
	/** @return is transmitted */
	virtual bool    getSample( dword               part,
	                           Float01             roulette,
	                           const Vector2f01&   stochastic,
	                           const UnitVector3f& inDirection,
	                           UnitVector3f&       outDirection,
	                           Light&              scaling )                               const;

	/// clone
	virtual Brdf* clone()                                                                  const =0;

	/// coordinate frame
	        const UnitVector3f& getNormal()                                                const;
	        const UnitVector3f& getTangentX()                                              const;
	        const UnitVector3f& getTangentY()                                              const;

	virtual void  getDirectionFromFrameCoords( const Vector3f& frameCoords,
	                                           UnitVector3f&   direction )                 const;
	virtual void  getFrameCoordsFromDirection( const UnitVector3f& direction,
	                                           Vector3f&           frameCoords )           const;
	virtual void  getDirectionFrom2DCoords( const Vector2f01& coords,
	                                        UnitVector3f&     direction )                  const;
	virtual void  get2DCoordsFromDirection( const UnitVector3f& direction,
	                                        Vector2f01&         coords )                   const;

	virtual void  getUniformSampleDirection( const Vector2f01& stochastic,
	                                         UnitVector3f&     direction )                 const;

	/// utility
	static const float PI;
	static const float TWO_PI;
	static const float FOUR_PI;
	static const float ONE_OVER_PI;
	static const float ONE_OVER_TWO_PI;
	static const float ONE_OVER_FOUR_PI;

	/// test
	virtual bool  test( bool isLogging )                                                   const;
	virtual void  testView( bool                 isBrdfNotSdf,
	                        dword                partitionsBits,
	                        bool                 isCos,
	                        bool                 isTransmission,
	                        const Sheet<CieXyz>& image )                                   const;


/// implementation ---------------------------------------------------------------------------------
protected:
	/// customization
	virtual Float01 getReflectivity_()                                                     const =0;
	virtual void    getPartitions_( float   proportions[PARTITIONS_MAX],
	                                Float01 peakness[PARTITIONS_MAX],
	                                dword&  length )                                       const =0;
	virtual Float01 getCos_( dword               part,
	                         const UnitVector3f& outDirection )                            const =0;
	/** @precondition (normal.dot( inDirection ) >= 0.0f) & (normal.dot( outDirection ) >= 0.0f) */
	virtual void    getScaling_( dword               part,
	                             float               IdotN,
	                             float               OdotN,
	                             const UnitVector3f& inDirection,
	                             const UnitVector3f& outDirection,
	                             Light&              scaling )                             const =0;
	/** @precondition normal.dot( inDirection ) >= 0.0f */
	virtual void    getSample_( dword               part,
	                            const Vector2f01&   stochastic,
	                            const UnitVector3f& inDirection,
	                            UnitVector3f&       outDirection,
	                            Light&              scaling )                              const =0;


	/// test
	static  bool  testAll( const Brdf& brdf,
	                       bool        isLogging );
	static  bool  testBrdf( const Brdf& brdf,
	                        bool        isLogging );
	static  bool  testSdf( const Brdf& brdf,
	                       bool        isLogging );
	static  bool  testOrientation( const Brdf& brdf,
	                               bool        isLogging );
	static  void  viewAll( const Brdf&          brdf,
	                       bool                 isBrdfNotSdf,
	                       dword                partitionsBits,
	                       bool                 isCos,
	                       bool                 isTransmission,
	                       const Sheet<CieXyz>& image );
	static  void  viewBrdf( const Brdf&          brdf,
	                        dword                partitionsBits,
	                        bool                 isCos,
	                        bool                 isTransmission,
	                        const Sheet<CieXyz>& image );
	static  void  viewSdf( const Brdf&          brdf,
	                       dword                partitionsBits,
	                       bool                 isCos,
	                       bool                 isTransmission,
	                       const Sheet<CieXyz>& image );


/// fields -----------------------------------------------------------------------------------------
private:
	/// 3d frame
	UnitVector3f normal_m;
	UnitVector3f tangentX_m;
	UnitVector3f tangentY_m;
};








/// inlines ///

/// queries ----------------------------------------------------------------------------------------
inline
const UnitVector3f& Brdf::getNormal() const
{
	return normal_m;
}


inline
const UnitVector3f& Brdf::getTangentX() const
{
	return tangentX_m;
}


inline
const UnitVector3f& Brdf::getTangentY() const
{
	return tangentY_m;
}


inline
void Brdf::getUniformSampleDirection
(
	const Vector2f01& stochastic,
	UnitVector3f&     direction
) const
{
	Brdf::getDirectionFrom2DCoords( stochastic, direction );
}


}}




#endif//Brdf_h
