/*--------------------------------------------------------------------

   Perceptuum renderer, version 2.1
   Copyright (c) 2002-2003,  Harrison Ainsworth.

   www.hxa7241.org

--------------------------------------------------------------------*/

/*--------------------------------------------------------------------

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston,
   MA 02111-1307, USA

--------------------------------------------------------------------*/


#include "Light.h"
#include "Polar.h"

#include "Bssdf.h"   /// own header is included last


using namespace perceptuum2::interaction;




/// standard object services -----------------------------------------------------------------------
Bssdf::Bssdf()
 :	Brdf()
{
	Bssdf::setBssdf( 1.0f, 1.0f, 0.0f );
}


Bssdf::Bssdf
(
	const UnitVector3f& normal,
	const UnitVector3f& tangent,
	const float         refractiveIndexOutside,
	const float         refractiveIndexInside,
	const Float01       transparency
)
 :	Brdf( normal, tangent )
{
	Bssdf::setBssdf( refractiveIndexOutside, refractiveIndexInside, transparency );
}


Bssdf::~Bssdf()
{
}


Bssdf::Bssdf( const Bssdf& other )
{
	Bssdf::operator=( other );
}


Bssdf& Bssdf::operator=( const Bssdf& other )
{
	if( &other != this )
	{
		Brdf::operator=( other );

		refractiveIndexOutside_m  = other.refractiveIndexOutside_m;
		refractiveIndexInside_m   = other.refractiveIndexInside_m;
		transparency_m            = other.transparency_m;

		relativeRefractiveIndex_m = other.relativeRefractiveIndex_m;
	}

	return *this;
}




/// commands ---------------------------------------------------------------------------------------
void Bssdf::setBssdf
(
	const float   refractiveIndexOutside,
	const float   refractiveIndexInside,
	const Float01 transparency
)
{
	refractiveIndexOutside_m = refractiveIndexOutside;
	/// (superluminal phase velocity is not supported)
	if( refractiveIndexOutside_m < 1.0f )
	{
		refractiveIndexOutside_m = 1.0f;
	}

	refractiveIndexInside_m  = refractiveIndexInside;
	/// (superluminal phase velocity is not supported)
	if( refractiveIndexInside_m < 1.0f )
	{
		refractiveIndexInside_m = 1.0f;
	}

	transparency_m = transparency;

	relativeRefractiveIndex_m = refractiveIndexOutside_m / refractiveIndexInside_m;
}




/// queries ----------------------------------------------------------------------------------------
void Bssdf::getScaling
(
	const dword         part,
	const UnitVector3f& inDirection,
	const UnitVector3f& outDirection,
	Light&              scaling
) const
{
	/// accelerate dominant special case
	if( transparency_m == 0.0f )
	{
		Brdf::getScaling( part, inDirection, outDirection, scaling );
	}
	else
	{
		UnitVector3f inReversed( inDirection );
		inReversed.negate();

		float transportScaling;// = 0.0f;

		/// make fake directions
		UnitVector3f inReversedT( inReversed );
		UnitVector3f outDirectionT( outDirection );
		{
			const float NdotI     = getNormal().dot( inReversed );
			const float NdotO     = getNormal().dot( outDirection );
			const bool  inInside  = NdotI < 0.0f;
			const bool  outInside = NdotO < 0.0f;
			const bool  isTir     = inReversedT.isInTirZone( getNormal(), relativeRefractiveIndex_m );

			/// if rays on either side - maybe transmission
			if( inInside ^ outInside )
			{
				/// if inward in tir zone, no transmission
				if( isTir || outDirection.isInTirZone( getNormal(), relativeRefractiveIndex_m ) )
				{
					transportScaling = 0.0f;
				}
				/// else, bring the inside vector to the outside
				else
				{
					transportScaling = transparency_m;

					/// pick inside vector
					UnitVector3f* pDirection = inInside ? &inReversedT : &outDirectionT;

					/// refract, and change side across normal
					pDirection->refractEq( getNormal(), relativeRefractiveIndex_m );
					pDirection->reflectEq( getNormal() );
				}
			}
			/// else rays on same side - reflection, internal or external
			else
			{
				/// if inward in tir zone, entirely reflection
				if( isTir )
				{
					transportScaling = 1.0f;
				}
				/// else partial reflection
				else
				{
					transportScaling = 1.0f - transparency_m;
				}

				/// if inside bring to the outside
				if( inInside )
				{
					/// change side: reflect thru surface
					inReversedT.reflectAlongEq( getNormal() );
					outDirectionT.reflectAlongEq( getNormal() );
				}
			}
		}

		if( transportScaling > 0.0f )
		{
			Brdf::getScaling( part, -inReversedT, outDirectionT, scaling );
		}

		scaling *= transportScaling;
	}
}


bool Bssdf::getSample
(
	const dword         part,
	const Float01       roulette,
	const Vector2f01&   stochastic,
	const UnitVector3f& inDirection,
	UnitVector3f&       outDirection,
	Light&              scaling
) const
{
	/// accelerate dominant special case
	if( transparency_m == 0.0f )
	{
		Brdf::getSample( part, roulette, stochastic, inDirection, outDirection, scaling );
		return false;
	}
	else
	{
		UnitVector3f inReversed( inDirection );
		inReversed.negate();

		const bool doReflection = (roulette >= transparency_m) ||
								  inReversed.isInTirZone( getNormal(), relativeRefractiveIndex_m );
		float rouletteT;
		if( roulette >= transparency_m )
		{
			rouletteT = (roulette - transparency_m) * (1.0f / (1.0f - transparency_m));
		}
		else
		{
			rouletteT = roulette * (1.0f / transparency_m);
		}

		/// from inside
		if( getNormal().dot( inReversed ) < 0.0f )
		{
			if( doReflection )
			{
				/// change side: reflect thru surface
				inReversed.reflectAlongEq( getNormal() );

				Brdf::getSample( part, rouletteT, stochastic, -inReversed, outDirection, scaling );

				/// change side back: reflect thru surface
				outDirection.reflectAlongEq( getNormal() );
			}
			else
			{
				/// refract and change side across normal
				inReversed.refractEq( getNormal(), relativeRefractiveIndex_m );
				inReversed.reflectEq( getNormal() );

				Brdf::getSample( part, rouletteT, stochastic, -inReversed, outDirection, scaling );
			}
		}
		/// from outside
		else
		{
			Brdf::getSample( part, rouletteT, stochastic, -inReversed, outDirection, scaling );

			if( !doReflection )
			{
				/// refract and change side across normal
				outDirection.refractEq( getNormal(), relativeRefractiveIndex_m );
				outDirection.reflectEq( getNormal() );
			}
		}

		return !doReflection;
	}
}




void Bssdf::getDirectionFromCoords
(
	const Vector2f01& coords,
	const bool        isUpper,
	UnitVector3f&     direction
) const
{
	float x, y, z;
	hxa7241::graphics::squareToHemisphere( coords.getX(), coords.getY(), x, y, z );

	if( !isUpper )
	{
		z = -z;
	}
	//z *= float((dword(isUpper) << 1) - 1);

	getDirectionFromFrameCoords( Vector3f(x,y,z), direction );
}


void Bssdf::getCoordsFromDirection
(
	const UnitVector3f& direction,
	Vector2f01&         coords,
	bool&               isUpper
) const
{
	Vector3f frameCoords;
	getFrameCoordsFromDirection( direction, frameCoords );

	isUpper = frameCoords.getZ() >= 0.0f;
	if( !isUpper )
	{
		frameCoords.z_m = -frameCoords.z_m;
	}

	float x, y;
	hxa7241::graphics::hemisphereToSquare( frameCoords.getX(), frameCoords.getY(), frameCoords.getZ(), x, y );
	coords.setXY( x, y );
}
