/*--------------------------------------------------------------------

   Perceptuum renderer, version 2.1
   Copyright (c) 2002-2003,  Harrison Ainsworth.

   www.hxa7241.org

--------------------------------------------------------------------*/

/*--------------------------------------------------------------------

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston,
   MA 02111-1307, USA

--------------------------------------------------------------------*/


#include "BssdfConstant.h"   /// own header is included last


using namespace perceptuum2::interaction;




/// standard object services -----------------------------------------------------------------------
BssdfConstant::BssdfConstant
(
	const UnitVector3f& normal,
	const UnitVector3f& tangent,
	const float         refractiveIndexOutside,
	const float         refractiveIndexInside,
	const Float01       transparency,
	const Light&        reflectivity
)
 :	Bssdf( normal, tangent, refractiveIndexOutside, refractiveIndexInside, transparency )
{
	BssdfConstant::setBssdfConstant( reflectivity );
}


BssdfConstant::~BssdfConstant()
{
}


BssdfConstant::BssdfConstant( const BssdfConstant& other )
{
	BssdfConstant::operator=( other );
}


BssdfConstant& BssdfConstant::operator=( const BssdfConstant& other )
{
	if( &other != this )
	{
		Bssdf::operator=( other );

		reflectivitySpectral_m = other.reflectivitySpectral_m;

		reflectivity_m = other.reflectivity_m;
		color_m        = other.color_m;
	}

	return *this;
}




/// commands ---------------------------------------------------------------------------------------
void BssdfConstant::setBssdfConstant
(
	const Light& reflectivity
)
{
	/// precondition reflectivity
	reflectivitySpectral_m = reflectivity;
	reflectivitySpectral_m.clampTo01Positive();

	/// set auxilliarys
	{
		float mean;
		reflectivitySpectral_m.getMeanAndColor( mean, color_m );
		reflectivity_m = mean;
	}
}




/// queries ----------------------------------------------------------------------------------------
Brdf* BssdfConstant::clone() const
{
	Brdf* pClone = new BssdfConstant( *this );
	if( pClone == 0 )
	{
		throw "BssdfConstant::clone()\ncouldnt get memory";
	}

	return  pClone;
}




/// implementation ---------------------------------------------------------------------------------
Float01 BssdfConstant::getReflectivity_() const
{
	return reflectivity_m;
}


void BssdfConstant::getPartitions_
(
	float   proportions[],
	Float01 peakness[],
	dword&  length
) const
{
	proportions[0] = 1.0f;
	peakness[0]    = Float01::MIN();

	length = 1;
}


Float01 BssdfConstant::getCos_
(
	const dword         ,//part,
	const UnitVector3f& outDirection
) const
{
	float cos = getNormal().dot( outDirection );
	if( cos < 0.0f )
	{
		cos = -cos;
	}

	return cos;
}


void BssdfConstant::getScaling_
(
	const dword         ,//part,
	const float         ,//IdotN,
	const float         ,//OdotN,
	const UnitVector3f& ,//inDirection,
	const UnitVector3f& ,//outDirection,
	Light&              scaling
) const
{
	(scaling = color_m) *= ONE_OVER_PI;
}


void BssdfConstant::getSample_
(
	const dword         ,//part,
	const Vector2f01&   stochastic,
	const UnitVector3f& ,//inDirection,
	UnitVector3f&       outDirection,
	Light&              scaling
) const
{
	getUniformSampleDirection( stochastic, outDirection );

	scaling = color_m;
}








/*void BssdfConstant::sdfCos
(
	const UnitVector3f& inDirection,
	const Vector2f01&   stochastic,
	const Float01       roulette,
	const bool          isScaleByReflectivity,
	UnitVector3f&       outDirection,
	Light&              scaling,
) const
{
	static const float NORMALISATION = (3.0 / TWO_PI);

	if( ((roulette < (reflectivity_m * NORMALISATION)) | isScaleByReflectivity) )
	{
		{
			float diskX, diskY;
			hxa7241::graphics::squareToDisk( stochastic.getX(), stochastic.getY(), diskX, diskY );

			const float z = ::sqrt( 1.0f - ((diskX * diskX) + (diskY * diskY)) );
			getDirectionFromFrameCoords( Vector3f(diskX, diskY, z), outDirection );
		}

		if( isScaleByReflectivity )
		{
			(scaling = reflectivitySpectral_m) *= NORMALISATION;
		}
		else
		{
			scaling = color_m;
		}
	}
	else
	{
		/// no transmission, no importance sample
		outDirection = UnitVector3f::ZERO();
		scaling      = Light::ZERO();
	}
}*/
