/*--------------------------------------------------------------------

   Perceptuum renderer, version 2.1
   Copyright (c) 2002-2003,  Harrison Ainsworth.

   www.hxa7241.org

--------------------------------------------------------------------*/

/*--------------------------------------------------------------------

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston,
   MA 02111-1307, USA

--------------------------------------------------------------------*/


#include "Polar.h"
#include "hxa7241math.h"

#include "BssdfPhong.h"   /// own header is included last


using namespace perceptuum2::interaction;




/// standard object services -----------------------------------------------------------------------
BssdfPhong::BssdfPhong()
 :	Bssdf()
{
	BssdfPhong::setBssdfPhong( 0.0f, Light::ONE(), Light::ONE(), 1.0f );
}


BssdfPhong::BssdfPhong
(
	const UnitVector3f& normal,
	const UnitVector3f& tangent,
	const float         refractiveIndexOutside,
	const float         refractiveIndexInside,
	const Float01       transparency,
	const Float01       reflectivity,
	const Light&        colorDiffuse,
	const Light&        colorSpecular,
	const Float01       roughness
)
 :	Bssdf( normal, tangent, refractiveIndexOutside, refractiveIndexInside, transparency )
{
	BssdfPhong::setBssdfPhong( reflectivity, colorDiffuse, colorSpecular, roughness );
}


BssdfPhong::~BssdfPhong()
{
}


BssdfPhong::BssdfPhong( const BssdfPhong& other )
{
	BssdfPhong::operator=( other );
}


BssdfPhong& BssdfPhong::operator=( const BssdfPhong& other )
{
	if( &other != this )
	{
		Bssdf::operator=( other );

		reflectivity_m  = other.reflectivity_m;
		colorDiffuse_m  = other.colorDiffuse_m;
		colorSpecular_m = other.colorSpecular_m;
		roughness_m     = other.roughness_m;

		colorGloss_m    = other.colorGloss_m;

		diffuseness_m   = other.diffuseness_m;
		glossyness_m    = other.glossyness_m;
		specularness_m  = other.specularness_m;

		exponent_m      = other.exponent_m;
		normalisation_m = other.normalisation_m;
	}

	return *this;
}




/// commands ---------------------------------------------------------------------------------------
void BssdfPhong::setBssdfPhong
(
	const Float01 reflectivity,
	const Light&  colorDiffuse,
	const Light&  colorSpecular,
	const Float01 roughness
)
{
	/// precondition colors
	{
		colorDiffuse_m  = colorDiffuse;
		colorDiffuse_m.setToMean1Color();

		colorSpecular_m = colorSpecular;
		colorSpecular_m.setToMean1Color();
	}

	reflectivity_m = reflectivity;
	roughness_m    = roughness;


	/// set auxilliarys
	{
		((colorGloss_m = colorDiffuse_m) += colorSpecular_m) *= 0.5f;

		glossyness_m   = (4.0f * roughness_m * (1.0f - roughness_m));
		diffuseness_m  = (roughness_m < 0.5f ? 0.0f : 1.0f - glossyness_m);
		specularness_m = (roughness_m < 0.5f ? 1.0f - glossyness_m : 0.0f);

		const float roughnessClamped = roughness_m > 0.001f ? roughness_m : 0.001;
		exponent_m      = /*hxa7241::utility::fpToInt*/( (1.0f / roughnessClamped) - 1.0f );
		normalisation_m = (exponent_m + 2.0f) / TWO_PI;
	}
}




/// queries ----------------------------------------------------------------------------------------
Brdf* BssdfPhong::clone() const
{
	Brdf* pClone = new BssdfPhong( *this );
	if( pClone == 0 )
	{
		throw "BssdfPhong::clone()\ncouldnt get memory";
	}

	return  pClone;
}




/// implementation ---------------------------------------------------------------------------------
Float01 BssdfPhong::getReflectivity_() const
{
	return reflectivity_m;
}


void BssdfPhong::getPartitions_
(
	float   proportions[],
	Float01 peakness[],
	dword&  length
) const
{
	proportions[DIFFUSE_PART]  = diffuseness_m;
	peakness[DIFFUSE_PART]     = Float01::MIN();

	proportions[GLOSSY_PART]   = glossyness_m;
	peakness[GLOSSY_PART]      = roughness_m.MAX() - roughness_m;

	proportions[SPECULAR_PART] = specularness_m;
	peakness[SPECULAR_PART]    = Float01::MAX();

	length = 3;
}


Float01 BssdfPhong::getCos_
(
	const dword         part,
	const UnitVector3f& outDirection
) const
{
	float cos;

	if( part != SPECULAR_PART )
	{
		cos = getNormal().dot( outDirection );
		if( cos < 0.0f )
		{
			cos = -cos;
		}

		if( part == GLOSSY_PART )
		{
			cos = 1.0f + (roughness_m * (cos - 1.0f));
		}
	}
	else
	{
		cos = 1.0f;
	}

	return cos;
}


void BssdfPhong::getScaling_
(
	const dword         part,
	const float         ,//IdotN,
	const float         ,//OdotN,
	const UnitVector3f& inDirection,
	const UnitVector3f& outDirection,
	Light&              scaling
) const
{
	switch( part )
	{
		case DIFFUSE_PART  :
		{
			brdfDiffuse( inDirection, outDirection, scaling );
			break;
		}
		case GLOSSY_PART   :
		{
			brdfGlossy( inDirection, outDirection, scaling );
			break;
		}
		case SPECULAR_PART :
		{
			brdfSpecular( inDirection, outDirection, scaling );
			break;
		}
		default :
		{
			scaling = Light::ZERO();
			break;
		}
	}
}


void BssdfPhong::getSample_
(
	const dword         part,
	const Vector2f01&   stochastic,
	const UnitVector3f& inDirection,
	UnitVector3f&       outDirection,
	Light&              scaling
) const
{
	switch( part )
	{
		case DIFFUSE_PART  :
		{
			sdfDiffuse( stochastic, inDirection, outDirection, scaling );
			break;
		}
		case GLOSSY_PART   :
		{
			sdfGlossy( stochastic, inDirection, outDirection, scaling );
			break;
		}
		case SPECULAR_PART :
		{
			sdfSpecular( stochastic, inDirection, outDirection, scaling );
			break;
		}
		default :
		{
			outDirection = UnitVector3f::ZERO();
			scaling      = Light::ZERO();
			break;
		}
	}
}




inline
void BssdfPhong::brdfDiffuse
(
	const UnitVector3f& ,//inDirection,
	const UnitVector3f& ,//outDirection,
	Light&              scaling
) const
{
	(scaling = colorDiffuse_m) *= ONE_OVER_PI;
}


void BssdfPhong::brdfGlossy
(
	const UnitVector3f& inDirection,
	const UnitVector3f& outDirection,
	Light&              scaling
) const
{
	/// precondition: (getNormal().dot( inDirection ) >= 0.0f) & (getNormal().dot( outDirection ) >= 0.0f) )

	float cosPower;// = 1.0f;
	{
		UnitVector3f mirrorDirection( inDirection );
		mirrorDirection.reflectEq( getNormal() );

		const float outDotMirror = mirrorDirection.dot( outDirection );

		if( outDotMirror > 0.0f )
		{
			cosPower = float( ::pow( outDotMirror, exponent_m ) );

			/*/// for integer exponents...
			switch( exponent_m )
			{
				default :
				{
					dword i = exponent_m;
					do
					{
						cosPower *= outDotMirror;
					}
					while( --i > 0 );

					break;
				}
				case  6 :  cosPower *= outDotMirror;
				case  5 :  cosPower *= outDotMirror;
				case  4 :  cosPower *= outDotMirror;
				case  3 :  cosPower *= outDotMirror;
				case  2 :  cosPower *= outDotMirror;
				case  1 :  cosPower *= outDotMirror;
				case  0 :  ;
			}*/
		}
		else
		{
			cosPower = 0.0f;
		}
	}

	///// derived by numerical experimentation... (violates energy conservation at some angles)
	///// maybe look more carefully into something better...
	//const float q = (roughness_m * 0.315f) + ((1.0f - roughness_m) * 0.51f);
	//const float undersurfaceLobeCompensation = 1.0f / (1.0f - q);

	(scaling = colorGloss_m) *= (normalisation_m * cosPower);// * undersurfaceLobeCompensation);
}


inline
void BssdfPhong::brdfSpecular
(
	const UnitVector3f& ,//inDirection,
	const UnitVector3f& ,//outDirection,
	Light&              scaling
) const
{
	/// not really do-able numerically...

	/*UnitVector3f mirrorDirection( inDirection );
	mirrorDirection.reflectEq( getNormal() );

	const float outDotMirror = mirrorDirection.dot( outDirection );
	if( outDotMirror >= 1.0f )
	{
		(scaling = colorSpecular_m) *= FLOAT_LARGE;
	}
	else
	{
		scaling = Light::ZERO();
	}*/


	scaling = Light::ZERO();
}


inline
void BssdfPhong::sdfDiffuse
(
	const Vector2f01&   stochastic,
	const UnitVector3f& ,//inDirection,
	UnitVector3f&       outDirection,
	Light&              scaling
) const
{
	getUniformSampleDirection( stochastic, outDirection );
	scaling = colorDiffuse_m;
}


void BssdfPhong::sdfGlossy
(
	const Vector2f01&   stochastic,
	const UnitVector3f& inDirection,
	UnitVector3f&       outDirection,
	Light&              scaling
) const
{
	/// precondition: getNormal().dot( inDirection ) >= 0.0f

	const float angleAroundMirror = TWO_PI * stochastic.getY();

	float sinAngleAroundMirror;
	float cosAngleAroundMirror;
	hxa7241::graphics::sinAndcos( angleAroundMirror, sinAngleAroundMirror, cosAngleAroundMirror );

	/// make hemispherical coefficients
	const float z = float( ::pow( stochastic.getX(), 1.0f / (exponent_m + 1.0f) ) );
	const float s = float( ::sqrt( 1.0f - (z * z) ) );
	const float x = cosAngleAroundMirror * s;
	const float y = sinAngleAroundMirror * s;

	/// make direction vector
	{
		using hxa7241::graphics::Vector3f;

		/// make components
		UnitVector3f mirrorDir( inDirection );
		mirrorDir.reflectEq( getNormal() );

		UnitVector3f mDirA( getNormal() );
		{
			float d = mDirA.dot( mirrorDir );
			if( (d == 1.0f) | (d == -1.0f) )
			{
				mDirA = getTangentX();
				d = mDirA.dot( mirrorDir );
			}

			if( d != 0.0f )
			{
				mDirA = mDirA.getVector() - (mirrorDir.getVector() * d);
			}
		}

		UnitVector3f mDirB( mirrorDir );
		mDirB.crossEq( mDirA );

		/// scale components
		Vector3f dir( mirrorDir.getVector() * z );
		dir += (mDirA.getVector() * x);
		dir += (mDirB.getVector() * y);

		outDirection = dir;
	}

	/// set scaling
	scaling = colorGloss_m;


	/// clamp direction to hemisphere
	{
		const float nDotOut = getNormal().dot( outDirection );
		if( nDotOut < 0.0f )
		{
			outDirection = outDirection.getVector() - (getNormal().getVector() * nDotOut);
			if( outDirection.isZero() )
			{
				outDirection = getTangentX();
			}
		}
	}
}


inline
void BssdfPhong::sdfSpecular
(
	const Vector2f01&   stochastic,
	const UnitVector3f& inDirection,
	UnitVector3f&       outDirection,
	Light&              scaling
) const
{
	/// precondition: getNormal().dot( inDirection ) >= 0.0f

	outDirection = inDirection;
	outDirection.reflectEq( getNormal() );

	scaling = colorSpecular_m;
}
