/*--------------------------------------------------------------------

   Perceptuum renderer, version 2.1
   Copyright (c) 2002-2003,  Harrison Ainsworth.

   www.hxa7241.org

--------------------------------------------------------------------*/

/*--------------------------------------------------------------------

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston,
   MA 02111-1307, USA

--------------------------------------------------------------------*/


#include "Polar.h"
#include "hxa7241math.h"

#include "BssdfWard.h"   /// own header is included last


using namespace perceptuum2::interaction;




/// standard object services -----------------------------------------------------------------------
BssdfWard::BssdfWard()
 :	Bssdf()
{
	BssdfWard::setBssdfWard( Light::ZERO(), Light::ZERO(), 1.0f, 1.0f );
}


BssdfWard::BssdfWard
(
	const UnitVector3f& normal,
	const UnitVector3f& tangent,
	const float         refractiveIndexOutside,
	const float         refractiveIndexInside,
	const Float01       transparency,
	const Light&        reflectivityDiffuse,
	const Light&        reflectivitySpecular,
	const Float01       roughnessX,
	const Float01       roughnessY
)
 :	Bssdf( normal, tangent, refractiveIndexOutside, refractiveIndexInside, transparency )
{
	BssdfWard::setBssdfWard( reflectivityDiffuse, reflectivitySpecular, roughnessX, roughnessY );
}


BssdfWard::~BssdfWard()
{
}


BssdfWard::BssdfWard( const BssdfWard& other )
{
	BssdfWard::operator=( other );
}


BssdfWard& BssdfWard::operator=( const BssdfWard& other )
{
	if( &other != this )
	{
		Bssdf::operator=( other );

		reflectivityDiffuse_m  = other.reflectivityDiffuse_m;
		reflectivitySpecular_m = other.reflectivitySpecular_m;
		roughnessX_m           = other.roughnessX_m;
		roughnessY_m           = other.roughnessY_m;

		reflectivity_m  = other.reflectivity_m;
		colorDiffuse_m  = other.colorDiffuse_m;
		colorSpecular_m = other.colorSpecular_m;
		diffuseness_m   = other.diffuseness_m;
	}

	return *this;
}




/// commands ---------------------------------------------------------------------------------------
void BssdfWard::setBssdfWard
(
	const Light&  reflectivityDiffuse,
	const Light&  reflectivitySpecular,
	const Float01 roughnessX,
	const Float01 roughnessY
)
{
	/// precondition reflectivity
	{
		reflectivityDiffuse_m  = reflectivityDiffuse;
		reflectivitySpecular_m = reflectivitySpecular;

		reflectivityDiffuse_m.clampToPositive();
		reflectivitySpecular_m.clampToPositive();

		Light reflectivity;
		(reflectivity = reflectivityDiffuse_m) += reflectivitySpecular_m;
		const float mean = reflectivity.average();
		if( mean >= 1.0f )
		{
			const float rescaling = FLOAT_ALMOST_ONE / mean;
			reflectivityDiffuse_m  *= rescaling;
			reflectivitySpecular_m *= rescaling;
		}
	}

	/// precondition roughness
	{
		roughnessX_m = roughnessX;// * 0.25f;
		roughnessY_m = roughnessY;// * 0.25f;

		static const float ALPHA_MIN = 0.001;//0.035f;   /// numerical tests showed overall reflectance increasing with alpha < ~that
		if( roughnessX_m < ALPHA_MIN )
		{
			roughnessX_m = ALPHA_MIN;
		}
		if( roughnessY_m < ALPHA_MIN )
		{
			roughnessY_m = ALPHA_MIN;
		}
	}

	/// set auxilliarys
	{
		float rdMean;
		float rsMean;
		reflectivityDiffuse_m.getMeanAndColor( rdMean, colorDiffuse_m );
		reflectivitySpecular_m.getMeanAndColor( rsMean, colorSpecular_m );

		reflectivity_m = rdMean + rsMean;
		diffuseness_m  = (reflectivity_m > 0.0f) ? (rdMean / reflectivity_m) : 1.0f;
	}
}




/// queries ----------------------------------------------------------------------------------------
Brdf* BssdfWard::clone() const
{
	Brdf* pClone = new BssdfWard( *this );
	if( pClone == 0 )
	{
		throw "BssdfWard::clone()\ncouldnt get memory";
	}

	return  pClone;
}




/// implementation ---------------------------------------------------------------------------------
Float01 BssdfWard::getReflectivity_() const
{
	return reflectivity_m;
}


void BssdfWard::getPartitions_
(
	float   proportions[],
	Float01 peakness[],
	dword&  length
) const
{
	proportions[DIFFUSE_PART]  = diffuseness_m;
	peakness[DIFFUSE_PART]     = Float01::MIN();

	proportions[SPECULAR_PART] = 1.0f - diffuseness_m;
	peakness[SPECULAR_PART]    = roughnessX_m.MAX() - ((roughnessX_m + roughnessY_m) * 0.5f);

	length = 2;
}


Float01 BssdfWard::getCos_
(
	const dword         part,
	const UnitVector3f& outDirection
) const
{
	float cos = 1.0f;

	if( part != SPECULAR_PART )
	{
		cos = getNormal().dot( outDirection );
		if( cos < 0.0f )
		{
			cos = -cos;
		}
	}


	/*float cos = getNormal().dot( outDirection );
	if( cos < 0.0f )
	{
		cos = -cos;
	}

	if( part == SPECULAR_PART )
	{
		float roughness;
		if( roughnessX_m == roughnessY_m )
		{
			roughness = roughnessX_m;
		}
		else
		{
			const float OdotX = outDirection.dot( getTangentX() );
			const float OdotY = outDirection.dot( getTangentY() );

			/// make linear interpolation (OdotX^2 and OdotY^2 being 0-1 complements)
			roughness = ((OdotX * OdotX) * roughnessX_m) + ((OdotY * OdotY) * roughnessY_m);
		}
		cos = 1.0f + (roughness * (cos - 1.0f));
	}*/


	return cos;
}


void BssdfWard::getScaling_
(
	const dword         part,
	const float         IdotN,
	const float         OdotN,
	const UnitVector3f& inDirection,
	const UnitVector3f& outDirection,
	Light&              scaling
) const
{
	switch( part )
	{
		case DIFFUSE_PART  :
		{
			brdfDiffuse( inDirection, outDirection, scaling );
			break;
		}
		case SPECULAR_PART :
		{
			brdfSpecular( inDirection, outDirection, IdotN, OdotN, scaling );
			break;
		}
		default :
		{
			scaling = Light::ZERO();
			break;
		}
	}
}


void BssdfWard::getSample_
(
	const dword         part,
	const Vector2f01&   stochastic,
	const UnitVector3f& inDirection,
	UnitVector3f&       outDirection,
	Light&              scaling
) const
{
	switch( part )
	{
		case DIFFUSE_PART  :
		{
			sdfDiffuse( stochastic, inDirection, outDirection, scaling );
			break;
		}
		case SPECULAR_PART :
		{
			sdfSpecular( stochastic, inDirection, outDirection, scaling );
			break;
		}
		default :
		{
			outDirection = UnitVector3f::ZERO();
			scaling      = Light::ZERO();
			break;
		}
	}
}




inline
void BssdfWard::brdfDiffuse
(
	const UnitVector3f& ,//inDirection,
	const UnitVector3f& ,//outDirection,
	Light&              scaling
) const
{
	(scaling = colorDiffuse_m) *= ONE_OVER_PI;
}


void BssdfWard::brdfSpecular
(
	const UnitVector3f& inDirection,
	const UnitVector3f& outDirection,
	const float         ,//IdotN,
	const float         ,//OdotN,
	Light&              scaling
) const
{
	/// precondition: (getNormal().dot( inDirection ) >= 0.0f) & (getNormal().dot( outDirection ) >= 0.0f) )

	/// modification:
	/// the 1/sqrt(IdotN * OdotN) term in the original ward formulation is apparently to cancel the cos term
	/// in the reflection equation, but since this class sets the cos term to 1 for specular reflection, the
	/// 1/sqrt(IdotN * OdotN) term is not appropriate and has been removed.
	/// ok, it not quite the same, but very nearly. the peakness means the in and out directions will be
	/// similar where the function is not (near) zero.

	const UnitVector3f h( inDirection.getVector() + outDirection.getVector() );
	const float HdotN       = h.dot( getNormal() );
	//const float IdotNxOdotN = IdotN * OdotN;

	if( HdotN < 1e-9f )
	//if( (HdotN == 0.0f) | (IdotNxOdotN == 0.0f) )
	{
		/// grazing vectors
		if( h.isZero() )
		{
			/// straight across
			scaling = colorSpecular_m;
		}
		else
		{
			/// different tangent directions
			scaling = Light::ZERO();
		}
	}
	else
	{
		float tanExpr;
		if( roughnessX_m == roughnessY_m )
		{
			float oneMinusHdotN2 = 1.0f - (HdotN * HdotN);
			oneMinusHdotN2 = oneMinusHdotN2 >= 0.0f ? oneMinusHdotN2 : 0.0f;

			const float tanDelta = float(::sqrt( oneMinusHdotN2 )) / HdotN;
			//const float tanDelta = float(::tan( ::acos( HdotN ) ));

			/// equation (4) particular piece: inside the exp()
			tanExpr = (tanDelta * tanDelta) / (roughnessX_m * roughnessY_m);
		}
		else
		{
			const float HdotXoverAx = h.dot( getTangentX() ) / roughnessX_m;
			const float HdotYoverAy = h.dot( getTangentY() ) / roughnessY_m;

			/// equation (5b) particular piece: inside the exp()
			tanExpr = 2.0f * (((HdotXoverAx * HdotXoverAx) + (HdotYoverAy * HdotYoverAy)) / (1.0f + HdotN));
		}

		/// equation (4) and (5a) shared specular piece,
		scaling  = colorSpecular_m;
		scaling *= ( float(::exp( -tanExpr )) / (FOUR_PI * roughnessX_m * roughnessY_m) );
		//scaling *= ( float(::exp( -tanExpr )) / (float(::sqrt( IdotNxOdotN )) * (FOUR_PI * roughnessX_m * roughnessY_m)) );
	}
}


inline
void BssdfWard::sdfDiffuse
(
	const Vector2f01&   stochastic,
	const UnitVector3f& ,//inDirection,
	UnitVector3f&       outDirection,
	Light&              scaling
) const
{
	getUniformSampleDirection( stochastic, outDirection );
	scaling = colorDiffuse_m;
}


void BssdfWard::sdfSpecular
(
	const Vector2f01&   stochastic,
	const UnitVector3f& inDirection,
	UnitVector3f&       outDirection,
	Light&              scaling
) const
{
	/// precondition: getNormal().dot( inDirection ) >= 0.0f

	/// angle around normal
	float hAngleToTangentX;
	{
		if( roughnessX_m == roughnessY_m )
		{
			/// equation (7b) isotropic
			hAngleToTangentX = TWO_PI * stochastic.getY();
		}
		else
		{
			/// equation (7b) anisotropic (keeping the angle in its starting quadrant)
			hAngleToTangentX = float(::atan( (roughnessY_m / roughnessX_m) * float(::tan( TWO_PI * (stochastic.getY() - 0.25f) )) ));
			if( stochastic.getY() >= 0.5 )
			{
				hAngleToTangentX += PI;
			}
		}
	}

	/// angle to normal
	float cosHAngleToX;
	float sinHAngleToX;
	hxa7241::graphics::sinAndcos( hAngleToTangentX, sinHAngleToX, cosHAngleToX );
	float hAngleToNormal;
	{
		const float cosHAngleToXOverAx = cosHAngleToX / roughnessX_m;
		const float sinHAngleToXOverAy = sinHAngleToX / roughnessY_m;

		/// equation (7a)
		hAngleToNormal = float(::sqrt( -float(::log( stochastic.getX() )) /
						 ( (cosHAngleToXOverAx * cosHAngleToXOverAx) + (sinHAngleToXOverAy * sinHAngleToXOverAy) ) ));
	}

	/// make direction vector
	{
		float cosHAngleToNormal;
		float sinHAngleToNormal;
		hxa7241::graphics::sinAndcos( hAngleToNormal, sinHAngleToNormal, cosHAngleToNormal );

		const Vector3f tangentDirection( (getTangentX().getVector() * cosHAngleToX) += (getTangentY().getVector() * sinHAngleToX) );
		const Vector3f h( (tangentDirection * sinHAngleToNormal) += (getNormal().getVector() * cosHAngleToNormal) );

		Vector3f d( h );
		d *= h.dot( inDirection.getVector() ) * 2.0f;
		d -= inDirection.getVector();

		outDirection = d;
	}

	/// clamp direction to hemisphere
	{
		const float nDotOut = getNormal().dot( outDirection );
		if( nDotOut < 0.0f )
		{
			outDirection = outDirection.getVector() - (getNormal().getVector() * nDotOut);
			if( outDirection.isZero() )
			{
				outDirection = getTangentX();
			}
		}
	}

	scaling = colorSpecular_m;
}








/*void BssdfWard::brdfParts
(
	const UnitVector3f& inDirection,
	const UnitVector3f& outDirection,
	Light&              diffuse,
	Light&              specular
) const
{
	const float LdotN = getNormal().dot( inDirection );
	const float VdotN = getNormal().dot( outDirection );

	const bool isNotTransmission = (LdotN >= 0.0f) & (VdotN >= 0.0f);
	if( isNotTransmission )
	{
		/// equation (4) (5b) common pieces

		/// specular/directional
		{
			const UnitVector3f h( inDirection.getVector() + outDirection.getVector() );
			const float HdotN       = h.dot( getNormal() );
			const float LdotNxVdotN = LdotN * VdotN;

			if( (HdotN == 0.0f) | (LdotNxVdotN == 0.0f) )
			{
				/// grazing vectors
				if( h.isZero() )
				{
					/// straight across
					specular = Light::ONE();
				}
				else
				{
					/// different tangent directions
					specular = Light::ZERO();
				}
			}
			else
			{
				float tanExpr;
				if( roughnessX_m == roughnessY_m )
				{
					/// equation (4) particular piece: inside the exp()

					float oneMinusHdotN2 = 1.0f - (HdotN * HdotN);
					oneMinusHdotN2 = oneMinusHdotN2 >= 0.0f ? oneMinusHdotN2 : 0.0f;

					const float tanDelta = float(::sqrt( oneMinusHdotN2 )) / HdotN;
					//const float tanDelta = float(::tan( ::acos( HdotN ) ));

					tanExpr = -(tanDelta * tanDelta) * oneOverAlpha2_m;
				}
				else
				{
					/// equation (5b) particular piece: inside the exp()

					const float HdotXoverAx = xTangentOverAx_m.dot( h.getVector() );
					const float HdotYoverAy = yTangentOverAy_m.dot( h.getVector() );

					tanExpr = -2.0f * (((HdotXoverAx * HdotXoverAx) + (HdotYoverAy * HdotYoverAy)) / (1.0f + HdotN));
				}

				specular  = rsOver4PiAlpha2_m;
				specular *= float(::exp( tanExpr )) / float(::sqrt( LdotNxVdotN ));
			}
		}

		/// diffuse
		{
			diffuse = rdOverPi_m;
		}
	}
	else
	{
		diffuse  = Light::ZERO();
		specular = Light::ZERO();
	}
}


void BssdfWard::sdfGeneral
(
	const UnitVector3f& inDirection,
	const Vector2f01&   stochastic,
	const Float01       roulette,
	const bool          isScaleByReflectivity,
	UnitVector3f&       outDirection,
	Light&              scaling
) const
{
	if( (getNormal().dot( inDirection ) >= 0.0f) &
	    ((roulette < rMean_m) | isScaleByReflectivity) )
	{
		const float rouletteScaled = isScaleByReflectivity ? (roulette * rMean_m) : roulette;

		/// diffuse
		if( rouletteScaled < rdMean_m )
		{
			getUniformSampleDirection( stochastic, outDirection );
			scaling = colorDiffuse_m;
		}
		/// specular
		else
		{
			/// angle around normal
			float hAngleToTangentX;
			{
				if( roughnessX_m == roughnessY_m )
				{
					/// equation (7b) isotropic
					hAngleToTangentX = TWO_PI * stochastic.getY();
				}
				else
				{
					/// equation (7b) anisotropic (keeping the angle in its starting quadrant)
					hAngleToTangentX = float(::atan( (roughnessY_m / roughnessX_m) * float(::tan( TWO_PI * (stochastic.getY() - 0.25f) )) ));
					if( stochastic.getY() >= 0.5 )
					{
						hAngleToTangentX += PI;
					}
				}
			}

			/// angle to normal
			float cosHAngleToX;
			float sinHAngleToX;
			hxa7241::graphics::sinAndcos( hAngleToTangentX, sinHAngleToX, cosHAngleToX );
			float hAngleToNormal;
			{
				const float cosHAngleToXOverAx = cosHAngleToX / roughnessX_m;
				const float sinHAngleToXOverAy = sinHAngleToX / roughnessY_m;

				/// equation (7a)
				hAngleToNormal = float(::sqrt( -float(::log( stochastic.getX() )) /
								 ( (cosHAngleToXOverAx * cosHAngleToXOverAx) + (sinHAngleToXOverAy * sinHAngleToXOverAy) ) ));
			}

			/// make direction vector
			{
				float cosHAngleToNormal;
				float sinHAngleToNormal;
				hxa7241::graphics::sinAndcos( hAngleToNormal, sinHAngleToNormal, cosHAngleToNormal );

				const Vector3f tangentDirection( (getTangentX().getVector() * cosHAngleToX) += (getTangentY().getVector() * sinHAngleToX) );
				const Vector3f h( (tangentDirection * sinHAngleToNormal) += (getNormal().getVector() * cosHAngleToNormal) );

				Vector3f d( h );
				d *= h.dot( inDirection.getVector() ) * 2.0f;
				d -= inDirection.getVector();

				outDirection = d;
			}

			scaling = colorSpecular_m;
		}

		/// clamp direction to hemisphere
		{
			const float nDotOut = getNormal().dot( outDirection );
			if( nDotOut < 0.0f )
			{
				outDirection = outDirection.getVector() - (getNormal().getVector() * nDotOut);
				if( outDirection.isZero() )
				{
					outDirection = getTangentX();
				}
			}
		}

		if( isScaleByReflectivity )
		{
			scaling *= rMean_m;
		}
	}
	else
	{
		/// no transmission, no importance sample
		outDirection = UnitVector3f::ZERO();
		scaling      = Light::ZERO();
	}
}*/
