/*--------------------------------------------------------------------

   Perceptuum renderer, version 2.1
   Copyright (c) 2002-2003,  Harrison Ainsworth.

   www.hxa7241.org

--------------------------------------------------------------------*/

/*--------------------------------------------------------------------

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston,
   MA 02111-1307, USA

--------------------------------------------------------------------*/


#include "Ray.h"
#include "Float01.h"
#include "Float01c.h"
#include <iostream>

#include "EmitterPathNode.h"   /// own header is included last


using namespace perceptuum2::interaction;




/// standard object services -----------------------------------------------------------------------
EmitterPathNode::EmitterPathNode()
 :	SurfacePoint       (),
	isRootNode_m       ( false ),
	rootEmittance_m    (),
	inwardLightStream_m(),
	isInwardDiffuse_m  ( true )
{
}


EmitterPathNode::EmitterPathNode
(
	const Intersection&       intersection,
	const ObjectRenderableId& outsideObject,
	const Light&              emittance
)
 :	SurfacePoint       ( intersection, outsideObject ),
	isRootNode_m       ( true ),
	rootEmittance_m    ( emittance ),
	inwardLightStream_m(),
	isInwardDiffuse_m  ( true )
{
}


EmitterPathNode::EmitterPathNode
(
	const Intersection&       intersection,
	const ObjectRenderableId& outsideObject,
	const LightStream&        inward
)
 :	SurfacePoint       ( intersection, outsideObject ),
	isRootNode_m       ( false ),
	rootEmittance_m    (),
	inwardLightStream_m( inward ),
	isInwardDiffuse_m  ( true )
{
}


EmitterPathNode::~EmitterPathNode()
{
}


EmitterPathNode::EmitterPathNode( const EmitterPathNode& other )
 :	SurfacePoint       ( other ),
	isRootNode_m       ( other.isRootNode_m ),
	rootEmittance_m    ( other.rootEmittance_m ),
	inwardLightStream_m( other.inwardLightStream_m ),
	isInwardDiffuse_m  ( other.isInwardDiffuse_m )
{
}


EmitterPathNode& EmitterPathNode::operator=( const EmitterPathNode& other )
{
	assign( other );

	return *this;
}




/// streaming --------------------------------------------------------------------------------------
std::ostream& perceptuum2::interaction::operator<<( std::ostream& out, const EmitterPathNode& obj )
{
	return out << "{" << *static_cast<const SurfacePoint*>(&obj) << ", isrt:" <<
	              obj.isRootNode_m << ", rtemi:" << obj.rootEmittance_m << ", in:" << obj.inwardLightStream_m << ", isindf:" << obj.isInwardDiffuse_m << "}";
}




/// commands ---------------------------------------------------------------------------------------
void EmitterPathNode::setSurfacePoint
(
	const Intersection&       intersection,
	const ObjectRenderableId& outsideObject
)
{
	SurfacePoint::setSurfacePoint( intersection, outsideObject );

	isRootNode_m        = true;
	rootEmittance_m     = Light::ZERO();
	inwardLightStream_m = LightStream();
	isInwardDiffuse_m   = true;
}


void EmitterPathNode::setEmitterPathNode
(
	const Intersection&       intersection,
	const ObjectRenderableId& outsideObject,
	const Light&              emittance
)
{
 	SurfacePoint::setSurfacePoint( intersection, outsideObject );

	isRootNode_m        = true;
	rootEmittance_m     = emittance;
	inwardLightStream_m = LightStream();
	isInwardDiffuse_m   = true;
}


void EmitterPathNode::setEmitterPathNode
(
	const Intersection&       intersection,
	const ObjectRenderableId& outsideObject,
	const LightStream&        inward,
	const bool                isDiffuse
)
{
 	SurfacePoint::setSurfacePoint( intersection, outsideObject );

	isRootNode_m        = false;
	rootEmittance_m     = Light::ZERO();
	inwardLightStream_m = inward;
	isInwardDiffuse_m   = isDiffuse;
}




/// queries ----------------------------------------------------------------------------------------
void EmitterPathNode::getNextPathStep
(
	const bool    isRoulette,
	const Float01 stochastics[7],
	Ray&          ray,
	Light&        light,
	bool&         isDiffuse
) const
{
	if( isRootNode_m )
	{
		/// never executed
		/// functionality in SurfacePoint::getEmittedRay
		/// maybe should be refactored into here...
	}
	else
	{
		/// next step light = stochastically(inward light * reflectivity * sdf part * cos out)

		Light scaling;
		dword status;
		getPathStep( ((~(dword(isRoulette)-1) & IS_ROULETTE) | IS_DIFFUSE_STEP_ALLOWED), stochastics, inwardLightStream_m.direction_m,
		             ray, scaling, status );

		(light = inwardLightStream_m.light_m) *= scaling;
		isDiffuse = bool(status & IS_DIFFUSE);
	}
}


void EmitterPathNode::getSpreadLight
(
	const Vector3f&     toPosition,
	const UnitVector3f& outDirection,
	const Float01c      shinynessPoint,
	Light&              outLight
) const
{
	if( isRootNode_m )
	{
		/// spread light = (emittance * edf diffuse parts * cos outs / distance^2)

		getEmittedDiffuseLight( rootEmittance_m, outDirection, outLight );
	}
	else
	{
		/// spread light = ((inward light * brdf diffuse parts * cos outs * reflectivity) + (sub threshold emitted)) / distance^2

		getSpreadLight( inwardLightStream_m.direction_m, outDirection, inwardLightStream_m.light_m, shinynessPoint, outLight );

		/// add emitted light
		//outLight +=
	}

	outLight /= getDistance2( toPosition );
}


const LightStream& EmitterPathNode::getInwardLightstream() const
{
	return inwardLightStream_m;
}


bool EmitterPathNode::isInwardDiffuse() const
{
	return isInwardDiffuse_m;
}




/// implementation ---------------------------------------------------------------------------------
void EmitterPathNode::getSpreadLight
(
	const UnitVector3f& inDirection,
	const UnitVector3f& outDirection,
	const Light&        inLight,
	const Float01c      shinynessPoint,
	Light&              outLight
) const
{
	UnitVector3f inDirClamped( inDirection );
	clampNearToFakeSurface( inDirClamped );

	UnitVector3f outDirClamped( outDirection );
	clampNearToFakeSurface( outDirClamped );


	const Brdf& brdf = getBrdf();

	float   proportions[brdf.PARTITIONS_MAX];
	Float01 peakness[brdf.PARTITIONS_MAX];
	dword   partsCount;
	brdf.getPartitions( proportions, peakness, partsCount );

	/// sum diffuse parts of brdf
	Light diffuseScaling;
	for( dword part = partsCount;  part-- > 0; )
	{
		if( peakness[part] < shinynessPoint )
		{
			Light scaling;
			brdf.getScaling( part, inDirClamped, outDirClamped, scaling );
			const Float01 cos = brdf.getCos( part, outDirClamped );

			diffuseScaling += (scaling *= (cos * proportions[part]));
		}
	}

	const Float01 reflectivity = brdf.getReflectivity();
	diffuseScaling *= reflectivity;

	/// scale inward light by parts of brdf
	(outLight = inLight) *= diffuseScaling;
}


void EmitterPathNode::assign
(
	const SurfacePoint& other
)
{
	if( &other != this )
	{
		SurfacePoint::assign( other );

		isRootNode_m        = false;
		rootEmittance_m     = Light::ZERO();
		inwardLightStream_m = LightStream();
		isInwardDiffuse_m   = true;
	}
}


void EmitterPathNode::assign
(
	const EmitterPathNode& other
)
{
	if( &other != this )
	{
		SurfacePoint::assign( other );

		isRootNode_m        = other.isRootNode_m;
		rootEmittance_m     = other.rootEmittance_m;
		inwardLightStream_m = other.inwardLightStream_m;
		isInwardDiffuse_m   = other.isInwardDiffuse_m;
	}
}
