/*--------------------------------------------------------------------

   Perceptuum renderer, version 2.1
   Copyright (c) 2002-2003,  Harrison Ainsworth.

   www.hxa7241.org

--------------------------------------------------------------------*/

/*--------------------------------------------------------------------

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston,
   MA 02111-1307, USA

--------------------------------------------------------------------*/


#ifndef SurfacePoint_h
#define SurfacePoint_h


#include "Primitives.h"
#include "BssdfWard.h"
//#include "BssdfPhong.h"
#include "Light.h"
#include "LightStream.h"
#include "ObjectRenderableId.h"
#include "Vector3f.h"
#include "UnitVector3f.h"
#include <vector>
#include <iosfwd>

#include "hxa7241_graphics.h"
#include "perceptuum2_modelling.h"




#include "perceptuum2_interaction.h"
namespace perceptuum2 { namespace interaction {
	using namespace hxa7241::primitives;
	using perceptuum2::modelling::ObjectRenderableId;
	using perceptuum2::modelling::Intersection;
	using perceptuum2::modelling::Ray;
	using perceptuum2::modelling::Qualities;
	using hxa7241::graphics::Vector3f;
	using hxa7241::graphics::UnitVector3f;
	using hxa7241::utility::Float01;
	using hxa7241::utility::Histogram;
	using std::vector;


class SurfacePoint
{
/// standard object services -----------------------------------------------------------------------
protected:
	         SurfacePoint();
	         SurfacePoint( const Intersection&       intersection,
	                       const ObjectRenderableId& outsideObject );
public:
	virtual ~SurfacePoint();
protected:
	         SurfacePoint( const SurfacePoint& );
	SurfacePoint& operator=( const SurfacePoint& );

	friend  std::ostream& operator<<( std::ostream&, const SurfacePoint& );


/// commands ---------------------------------------------------------------------------------------
public:
	virtual void  setSurfacePoint( const Intersection&       intersection,
	                               const ObjectRenderableId& outsideObject );

protected:
	virtual void  setQualities( const ObjectRenderableId& object,
	                            const ObjectRenderableId& outsideObject,
	                            const Vector3f&           position,
	                            const Qualities&          qualities );


/// queries ----------------------------------------------------------------------------------------
public:
	virtual const Vector3f&           getPosition()                                           const;
	virtual const ObjectRenderableId& getObject()                                             const;
	virtual const ObjectRenderableId& getObject( const Vector3f& direction )                  const;

	virtual void  getRay( const Vector3f& toPosition,
	                      Ray&            ray )                                               const;
	virtual void  getRay( const UnitVector3f& direction,
	                      bool                isTransmitted,
	                      Ray&                ray )                                           const;

	/** @return is transmitted */
	static  bool  getEmittedRay( const Qualities&    qualities,
	                             const Vector2f01&   stochastic,
	                             const UnitVector3f& normalUnmodified,
	                             bool                isCosStochastic,
	                             UnitVector3f&       direction,
	                             Light&              light,
	                             bool&               isDiffuse );
//	virtual void  makeImportanceHistogram( const vector<LightStream>& illuminations,
//	                                       const UnitVector3f&        inDirection,
//	                                       vector<float>&             gridBuffer,
//	                                       Histogram&                 importance )            const;
//	              /** @return isTransmitted */
//	virtual bool  getSamplingDirection( const Histogram&    importance,
//	                                    const Float01       stochastics[3],
//	                                    const UnitVector3f& inDirection,
//	                                    UnitVector3f&       direction,
//	                                    Light&              scaling )                         const;

	/// test
	static  bool  test( bool isLogging );


/// implementation ---------------------------------------------------------------------------------
protected:
	enum EConditions { IS_ROULETTE =1, IS_DIFFUSE_STEP_ALLOWED =2 };
	enum EStatus { IS_DIFFUSE =65536, IS_DIRECT_SEPARATE =131072 };

	virtual void  getPathStep( dword               conditions,
	                           const Float01       stochastics[7],
	                           const UnitVector3f& inDirection,
	                           Ray&                ray,
	                           Light&              scaling,
	                           dword&              status )                                   const;
	virtual void  getPathStepScaling( const UnitVector3f& inDirection,
	                                  const UnitVector3f& outDirection,
	                                  Light&              scaling )                           const;
	virtual void  getEmittedLight( const UnitVector3f& outDirection,
	                               Light&              emitted )                              const;
	virtual void  getEmittedDiffuseLight( const Light&        emittance,
	                                      const UnitVector3f& outDirection,
	                                      Light&              outLight )                      const;

	virtual const Brdf& getBrdf()                                                             const;
	virtual float getDistance2( const Vector3f& position )                                    const;

	static  void  makeCoordFrame( const UnitVector3f& normal,
	                              const UnitVector3f& tangent,
	                              UnitVector3f&       normalZ,
	                              UnitVector3f&       tangentX,
	                              UnitVector3f&       tangentY );

	virtual bool  isFalseCrossing( const UnitVector3f& )                                      const;
	virtual bool  clampToFakeSurface( UnitVector3f& )                                         const;
	virtual bool  clampToTrueSurface( UnitVector3f& )                                         const;
	virtual bool  clampNearToFakeSurface( UnitVector3f& )                                     const;
	static  bool  clampToSurface( const UnitVector3f& normalTrue,
	                              const UnitVector3f& normalFake,
	                              const UnitVector3f& tangentTrue,
	                              UnitVector3f&       direction );

	virtual void  assign( const SurfacePoint& );


/// fields -----------------------------------------------------------------------------------------
private:
	ObjectRenderableId object_m;
	ObjectRenderableId outsideObject_m;
	Vector3f           position_m;
	UnitVector3f       normalUnmodified_m;
	Light              emitted_m;
	//BssdfPhong         brdf_m;
	BssdfWard          brdf_m;
	//BssdfPhong         emittedDistribution_m;
};




std::ostream& operator<<( std::ostream&, const SurfacePoint& );


}}




#endif//SurfacePoint_h
