/*--------------------------------------------------------------------

   Perceptuum renderer, version 2.1
   Copyright (c) 2002-2003,  Harrison Ainsworth.

   www.hxa7241.org

--------------------------------------------------------------------*/

/*--------------------------------------------------------------------

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston,
   MA 02111-1307, USA

--------------------------------------------------------------------*/


#ifndef Material_h
#define Material_h


#include "Primitives.h"
#include "Qualities.h"
#include "ImageConverter.h"
#include "ImageRgba.h"
#include "Sheet.h"
#include "Vector2.h"
#include <string>
#include <vector>

#include "hxa7241_utility.h"
#include "hxa7241_graphics.h"




#include "perceptuum2_modelling.h"
namespace perceptuum2 { namespace modelling {
	using namespace hxa7241::primitives;
	using perceptuum2::imaging::ImageRgba;
	using perceptuum2::imaging::ImageConverter;
	using hxa7241::graphics::Vector2;
	using hxa7241::graphics::Vector2f;
	using hxa7241::graphics::Vector2i;
	using hxa7241::graphics::Vector3f;
	using hxa7241::graphics::UnitVector3f;
	using hxa7241::graphics::CieXyz;
	using hxa7241::utility::Sheet;
	using hxa7241::utility::Float01;
	using hxa7241::utility::SerializerOut;
	using hxa7241::utility::SerializerIn;
	using std::string;
	using std::vector;


class Material
{
/// standard object services -----------------------------------------------------------------------
public:
	         Material();

	virtual ~Material();
	         Material( const Material& );
	Material& operator=( const Material& );


/// commands ---------------------------------------------------------------------------------------
public:
	virtual void  setImagePath( const string& );


/// queries ----------------------------------------------------------------------------------------
public:
	/// coords should be positive
	virtual void  getQualities( const Vector2f& barycentrics, const Vector2<word> cornerCoords[3],
	                            Qualities& )                                                  const;
	virtual void  getQualities( const Vector2f& coords, Qualities& )                          const;
	virtual const Qualities& getQualitiesBase()                                               const;

	virtual Float01 getTransparency( const Vector3f& position )                               const;
	virtual float   getRefractiveIndex( const Vector3f& position )                            const;

	virtual void  getEmittedMean( const Vector2<word> coords[3],      CieXyz& )               const;
	virtual void  getSpecularnessMean( const Vector2<word> coords[3], Float01& )              const;

	virtual void  getEmittedPosition( const Vector2<word> coords[3], const Vector3f& posRands,
	                                  vector<float>& storage, Vector2f& barycentrics )        const;


/// serialization ----------------------------------------------------------------------------------
public:
	friend  SerializerOut& operator<<( SerializerOut&, const Material& );
	friend  SerializerIn&  input     ( SerializerIn&,        Material&, Scene& );
	static const char* CLASS_NAME();
	static const char* CLASS_DESCRIPTION();


/// implementation ---------------------------------------------------------------------------------
protected:
	virtual void  getReflectivity( const Vector2f& coords, CieXyz&  rDiffuse,   CieXyz&  rSpecular )  const;
	virtual void  getRoughness   ( const Vector2f& coords, Float01& roughnessX, Float01& roughnessY ) const;
	virtual void  getTransparency( const Vector2f& coords, Float01&      transparency )               const;
	virtual void  getEmitted     ( const Vector2f& coords, CieXyz&       emitted )                    const;
	virtual void  getTangent     ( const Vector2f& coords, UnitVector3f& tangent )                    const;
	virtual void  getNormal      ( const Vector2f& coords, UnitVector3f& normal )                     const;
	virtual void  getSpecularness( const Vector2f& coords, Float01& specularness )                    const;

	static  bool  getTexelColor( const Vector2f& coords, const string& imagePath, const string& imageName,
	                             ImageRgba& image, ImageConverter&, bool& isLoadFailed, CieXyz& color );
	static  bool  getTexelVector( const Vector2f& coords, const string& imagePath, const string& imageName,
	                              Sheet<byte>& image, bool& isLoadFailed, UnitVector3f& vector );
	static  bool  getTexel( const Vector2f& coords, const string& imagePath, const string& imageName,
	                        Sheet<byte>& image, bool& isLoadFailed, Float01& texel );

	static  bool  loadImage( const string& imagePath, const string& imageName,
	                         Sheet<byte>* pImageByte, ImageRgba* pImageRgba, ImageConverter* pConverter,
	                         bool& isLoadFailed );
	static  float readElementInterpolated( const Vector2f& coords, const Sheet<byte>& image );
	static  void  makeInterpolantPositions( const Vector2f& coords, const Vector2f& imageSize,
	                                        dword& smallerOffset, Vector2i& increment, Vector2f& subPosition );
	virtual void  resetImages();
	virtual void  resetImageFlags();


/// fields -----------------------------------------------------------------------------------------
private:
	string name_m;

	Qualities baseQualities_m;

	string imagePath_m;

	string colorImageName_m;
	string colorSpecImageName_m;
	string specDiffImageName_m;
	string roughnessXImageName_m;
	string roughnessYImageName_m;
	string transparencyImageName_m;
	string emittedImageName_m;
	string tangentImageName_m;
	string normalImageName_m;

	mutable ImageConverter colorConverter_m;

	mutable ImageRgba   color_m;
	mutable ImageRgba   colorSpec_m;
	mutable Sheet<byte> specDiff_m;
	mutable Sheet<byte> roughnessX_m;
	mutable Sheet<byte> roughnessY_m;
	mutable Sheet<byte> transparency_m;
	mutable Sheet<byte> emitted_m;
	mutable Sheet<byte> tangent_m;
	mutable Sheet<byte> normal_m;

	mutable bool isColorLoadFailed_m;
	mutable bool isColorSpecLoadFailed_m;
	mutable bool isSpecDiffLoadFailed_m;
	mutable bool isRoughnessXLoadFailed_m;
	mutable bool isRoughnessYLoadFailed_m;
	mutable bool isTransparencyLoadFailed_m;
	mutable bool isEmittedLoadFailed_m;
	mutable bool isTangentLoadFailed_m;
	mutable bool isNormalLoadFailed_m;

	static const char LOAD_FAIL_MESSAGE[];
};




/// serialization ----------------------------------------------------------------------------------
SerializerOut& operator<<( SerializerOut&, const Material& );
SerializerIn&  operator>>( SerializerIn&,        Material& );


}}




#endif//Material_h
