/*--------------------------------------------------------------------

   Perceptuum renderer, version 2.1
   Copyright (c) 2002-2003,  Harrison Ainsworth.

   www.hxa7241.org

--------------------------------------------------------------------*/

/*--------------------------------------------------------------------

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston,
   MA 02111-1307, USA

--------------------------------------------------------------------*/


#include "Ray.h"
#include "Intersection.h"
#include "Scene.h"
#include "ProjectableObjectRef.h"
#include "Vertex.h"
#include "TriangleImplicit.h"
#include "ClipFlags.h"
#include "Histogram.h"
#include "Serialization.h"
#include "Logger.h"

#include "ObjectInstance.h"   /// own header is included last


using namespace perceptuum2::modelling;




/// statics ----------------------------------------------------------------------------------------
const char ObjectInstance::DESERIALIZATION_WARNING_1[] = "object nesting too deep";




/// standard object services -----------------------------------------------------------------------
ObjectInstance::ObjectInstance
(
	const Scene*  pScene,
	SerializerIn& in,
	const bool    b
)
 :	ObjectRenderable    ( pScene ),
	//subParts_m          (),
	previousUpdateTime_m( -0.0f ),
	previousIsChanging_m( false )
{
	if( b )
	{
		in.read( "object", *this );
	}
	else
	{
		in >> *this;
	}
}


ObjectInstance::ObjectInstance
(
	const Scene* pScene
)
 :	ObjectRenderable    ( pScene ),
	//subParts_m          (),
	previousUpdateTime_m( -0.0f ),
	previousIsChanging_m( false )
{
}


ObjectInstance::~ObjectInstance()
{
}


ObjectInstance::ObjectInstance( const ObjectInstance& other )
 :	ObjectRenderable    ( other ),
	subParts_m          ( other.subParts_m ),
	previousUpdateTime_m( other.previousUpdateTime_m ),
	previousIsChanging_m( other.previousIsChanging_m )
{
}


ObjectInstance& ObjectInstance::operator=( const ObjectInstance& other )
{
	assign( other );

	return *this;
}




/// commands ---------------------------------------------------------------------------------------
bool ObjectInstance::update
(
	const float time
)
{
	bool isChanging = false;

	if( time != previousUpdateTime_m )
	{
		/// step thru subparts
		{
			const Scene* pScene = getScene();

			deque<SubPart>::iterator itSubpart( subParts().begin() );
			deque<SubPart>::iterator itEnd( subParts().end() );
			for( ;  itSubpart != itEnd;  ++itSubpart )
			{
				ObjectRenderable*const pObject = pScene->getObjectNonConst(itSubpart->objectIndex_m);
				if( pObject != 0 )
				{
					/// update subpart, and note if its changing
					isChanging |= pObject->update( time );
				}
			}
		}

		/// reset base if any subparts changing
		if( isChanging )
		{
			ObjectInstance::setBase();
		}

		ObjectInstance::setUpdateNote( time, isChanging );
	}
	else
	{
		isChanging = previousIsChanging_m;
	}

	return isChanging;
}




/// queries ----------------------------------------------------------------------------------------
bool ObjectInstance::getEmitted
(
	const Vector3f& posRands,
	vector<float>&  compositionStorage,
	Intersection&   intersection,
	Qualities&      qualities
) const
{
	/// choose sub object with roulette and emittance composition
	dword subPartIndex     = 0;
	float rouletteAdjusted = 0.0f;
	bool  isAnyEmitted;//     = false;
	{
		const dword subPartsSize = subParts_m.size();
		hxa7241::utility::Histogram composition( compositionStorage, subPartsSize );

		const deque<SubPart>::const_iterator end( subParts_m.end() );
		for( deque<SubPart>::const_iterator it( subParts_m.begin() );  it != end;  ++it )
		{
			const float emittance = getScene()->getObject(it->objectIndex_m)->getEmittanceSum().average();
			composition.addElement( emittance );
		}

		isAnyEmitted = composition.findElement( posRands.getZ(), subPartIndex, rouletteAdjusted );
	}


	if( isAnyEmitted )
	{
		/// push objectid
		intersection.object_m.push( byte(subPartIndex) );

		const SubPart& subPart = subParts_m[subPartIndex];

		/// recurse to sub object
		{
			const ObjectRenderable* const pObject = getScene()->getObject(subPart.objectIndex_m);
			const Vector3f                posRandsSubPart( posRands.getX(), posRands.getY(), rouletteAdjusted );
			isAnyEmitted = pObject->getEmitted( posRandsSubPart, compositionStorage, intersection, qualities );
		}

		if( isAnyEmitted )
		{
			/// outward transform intersection and qualities
			const Transform& transform = subPart.transform_m;
			intersection.transformEq( transform );
			qualities.transformEq( transform );
		}
	}

	return isAnyEmitted;
}


bool ObjectInstance::getIntersection
(
	const ObjectRenderableId& objectId,
	const Ray&                ray,
	float&                    distance,
	Intersection&             intersection
) const
{
	bool anyHits = false;

	/// check bound intersection first, but skip if less than 2 sub parts
	if( (subParts_m.size() <= 1) || getBound().getIntersection( ray ) )
	{
		float          nearestDistance = FLOAT_MAX;
		Intersection   nearestIntersection;
		const SubPart* pSubPart = 0;

		/// non-indexed search, by comparing all subobject intersection distances to find minimum
		const deque<SubPart>::const_iterator begin( subParts_m.begin() );
		const deque<SubPart>::const_iterator end( subParts_m.end() );
		for( deque<SubPart>::const_iterator it( begin );  it != end;  ++it )
		{
			const SubPart& subPart = *it;

			/// inward transform ray
			Ray rayTransformed( ray );
			rayTransformed.transformEq( Transplace( subPart.transform_m, true ) );

			/// make object id of sub part
			ObjectRenderableId objectIdNext( objectId );
			const bool overflow = objectIdNext.push( byte(it - begin) );
			if( !overflow )
			{
				/// intersect sub part
				float        distanceTrial = 0.0f;
				Intersection intersectionTrial;
				const ObjectRenderable* const pObject = getScene()->getObject( subPart.objectIndex_m );
				const bool isHit = (pObject != 0) && pObject->getIntersection( objectIdNext, rayTransformed, distanceTrial, intersectionTrial );

				subPart.transform_m.transform( distanceTrial );

				/// keep nearest hit data
				anyHits |= isHit;
				if( isHit & (distanceTrial < nearestDistance) )
				{
					nearestDistance     = distanceTrial;
					nearestIntersection = intersectionTrial;
					pSubPart            = &subPart;
				}
			}
		}

		/// set intersection output params
		if( anyHits )
		{
			distance     = nearestDistance;
			intersection = nearestIntersection;

			/// outward transform intersection
			const Transform& transform = pSubPart->transform_m;
			intersection.transformEq( transform );
		}
	}

	return anyHits;
}


void ObjectInstance::getObjectsForProjection
(
	const Transclip&              transform,
	const ClipFlags               enclosingClipFlags,
	vector<ProjectableObjectRef>& objects
) const
{
	const dword subPartCount = subParts_m.size();
	/// general case
	if( subPartCount > 1 )
	{
		/// if wholly or partially inside
		ClipFlags nextClipflags;
		if( getBound().getClipping( enclosingClipFlags, transform, nextClipflags ) )
		{
			/// ordering (hmmm, probably leave it and wait for octree indexing)
			/// 	step through sub parts
			/// 		get bound center point, transform it to view space
			/// 	sort transformed bound centers plus radiuss in z

			/// step through sub parts
			const deque<SubPart>::const_iterator begin( subParts_m.begin() );
			const deque<SubPart>::const_iterator end( subParts_m.end() );
			for( deque<SubPart>::const_iterator it( begin );  it != end;  ++it )
			{
				const SubPart& subPart = *it;

				/// concatenate transform param with sub part transform
				Transclip nextTransform( transform );
				nextTransform += subPart.transform_m.getMatrix();

				/// call sub part getObjectsForProjection with concatenated transform and new clipflags
				const ObjectRenderable* const pObject = getScene()->getObject( subPart.objectIndex_m );

				pObject->getObjectsForProjection( nextTransform, nextClipflags, objects );
			}
		}
	}
	/// accelerated special case
	else if( subPartCount == 1 )
	{
		const SubPart& subPart = subParts_m.front();

		/// concatenate transform param with sub part transform
		Transclip nextTransform( transform );
		nextTransform += subPart.transform_m.getMatrix();

		/// call sub part getObjectsForProjection with concatenated transform
		const ObjectRenderable* const pObject = getScene()->getObject( subPart.objectIndex_m );

		pObject->getObjectsForProjection( nextTransform, enclosingClipFlags, objects );
	}
}


void ObjectInstance::getMesh
(
	Mesh&        ,//meshBuffer,
	const Mesh** ppMesh
) const
{
	*ppMesh = 0;


	/// needs to copy all sub parts items into one collection...

//	const deque<SubPart>::const_iterator begin( subParts_m.begin() );
//	const deque<SubPart>::const_iterator end( subParts_m.end() );
//	for( deque<SubPart>::const_iterator it( begin );  it != end;  ++it )
//	{
//		const SubPart& subPart = *it;
//
//	}
}


/*void ObjectInstance::getQualities
(
	const ObjectRenderableId& objectId,
	const Vector3f&           position,
	Qualities&                qualities
) const
{
	ObjectRenderableId subObjectId;
	const Transplace*  pTransform = 0;
	getSubObject( objectId, subObjectId, pTransform );

	if( subObjectId.isValid() )
	{
		/// inward transform position
		Vector3f positionTransformed( position );
		Transplace( *pTransform, true ).transform( positionTransformed );

		subObjectId->getQualities( subObjectId, positionTransformed, qualities );

		/// outward transform qualities
		qualities.transformEq( *pTransform );
	}
	else
	{
		qualities.setToDefault();
	}
}*/


void ObjectInstance::getQualities
(
	const Intersection& intersection,
	Qualities&          qualities
) const
{
	ObjectRenderableId subObjectId;
	const Transplace*  pTransform = 0;
	getSubObject( intersection.object_m, subObjectId, pTransform );

	if( subObjectId.isValid() )
	{
		/// inward transform intersection
		Intersection subIntersection( intersection );
		subIntersection.object_m = subObjectId;
		subIntersection.transformEq( Transplace( *pTransform, true ) );

		subObjectId->getQualities( subIntersection, qualities );

		/// outward transform qualities
		qualities.transformEq( *pTransform );
	}
	else
	{
		qualities.setToDefault();
	}
}


Float01 ObjectInstance::getTransparency
(
	const ObjectRenderableId& objectId,
	const Vector3f&           position
) const
{
	Float01 transparency;

	ObjectRenderableId subObjectId;
	const Transplace*  pTransform = 0;
	getSubObject( objectId, subObjectId, pTransform );

	if( subObjectId.isValid() )
	{
		/// inward transform position
		Vector3f positionTransformed( position );
		Transplace( *pTransform, true ).transform( positionTransformed );

		transparency = subObjectId->getTransparency( subObjectId, positionTransformed );
	}

	return transparency;
}


float ObjectInstance::getRefractiveIndex
(
	const ObjectRenderableId& objectId,
	const Vector3f&           position
) const
{
	float refractiveIndex = 1.0f;

	ObjectRenderableId subObjectId;
	const Transplace*  pTransform = 0;
	getSubObject( objectId, subObjectId, pTransform );

	if( subObjectId.isValid() )
	{
		/// inward transform position
		Vector3f positionTransformed( position );
		Transplace( *pTransform, true ).transform( positionTransformed );

		refractiveIndex = subObjectId->getRefractiveIndex( subObjectId, positionTransformed );
	}

	return refractiveIndex;
}


const CieXyz& ObjectInstance::getEmittance
(
	const ObjectRenderableId& objectId
) const
{
	ObjectRenderableId subObjectId;
	const Transplace*  pTransform = 0;
	getSubObject( objectId, subObjectId, pTransform );

	if( subObjectId.isValid() )
	{
		return subObjectId->getEmittance( subObjectId );
	}
	else
	{
		return CieXyz::ZERO();
	}
}


const char* ObjectInstance::getClassName() const
{
	return ObjectInstance::CLASS_NAME();
}




/// implementation ---------------------------------------------------------------------------------
void ObjectInstance::assign
(
	const ObjectRenderable& other
)
{
	if( &other != this )
	{
		ObjectRenderable::assign( other );

		subParts_m.resize( 0 );
		previousUpdateTime_m = -0.0f;
		previousIsChanging_m = false;
	}
}


void ObjectInstance::assign
(
	const ObjectInstance& other
)
{
	if( &other != this )
	{
		ObjectRenderable::assign( other );

		subParts_m           = other.subParts_m;
		previousUpdateTime_m = other.previousUpdateTime_m;
		previousIsChanging_m = other.previousIsChanging_m;
	}
}


void ObjectInstance::setBase()
{
	const dword subPartCount = subParts_m.size();

	/// aggregate sub part properties
	CieXyz   emittanceSum;
	float    specularnessSum   = 0.0f;
	dword    hierarchyLevelMax = 0;
	dword    subPartCountSum   = 0;
	Vector3f boundMin( Vector3f::MAX() );
	Vector3f boundMax( Vector3f::MIN() );

	for( deque<SubPart>::const_iterator itSubpart( subParts_m.begin() );  itSubpart != subParts_m.end();  ++itSubpart )
	{
		const ObjectInstance::SubPart& subPart = *itSubpart;
		const ObjectRenderable* const  pObject = ObjectRenderable::getScene()->getObject( subPart.objectIndex_m );

		/// check for valid object ptr
		if( pObject != 0 )
		{
			/// accumulate qualities
			emittanceSum    += pObject->getEmittanceSum();
			specularnessSum += pObject->getSpecularness();

			/// maximize level
			{
				const dword level = pObject->getHierarchyLevel();
				if( hierarchyLevelMax < level  )
				{
					hierarchyLevelMax = level;
				}
			}

			/// accumulate sub part count
			subPartCountSum += pObject->getSubPartCount();

			/// minimise and maximise bounds
			{
				Vector3f corners[8];
				pObject->getBound().getCorners( corners );

				const Transform& transform = subPart.transform_m;

				for( int j = 8;  j-- > 0; )
				{
					Vector3f& corner = corners[j];
					transform.transform( corner );

					boundMin.clampToMaxOf( corner );
					boundMax.clampToMinOf( corner );
				}
			}
		}
	}

	/// set properties
	{
		ObjectRenderable::setEmittanceSum( emittanceSum );

		const float specularnessMean = subPartCount != 0 ? specularnessSum / float(subPartCount) : 0.0f;
		ObjectRenderable::setSpecularness( specularnessMean );

		ObjectRenderable::setHierarchyLevel( hierarchyLevelMax + 1 );
		ObjectRenderable::setSubPartCount( subPartCountSum );

		ObjectRenderable::bound().setMinMax( boundMin, boundMax );
	}
}


deque<ObjectInstance::SubPart>& ObjectInstance::subParts()
{
	return subParts_m;
}


float ObjectInstance::getUpdateTime() const
{
	return previousUpdateTime_m;
}


bool ObjectInstance::getUpdateIsChanging() const
{
	return previousIsChanging_m;
}


void ObjectInstance::setUpdateNote
(
	const float time,
	const bool  isChanging
)
{
	previousUpdateTime_m = time;
	previousIsChanging_m = isChanging;
}


void ObjectInstance::getSubObject
(
	const ObjectRenderableId& objectId,
	ObjectRenderableId&       subObjectId,
	const Transplace*&        pTransform
) const
{
	const ObjectRenderable* pObject  = 0;
	const SubPart*          pSubPart = 0;

	/// check for valid sub part id
	const dword subPartId = dword(ubyte( objectId.getBottom() ));
	if( (subPartId >= 0) & (udword(subPartId) < subParts_m.size()) )
	{
		pSubPart = &(subParts_m[subPartId]);
		pObject = getScene()->getObject( pSubPart->objectIndex_m );
	}

	///
	subObjectId = objectId;
	subObjectId.shift( pObject );

	if( pObject != 0 )
	{
		pTransform = &(pSubPart->transform_m);
	}
	else
	{
		pTransform = 0;
	}
}




/// serialization ----------------------------------------------------------------------------------
template<>
inline
const char* SerializerOut::typeName<std::deque<ObjectInstance::SubPart> >()
{
	return "std::deque";
}


template<>
inline
const char* SerializerOut::typeDescription<std::deque<ObjectInstance::SubPart> >()
{
	return "standard deque collection";
}


SerializerOut& perceptuum2::modelling::operator<<( SerializerOut& out, const ObjectInstance& obj )
{
	/*const dword subPartCount = obj.subParts_m.size();

	out.write( "subPartCount", subPartCount, "0-256" );

	for( int i = 0;  i < subPartCount;  ++i )
	{
		out.write( "subPart", obj.subParts_m[i] );

		const Transplace& trans = obj.transforms_m[i];
		out.write( "scale",    trans.getScale() );
		out.write( "rotation", trans.getRotation() );
		out.write( "position", trans.getPosition() );
	}*/

	out.write( "subParts", obj.subParts_m, *(obj.getScene()), 0 );

	return out;
}


SerializerIn& perceptuum2::modelling::operator>>( SerializerIn& in, ObjectInstance& obj )
{
	/// read object
	{
		/*/// read number of sub parts
		in.read( "subPartCount", subPartCount );
		if( subPartCount < 0 )
		{
			subPartCount = 0;
		}
		else if( subPartCount > ObjectRenderableId::getMaxStackWidth() )
		{
			subPartCount = ObjectRenderableId::getMaxStackWidth();
		}


		/// read in sub parts and transforms
		{
			obj.subParts_m.resize( 0 );
			obj.subParts_m.reserve( subPartCount );
			obj.transforms_m.resize( 0 );
			obj.transforms_m.reserve( subPartCount );

			for( int i = 0;  i < subPartCount;  ++i )
			{
				dword objectIndex = 0;
				in.read( "subPart", objectIndex );

				float    scale;
				Vector3f rotation;
				Vector3f position;
				in.read( "scale",    scale    );
				in.read( "rotation", rotation );
				in.read( "position", position );

				const ObjectRenderable* const pObject = obj.getScene()->getObject( objectIndex );
				/// check object index and object pointer validity
				if( pObject != 0 )
				{
					/// check nesting limit
					if( pObject->getHierarchyLevel() < ObjectRenderableId::getMaxStackSize() )
					{
						obj.subParts_m.push_back( objectIndex );
						obj.transforms_m.push_back( Transplace( scale, rotation, position ) );
					}
					else
					{
						hxa7241::utility::Logger() << hxa7241::utility::Logger::EXC1 << obj.DESERIALIZATION_WARNING_1 << " (" << i << ")";
					}
				}
			}

			subPartCount = obj.subParts_m.size();
		}*/


		/// read sub parts
		obj.subParts_m.resize( 0 );
		in.read( "subParts", obj.subParts_m, *(obj.getScene()) );

		/// check sub parts
		for( deque<ObjectInstance::SubPart>::iterator it = obj.subParts_m.begin();  it != obj.subParts_m.end();  ++it )
		{
			/// check object index and object pointer validity
			const ObjectRenderable* const pObject = obj.getScene()->getObject( it->objectIndex_m );
			if( pObject != 0 )
			{
				/// check nesting limit
				if( !(pObject->getHierarchyLevel() < ObjectRenderableId::getMaxStackSize()) )
				{
					/// remove sub part
					it = obj.subParts_m.erase( it );
					hxa7241::utility::Logger() << hxa7241::utility::Logger::EXC1 << obj.DESERIALIZATION_WARNING_1 << " (" << (it - obj.subParts_m.begin()) << ")";
				}
			}
			else
			{
				/// remove sub part
				it = obj.subParts_m.erase( it );
			}
		}

		if( obj.subParts_m.size() > udword(ObjectRenderableId::getMaxStackWidth()) )
		{
			obj.subParts_m.resize( ObjectRenderableId::getMaxStackWidth() );
		}
	}


	/// set base
	obj.ObjectInstance::setBase();


	return in;
}


const char* ObjectInstance::CLASS_NAME()
{
	static const char CLASS_NAME[] = "perceptuum2::modelling::ObjectInstance";
	return CLASS_NAME;
}


const char* ObjectInstance::CLASS_DESCRIPTION()
{
	static const char CLASS_DESCRIPTION[] = "transplaced instance(s) of renderable object(s)";
	return CLASS_DESCRIPTION;
}








/// serialization ----------------------------------------------------------------------------------
SerializerOut& perceptuum2::modelling::output( SerializerOut& out, const ObjectInstance::SubPart& obj, const Scene& context )
{
	{
		const string objectRef( context.getObjectName( obj.objectIndex_m ) );
		out.write( "objectRef", objectRef );
	}

	out.write( "transform", obj.transform_m );

	return out;
}

SerializerIn& perceptuum2::modelling::input( SerializerIn& in, ObjectInstance::SubPart& obj, const Scene& context )
{
	{
		string objectRef;
		in.read( "objectRef", objectRef );
		obj.objectIndex_m = context.getObjectIndex( objectRef );
	}

	in.read( "transform", obj.transform_m );

	return in;
}


const char* ObjectInstance::SubPart::CLASS_NAME()
{
	static const char CLASS_NAME[] = "perceptuum2::modelling::ObjectInstance::SubPart";
	return CLASS_NAME;
}


const char* ObjectInstance::SubPart::CLASS_DESCRIPTION()
{
	static const char CLASS_DESCRIPTION[] = "internal element from an ObjectInstance";
	return CLASS_DESCRIPTION;
}
