/*--------------------------------------------------------------------

   Perceptuum renderer, version 2.1
   Copyright (c) 2002-2003,  Harrison Ainsworth.

   www.hxa7241.org

--------------------------------------------------------------------*/

/*--------------------------------------------------------------------

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston,
   MA 02111-1307, USA

--------------------------------------------------------------------*/


#ifndef ObjectMesh_h
#define ObjectMesh_h


#include "Primitives.h"
#include "ObjectRenderable.h"
#include "Mesh.h"
#include "CieXyz.h"
#include "Vector3f.h"
#include "UnitVector3f.h"
#include "Vertex.h"
#include "TriangleImplicit.h"
#include <vector>
#include <map>
#include <string>

#include "hxa7241_graphics.h"
#include "hxa7241_utility.h"




#include "perceptuum2_modelling.h"
namespace perceptuum2 { namespace modelling {
	using namespace hxa7241::primitives;
	using hxa7241::graphics::CieXyz;
	using hxa7241::graphics::Vector3f;
	using hxa7241::graphics::UnitVector3f;
	using hxa7241::graphics::Transclip;
	using hxa7241::graphics::ClipFlags;
	using hxa7241::utility::SerializerOut;
	using hxa7241::utility::SerializerIn;
	using std::vector;
	using std::string;


/**
 * a single, non-intersecting, closed surface, made of triangle faces.
 *
 * validates structure when deserializing.
 */
class ObjectMesh
	: public ObjectRenderable
{
/// standard object services -----------------------------------------------------------------------
public:
	         ObjectMesh( const Scene*, SerializerIn&, bool );
protected:
	         ObjectMesh( const Scene* );
public:
	virtual ~ObjectMesh();
	         ObjectMesh( const ObjectMesh& );
	ObjectMesh& operator=( const ObjectMesh& );


/// commands ---------------------------------------------------------------------------------------
public:
	virtual bool  update( float time );


/// queries ----------------------------------------------------------------------------------------
public:
	virtual bool    getEmitted( const Vector3f& posRands,
	                            vector<float>&  compositionStorage,
	                            Intersection&,
	                            Qualities& )                                                  const;

	virtual bool    getIntersection( const ObjectRenderableId&,
	                                 const Ray&,
	                                 float& distance,
	                                 Intersection& )                                          const;
	virtual void    getObjectsForProjection( const Transclip& transform,
	                                         ClipFlags        clipFlags,
	                                         vector<ProjectableObjectRef>& )                  const;
	virtual void    getMesh( Mesh&, const Mesh** )                                            const;

	virtual void    getQualities( const Intersection&, Qualities& )                           const;

	virtual Float01 getTransparency( const ObjectRenderableId&,
	                                 const Vector3f& position )                               const;
	virtual float   getRefractiveIndex( const ObjectRenderableId&,
	                                    const Vector3f& position )                            const;
	virtual const CieXyz& getEmittance( const ObjectRenderableId& )                           const;

	virtual const char* getClassName()                                                        const;


/// serialization ----------------------------------------------------------------------------------
	friend  SerializerOut& operator<<( SerializerOut&, const ObjectMesh& );
	friend  SerializerIn&  operator>>( SerializerIn&,        ObjectMesh& );
	static const char* CLASS_NAME();
	static const char* CLASS_DESCRIPTION();


/// implementation ---------------------------------------------------------------------------------
protected:
	virtual void  setMesh( const vector<Vector3f>&         vertexPositions,
	                       const vector<TriangleImplicit>& triangles,
	                       dword                           materialId );
	virtual void  setMesh();

	virtual void  makeBound( const Vector3f& min,
	                         const Vector3f& max,
	                         Bound& )                                                         const;
	virtual CieXyz  makeEmittance()                                                           const;
	virtual Float01 makeSpecularness()                                                        const;

	virtual void  zeroObject();

	virtual dword  getMaterialId()                                                            const;
	virtual string getMaterialName()                                                          const;
	virtual dword  getMaterialIndex( const string& name )                                     const;

	virtual const vector<TriangleImplicit>& triangles()                                       const;

	virtual bool  isFalseIntersection( const Ray&,
	                                   const ObjectRenderableId&,
	                                   const TriangleImplicit&,
	                                   const TriangleExplicit&,
	                                   const UnitVector3f& normal )                           const;

	virtual void  assign( const ObjectRenderable& );
	virtual void  assign( const ObjectMesh& );


/// fields -----------------------------------------------------------------------------------------
private:
	/// single, non-intersecting, closed surface
	Mesh  mesh_m;
	dword materialId_m;
};




/// serialization ----------------------------------------------------------------------------------
SerializerOut& operator<<( SerializerOut&, const ObjectMesh& );
SerializerIn&  operator>>( SerializerIn&,        ObjectMesh& );


}}




#endif//ObjectMesh_h
