/*--------------------------------------------------------------------

   Perceptuum renderer, version 2.1
   Copyright (c) 2002-2003,  Harrison Ainsworth.

   www.hxa7241.org

--------------------------------------------------------------------*/

/*--------------------------------------------------------------------

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston,
   MA 02111-1307, USA

--------------------------------------------------------------------*/


#include "UnitVector3f.h"
#include "Polar.h"
#include "Serialization.h"

#include "ObjectMeshCylinder.h"   /// own header is included last


using namespace perceptuum2::modelling;




/// standard object services -----------------------------------------------------------------------
ObjectMeshCylinder::ObjectMeshCylinder
(
	const Scene*  pScene,
	SerializerIn& in,
	const bool    b
)
 :	ObjectMesh( pScene ),
	radius_m(),
	height_m(),
	isCurved_m(),
	resolution_m()
{
	if( b )
	{
		in.read( "object", *this );
	}
	else
	{
		in >> *this;
	}
}


ObjectMeshCylinder::ObjectMeshCylinder
(
	const Scene* pScene
)
 :	ObjectMesh( pScene ),
	radius_m(),
	height_m(),
	isCurved_m(),
	resolution_m()
{
}


ObjectMeshCylinder::~ObjectMeshCylinder()
{
}


ObjectMeshCylinder::ObjectMeshCylinder( const ObjectMeshCylinder& other )
 :	ObjectMesh  ( other ),
	radius_m    ( other.radius_m ),
	height_m    ( other.height_m ),
	isCurved_m  ( other.isCurved_m ),
	resolution_m( other.resolution_m )
{
}


ObjectMeshCylinder& ObjectMeshCylinder::operator=( const ObjectMeshCylinder& other )
{
	assign( other );

	return *this;
}




/// implementation ---------------------------------------------------------------------------------
void ObjectMeshCylinder::assign
(
	const ObjectRenderable& other
)
{
	if( &other != this )
	{
		ObjectMesh::assign( other );

		radius_m     = 0.0f;
		height_m     = 0.0f;
		isCurved_m   = false;
		resolution_m = 0.0f;
	}
}


void ObjectMeshCylinder::assign
(
	const ObjectMesh& other
)
{
	if( &other != this )
	{
		ObjectMesh::assign( other );

		radius_m     = 0.0f;
		height_m     = 0.0f;
		isCurved_m   = false;
		resolution_m = 0.0f;
	}
}


void ObjectMeshCylinder::assign
(
	const ObjectMeshCylinder& other
)
{
	if( &other != this )
	{
		ObjectMesh::assign( other );

		radius_m     = other.radius_m;
		height_m     = other.height_m;
		isCurved_m   = other.isCurved_m;
		resolution_m = other.resolution_m;
	}
}


void ObjectMeshCylinder::makeCylinder
(
	const float               radius,
	const float               height,
	const bool                isCurved,
	const dword               resolution,
	vector<Vector3f>&         vertexs,
	vector<TriangleImplicit>& faces
)
{
	/// center base at 0,0,0, axis in y
	/// resolution number of vertical panels, each two triangles
	///	resolution number of short rim vertical panels top and bottom, each two triangles
	/// capped each end by resolution number of triangles radiating from axis


	/// setup storage
	vertexs.resize( (resolution * 4) + 2 );
	faces.resize( resolution * 8 );


	/// make column vertexs heights
	static const float RIM = 0.001f;
	const float columnHeights[4] = { 0.0f, RIM, height - RIM, height };

	dword vertexIndex = 0;

	/// make first four vertexs, in column at radius distance
	for( int j = 0;  j < 4;  ++j )
	{
		vertexs[vertexIndex++].setXYZ( 0.0f, columnHeights[j], radius );
	}

	/// make vertical side panels
	dword faceIndex = 0;
	for( dword i = 1;  i <= resolution;  ++i )
	{
		/// make next four vertexs, in column at radius distance
		if( i < resolution )
		{
			const float angle = -6.283185307179586476925286766559f * (float(i) / float(resolution));
			float x;
			float z;
			hxa7241::graphics::polarToRectangular( angle, radius, z, x );
			for( int j = 0;  j < 4;  ++j )
			{
				vertexs[vertexIndex++].setXYZ( x, columnHeights[j], z );
			}
		}
		else
		{
			vertexIndex += 4;
		}

		/// make triangles using four vertexs and previous four, and cap center vertexs
		{
			/// a0, -4, -8
			/// a1, -1, -5
			/// -4, -8, -7
			/// -7, -3, -4
			/// -3, -7, -6
			/// -6, -2, -3
			/// -2, -6, -5
			/// -5, -1, -2
			static const byte VI[2][3] = { {2, 6, 5}, {5, 1, 2} };
			const dword vertexsLength = vertexs.size() - 2;
			for( dword j = 3;  j-- > 0; )
			{
				const dword vi = vertexIndex - j;
				for( dword k = 2;  k-- > 0; )
				{
					TriangleImplicit& face = faces[faceIndex++];
					face.set( word((vi - VI[k][0]) % vertexsLength),
					          word((vi - VI[k][1]) % vertexsLength),
					          word((vi - VI[k][2]) % vertexsLength) );
					face.set( isCurved );

					//Vector2<word> txy[3];
					//const float textureStepX = textureScale / float(resolution);
					//for( dword t = 3;  t-- > 0; )
					//{
					//	const float tx = textureStepX * float(i - ((VI[k][t] - 1) >> 2));
					//	const float ty =  * columnHeights[ (VI[k][t] - 1) & 3 ];
					//	txy[t].set( tx, ty );
					//}
					//face.set( txy );
				}
			}

			static const dword VIC[2][3] = { {vertexsLength, 8, 4}, {vertexsLength + 1, 1, 5} };
			for( dword k = 2;  k-- > 0; )
			{
				TriangleImplicit& face = faces[faceIndex++];
				face.set( word(VIC[k][0]),
				          word((vertexIndex - VIC[k][1]) % vertexsLength),
				          word((vertexIndex - VIC[k][2]) % vertexsLength) );
				face.set( false );
			}
		}
	}

	/// make two cap center vertexs
	vertexs[vertexs.size() - 2].setXYZ( 0.0f, 0.0f, 0.0f );
	vertexs[vertexs.size() - 1].setXYZ( 0.0f, height, 0.0f );
}




/// serialization ----------------------------------------------------------------------------------
SerializerOut& perceptuum2::modelling::operator<<( SerializerOut& out, const ObjectMeshCylinder& obj )
{
	/// write radius
	out.write( "diameter", obj.radius_m * 2.0f );

	/// write height
	out.write( "height", obj.height_m );

	/// write curvedness
	out.write( "isCurved", obj.isCurved_m );

	/// write resolution
	out.write( "resolution", obj.resolution_m );

	/// write material id
	out.write( "materialRef", obj.getMaterialName() );

	return out;
}


SerializerIn& perceptuum2::modelling::operator>>( SerializerIn& in, ObjectMeshCylinder& obj )
{
	try
	{
		/// read
		dword materialId;// = 0;
		{
			/// read radius
			{
				float diameter = 0.0f;
				in.read( "diameter", diameter );
				if( diameter <= 0.0f )
				{
					diameter = 1.0f;
				}
				obj.radius_m = diameter * 0.5f;
			}

			/// read height
			{
				float height = 0.0f;
				in.read( "height", height );
				if( height <= 0.0f )
				{
					height = 1.0f;
				}
				obj.height_m = height;
			}

			/// read curvedness
			in.read( "isCurved", obj.isCurved_m );

			/// read resolution
			in.read( "resolution", obj.resolution_m );
			static const dword RESOLUTION_MAX = 1000;
			if( obj.resolution_m < 3 )
			{
				obj.resolution_m = 3;
			}
			else if( obj.resolution_m > RESOLUTION_MAX )
			{
				obj.resolution_m = RESOLUTION_MAX;
			}

			/// read material id
			{
				string materialName;
				in.read( "materialRef", materialName );
				materialId = obj.getMaterialIndex( materialName );
			}
		}


		/// make
		vector<Vector3f>         vertexs;
		vector<TriangleImplicit> faces;
		obj.makeCylinder( obj.radius_m, obj.height_m, obj.isCurved_m, obj.resolution_m,
						  vertexs, faces );

		/// set base
		obj.setMesh( vertexs, faces, materialId );


		return in;
	}
	catch( ... )
	{
		/// zero object
		obj.radius_m     = 0.0f;
		obj.height_m     = 0.0f;
		obj.resolution_m = 0;

		obj.zeroObject();

		throw;
	}
}


const char* ObjectMeshCylinder::CLASS_NAME()
{
	static const char CLASS_NAME[] = "perceptuum2::modelling::ObjectMeshCylinder";
	return CLASS_NAME;
}


const char* ObjectMeshCylinder::CLASS_DESCRIPTION()
{
	static const char CLASS_DESCRIPTION[] = "a cylinder, made of a ObjectMesh";
	return CLASS_DESCRIPTION;
}
