/*--------------------------------------------------------------------

   Perceptuum renderer, version 2.1
   Copyright (c) 2002-2003,  Harrison Ainsworth.

   www.hxa7241.org

--------------------------------------------------------------------*/

/*--------------------------------------------------------------------

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston,
   MA 02111-1307, USA

--------------------------------------------------------------------*/


#include "SpacetimeIndex.h"
#include "SamplingControl.h"
#include "SamplingControlEmitter.h"
#include "MediaStack.h"
#include "Light.h"
#include "LightStream.h"
#include "Medium.h"
#include "EmitterPathNode.h"
#include "EyePathNode.h"
#include "Scene.h"
#include "ObjectRenderableId.h"
#include "Intersection.h"
#include "Qualities.h"
#include "Ray.h"
#include "Vector2.h"
#include "UnitVector3f.h"
#include "CieXyz.h"
#include "Float01.h"
#include "Float01c.h"
#include "Logger.h"

#include "EmitterPath.h"   /// own header is included last


using namespace perceptuum2::transport;




/// standard object services -----------------------------------------------------------------------
EmitterPath::EmitterPath()
 :	pathLengthLimit_m ( 0 ),
	pathEndThreshold_m( 0 ),
	isRoulette_m      ( true ),
	nodes_m           (),
	nodesLength_m     ( 0 )
{
}


EmitterPath::EmitterPath
(
	const dword    pathLengthLimit,
	const Float01c pathEndThreshold,
	const bool     isRoulette
)
 :	pathLengthLimit_m ( 0 ),
	pathEndThreshold_m( 0 ),
	isRoulette_m      ( true ),
	nodes_m           (),
	nodesLength_m     ( 0 )
{
	EmitterPath::set( pathLengthLimit, pathEndThreshold, isRoulette );
}


EmitterPath::~EmitterPath()
{
}


EmitterPath::EmitterPath( const EmitterPath& other )
{
	EmitterPath::assign( other );
}


EmitterPath& EmitterPath::operator=( const EmitterPath& other )
{
	assign( other );

	return *this;
}




/// commands ---------------------------------------------------------------------------------------
void EmitterPath::set
(
	const dword    pathLengthLimit,
	const Float01c pathEndThreshold,
	const bool     isRoulette
)
{
	pathLengthLimit_m  = pathLengthLimit >= 0 ? pathLengthLimit : 0;
	pathEndThreshold_m = pathEndThreshold;
	isRoulette_m       = isRoulette;

	nodes_m.resize( pathLengthLimit_m + 1 );
}


void EmitterPath::trace
(
	const SpacetimeIndex&         sceneIndex,
	const SamplingControlEmitter& emitterSamplingControl,
	vector<float>&                lightsourceCompositionStorage,
	const float                   emittanceScaling
)
{
	/// choose light source emit position and first direction
	Intersection sourceIntersection;
	Light        sourceEmittance;
	LightStream  sourcePathStep;
	bool         isSourceTransmitted = false;
	bool         isSourceDiffuse     = true;
	bool         isAnyEmit;
	{
		const ObjectRenderableId rootObject( sceneIndex.getScene().getRootObject() );
		isAnyEmit = chooseLightSourceStart( rootObject, emitterSamplingControl, lightsourceCompositionStorage,
		                                    sourceIntersection, sourceEmittance, sourcePathStep, isSourceTransmitted, isSourceDiffuse );
		sourceEmittance        *= emittanceScaling;
		sourcePathStep.light_m *= emittanceScaling;
	}


	/// reset nodes length
	nodesLength_m = 0;

	/// skip everything if nothing emitted
	if( isAnyEmit )
	{
		//hxa7241::utility::Logger() << "EmitterPath -  tracing...";

		const ObjectRenderableId defaultObject( sceneIndex.getScene().getDefaultObject() );

		/// add root path node to path list
		nodes_m[0].setEmitterPathNode( sourceIntersection, defaultObject, sourceEmittance );
		++nodesLength_m;

		//hxa7241::utility::Logger() << "EmitterPath -  node " << (nodesLength_m - 1) << "  " << nodes_m[(nodesLength_m - 1)];

		if( pathLengthLimit_m > 0 )
		{
			const float oneOverSourcePathStepMean = (sourcePathStep.light_m.average() != 0.0f) ? (1.0f / sourcePathStep.light_m.average()) : 0.0f;

			/// make some path tracing things
			MediaStack media( defaultObject );
			perceptuum2::modelling::Ray     ray;
			perceptuum2::interaction::Light light;
			bool                            isDiffuse = isSourceDiffuse;

			/// make first step with pre-made outward lightstream
			nodes_m[0].getRay( sourcePathStep.direction_m, isSourceTransmitted, ray );
			light = sourcePathStep.light_m;

			/// step along path adding nodes
			for( ;  !( ray.getDirection().isZero() || light.isZero() ||
			           ((light.average() * oneOverSourcePathStepMean) < pathEndThreshold_m) );  ++nodesLength_m )
			{
				//hxa7241::utility::Logger() << "EmitterPath -  ray " << ray;

				/// update media stack
				media.update( ray );

				//hxa7241::utility::Logger() << "EmitterPath -  media " << media;

				/// intersect ray, and end path if no intersection
				Intersection intersection;
				if( !sceneIndex.getNextIntersection( ray, intersection ) )
				{
					break;
				}

				/// get path interaction stochastics
				hxa7241::utility::Float01 stochastics[8];
				emitterSamplingControl.getInteractionCoords( nodesLength_m - 1, stochastics );

				/// calc light change through medium (from path node to intersection)
				{
					const perceptuum2::interaction::Medium medium( media.getTop() );
					const bool isKilled = !medium.getParticipationRoulette( stochastics[7], ray.getPosition(), intersection.position_m, light );

					/// maybe end path with roulette
					if( isKilled )
					{
						break;
					}
				}

				/// add next path node
				{
					const ObjectRenderableId outsideObject( media.getOutsideObject( ray.getDirection().getVector(), intersection.normal_m.getVector(), intersection.object_m ) );
					nodes_m[nodesLength_m].setEmitterPathNode( intersection, outsideObject, LightStream( ray.getDirection(), light ), isDiffuse );

					//hxa7241::utility::Logger() << "EmitterPath -  node " << nodesLength_m << "  " << nodes_m[nodesLength_m];
				}

				/// check if path length limit is reached
				if( (nodesLength_m + 1) > pathLengthLimit_m )
				{
					++nodesLength_m;
					break;
				}

				/// get bounced ray from last path node
				nodes_m[nodesLength_m].getNextPathStep( isRoulette_m, stochastics, ray, light, isDiffuse );
			}
		}
	}
}




/// queries ----------------------------------------------------------------------------------------
void EmitterPath::getSpreadLight
(
	const SpacetimeIndex& sceneIndex,
	const EyePathNode&    eyePathNode,
	const Float01c        shinynessPoint,
	vector<LightStream>&  spreadStorage
) const
{
	/// prepare illuminations collection
	spreadStorage.resize( 0 );
	spreadStorage.reserve( nodesLength_m );

	///	step thru emitter path nodes, building illumination array
	for( int j = 0;  j < nodesLength_m;  ++j )
	{
		const EmitterPathNode& emitterPathNode = nodes_m[j];

		/// find if emitter path node has unobscured view of eye path node
		perceptuum2::modelling::Ray rayFromEmitterPathNode;
		emitterPathNode.getRay( eyePathNode.getPosition(), rayFromEmitterPathNode );
		const bool isIlluminated = !sceneIndex.isShadowed( rayFromEmitterPathNode, eyePathNode.getPosition(), eyePathNode.getObject() );

		/// get light from emitter path node to eye path node
		if( isIlluminated )
		{
			Light illumination;
			emitterPathNode.getSpreadLight( eyePathNode.getPosition(), rayFromEmitterPathNode.getDirection(), shinynessPoint, illumination );

			/// calc light change through medium from emitter path node to eye path node
			{
				using perceptuum2::modelling::ObjectRenderableId;
				const ObjectRenderableId& mediumObject = eyePathNode.getObject( (-rayFromEmitterPathNode.getDirection()).getVector() );
				perceptuum2::interaction::Medium medium( mediumObject );
				medium.participateLight( emitterPathNode.getPosition(), eyePathNode.getPosition(), illumination );
			}

			/// append illumination to collection
			spreadStorage.push_back( LightStream( rayFromEmitterPathNode.getDirection(), illumination ) );
		}
	}
}


dword EmitterPath::getLength() const
{
	return nodesLength_m;
}


const vector<EmitterPathNode>& EmitterPath::getNodes() const
{
	return nodes_m;
}


const vector<EmitterPathNode>& EmitterPath::getNodes
(
	dword& length
) const
{
	length = nodesLength_m;
	return nodes_m;
}




/// implementation ---------------------------------------------------------------------------------
bool EmitterPath::chooseLightSourceStart
(
	const ObjectRenderableId&     rootObject,
	const SamplingControlEmitter& samplingControl,
	vector<float>&                lightsourceCompositionStorage,
	Intersection&                 intersection,
	Light&                        sourceEmittance,
	LightStream&                  firstPathStep,
	bool&                         isTransmitted,
	bool&                         isDiffuse
) const
{
	/// get emitting point
	perceptuum2::modelling::Qualities qualities;
	bool isAnyEmit;// = false;
	{
		const Vector3f posStochastic( samplingControl.getPositionCoords() );
		intersection.object_m = rootObject;
		isAnyEmit = rootObject->getEmitted( posStochastic, lightsourceCompositionStorage, intersection, qualities );
	}

	if( isAnyEmit )
	{
		/// make emittance with scene power and point color
		{
			/// entire scene emittance colorless magnitude * particular emitting point magnitudeless color
			const float sceneEmittanceMagnitude = rootObject->getEmittanceSum().average();
			const float pointEmittedMagnitude   = qualities.emitted_m.average();
			if( pointEmittedMagnitude != 0.0f )
			{
				qualities.emitted_m *= (sceneEmittanceMagnitude / pointEmittedMagnitude);
			}
			sourceEmittance = qualities.emitted_m;
		}

		/// make emitted light and direction
		if( pathLengthLimit_m > 0 )
		{
			using hxa7241::graphics::Vector2;
			using hxa7241::utility::Float01;
			const Vector2<Float01> dirStochastic( samplingControl.getDirectionCoords() );
			using perceptuum2::interaction::SurfacePoint;
			isTransmitted = SurfacePoint::getEmittedRay( qualities, dirStochastic, intersection.normal_m, isRoulette_m,
			                                             firstPathStep.direction_m, firstPathStep.light_m, isDiffuse );
		}
		else
		{
			firstPathStep = LightStream();
			isTransmitted = false;
			isDiffuse     = true;
		}

		//hxa7241::utility::Logger() << "EmitterPath -  start  " << intersection;
		//hxa7241::utility::Logger() << "EmitterPath -  first step  " << firstPathStep << "  " << isTransmitted;
	}
	else
	{
		//hxa7241::utility::Logger() << "EmitterPath -  no emit";

		intersection    = Intersection();
		sourceEmittance = Light();
		firstPathStep   = LightStream();
		isTransmitted   = false;
		isDiffuse       = true;
	}

	return isAnyEmit;



	/// make specular lightstreams
		/// make specular lightstreams
			/// step thru all scene objects
				/// if specularness > threshold
					/// project object bounding box onto hemisphere grid
			/// loop for all wanted specular rays (illuminations.size())
				/// pick grid cell
					/// grid cell = trunc(rand01 * specular cell count)
				/// pick subcell direction
				/// call lightsource to get emitted light for direction and position
				/// scale light by 1 / specular ray count

		/// scale lightstreams
			/// specular weighting = (specular area / total hemisphere area)
			/// general   *= 1 / (1 + specular weighting)
			/// speculars *= specular weighting / (1 + specular weighting)
}


void EmitterPath::assign
(
	const EmitterPath& other
)
{
	if( &other != this )
	{
		pathLengthLimit_m  = other.pathLengthLimit_m;
		pathEndThreshold_m = other.pathEndThreshold_m;
		isRoulette_m       = other.isRoulette_m;

		nodes_m       = other.nodes_m;
		nodesLength_m = other.nodesLength_m;
	}
}
