/*--------------------------------------------------------------------

   Perceptuum renderer, version 2.1
   Copyright (c) 2002-2003,  Harrison Ainsworth.

   www.hxa7241.org

--------------------------------------------------------------------*/

/*--------------------------------------------------------------------

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston,
   MA 02111-1307, USA

--------------------------------------------------------------------*/


#include "SpacetimeIndex.h"
#include "SamplingControl.h"
#include "SamplingControlPath.h"
#include "MediaStack.h"
#include "EmitterPathNode.h"
#include "EyePathNode.h"
#include "Medium.h"
#include "Ray.h"
#include "Intersection.h"
#include "Light.h"
#include "Float01.h"
#include "Float01c.h"
#include "Histogram.h"
#include "Logger.h"

#include "Illumination.h"   /// own header is included last


using namespace perceptuum2::transport;




/// standard object services -----------------------------------------------------------------------
Illumination::Illumination
(
	const dword    emitterPathNumber,
	const dword    photonMapGeneralSize,
	//const dword    photonMapCausticSize,
	//const dword    gatherRayCount,
	const float    ,//lightsourceEmittanceThreshold,
	const dword    pathLengthLimit,
	const Float01c pathEndThreshold,
	const bool     isLogging
)
 :	emitterPaths_m                 ( emitterPathNumber >= 0 ? emitterPathNumber : 0 ),
	photonMapBuffers_m             (),
	photonMap_m                    ( &photonMapBuffers_m, photonMapGeneralSize, true ),
	photonMapDirect_m              ( &photonMapBuffers_m, emitterPaths_m.size() == 0 ? photonMapGeneralSize : 0, true ),
	photonMapCaustic_m             ( &photonMapBuffers_m, 0, true ),
	photonPathBuffer_m             ( pathLengthLimit, pathEndThreshold, true ),
	lightsourceHistogramStorage_m  (),
	spreadStorage_m                (),
//	gridStorage_m                  (),
//	importanceStorage_m            (),
//	gatherRayCount_m               ( 0 ),//gatherRayCount >= 0 ? gatherRayCount : 0 ),
	totalGeneralPathCount_m        ( 0 ),
	totalGeneralNodeCount_m        ( 0 ),
	totalCausticPathCount_m        ( 0 ),
	totalCausticNodeCount_m        ( 0 ),
	totalIlluminationQueriesCount_m( 0 ),
	isLogging_m                    ( isLogging )
{
	/// init emitter paths
	for( dword i = emitterPaths_m.size();  i-- > 0; )
	{
		emitterPaths_m[i].set( 0, pathEndThreshold, true );//pathLengthLimit/2, false );
	}

	if(isLogging_m) hxa7241::utility::Logger() << "Illumination -  constructed,  emitter paths: " << emitterPaths_m.size() <<
	                                              "  general photon map size: " << photonMap_m.getCapacity() <<
	                                              "  direct photon map size: " << photonMapDirect_m.getCapacity() <<
	                                              "  caustic photon map size: " << photonMapCaustic_m.getCapacity();
}


Illumination::~Illumination()
{
}




/// commands ---------------------------------------------------------------------------------------
void Illumination::makeIllumination
(
	const SpacetimeIndex& sceneIndex,
	SamplingControl&      samplingControl
)
{
	makeDirectIllumination( sceneIndex, samplingControl );

	makePhotonMapIllumination( sceneIndex, true, samplingControl );
}




/// queries ----------------------------------------------------------------------------------------
void Illumination::getIllumination
(
	const SamplingControlDiffuse& samplingControl,
	const SpacetimeIndex&         sceneIndex,
	const EyePathNode&            lastEyePathNode,
	const MediaStack&             lastEyePathNodeMedia,
	const UnitVector3f&           eyeDirection,
	const Float01c                shinynessPoint,
	Light&                        illumination
) const
{
	/// direct illumination
	Light directIllumination;
	getDirectIllumination( samplingControl, sceneIndex, lastEyePathNode, lastEyePathNodeMedia, eyeDirection, shinynessPoint, directIllumination );


	/// indirect from photon maps
	Light photonMapIllumination;
	{
		getPhotonMapIllumination( photonMap_m, lastEyePathNode, eyeDirection, shinynessPoint, photonMapIllumination );

		///// from caustic photon map
		//Light photonMapCausticIllumination;
		//getPhotonMapIllumination( photonMapCaustic_m, lastEyePathNode, eyeDirection, shinynessPoint, photonMapCausticIllumination );
		//photonMapIllumination += photonMapCausticIllumination;
	}


	/// sum the parts
	illumination = (directIllumination += photonMapIllumination);


	/// track overall stats
	++totalIlluminationQueriesCount_m;
}


void Illumination::getDirectIllumination
(
	const SamplingControlDiffuse& ,//samplingControl,
	const SpacetimeIndex&         sceneIndex,
	const EyePathNode&            lastEyePathNode,
	const MediaStack&             ,//lastEyePathNodeMedia,
	const UnitVector3f&           eyeDirection,
	const Float01c                shinynessPoint,
	Light&                        directIllumination
) const
{
	/// from emitter points
	Light emitterPathIllumination;
	for( dword i = emitterPaths_m.size();  i-- > 0; )
	{
		emitterPaths_m[i].getSpreadLight( sceneIndex, lastEyePathNode, shinynessPoint, spreadStorage_m );

		Light illumination;
		lastEyePathNode.getGatheredLightDistal( spreadStorage_m, eyeDirection, shinynessPoint, illumination );
		emitterPathIllumination += illumination;
	}

	/// from photon map
	Light photonMapIllumination;
	getPhotonMapIllumination( photonMapDirect_m, lastEyePathNode, eyeDirection, shinynessPoint, photonMapIllumination );

	/// sum the parts
	directIllumination = (emitterPathIllumination += photonMapIllumination);
}




/// implementation ---------------------------------------------------------------------------------
void Illumination::makeDirectIllumination
(
	const SpacetimeIndex& sceneIndex,
	SamplingControl&      samplingControl
)
{
	if(isLogging_m) hxa7241::utility::Logger() << "Illumination -  tracing for direct illumination";

	/// make emitter points
	if( emitterPaths_m.size() > 0 )
	{
		const float scaling = 1.0f / float(emitterPaths_m.size());
		for( dword i = emitterPaths_m.size();  i-- > 0; )
		{
			emitterPaths_m[i].trace( sceneIndex, samplingControl.getEmitterControl0(), lightsourceHistogramStorage_m, scaling );

			samplingControl.advanceIllumination0();
		}
	}

	/// make direct photon map
	if( photonMapDirect_m.getCapacity() > 0 )
	{
		photonMapDirect_m.switchToBuildingMode();

		EmitterPath photonPathBuffer( 1, 0.0f, true );

		dword pathCount = 0;
		for( dword i = photonMapDirect_m.getCapacity();
		     (photonMapDirect_m.getRemainingSpace() > 0) & (i-- > 0);
		     ++pathCount )
		{
			photonPathBuffer.trace( sceneIndex, samplingControl.getEmitterControl0(), lightsourceHistogramStorage_m, 1.0f );

			using perceptuum2::interaction::EmitterPathNode;
			const EmitterPathNode& node = photonPathBuffer.getNodes()[1];

			const LightStream& inward = node.getInwardLightstream();
			const dword        id     = 0;

			photonMapDirect_m.addElement( inward.light_m, node.getPosition(), inward.direction_m, id );

			samplingControl.advanceIllumination0();
		}

		if( pathCount > 0 )
		{
			photonMapDirect_m.scale( 1.0f / float(pathCount) );
		}

		photonMapDirect_m.switchToIlluminatingMode();
	}
}


void Illumination::makePhotonMapIllumination
(
	const SpacetimeIndex& sceneIndex,
	const bool            isDiscludeDirectLighting,
	SamplingControl&      samplingControl
)
{
	if( (photonMap_m.getCapacity() > 0) | (photonMapCaustic_m.getCapacity() > 0) )
	{
		if(isLogging_m) hxa7241::utility::Logger() << "Illumination -  tracing for photon maps";

		photonMap_m.switchToBuildingMode();
		photonMapCaustic_m.switchToBuildingMode();

		/// count paths (with an upper limit imposed)
		dword pathCountGeneral = 0;
		dword pathCountCaustic = 0;
		dword i = photonMap_m.getCapacity() >= photonMapCaustic_m.getCapacity() ? photonMap_m.getCapacity() : photonMapCaustic_m.getCapacity();
		for( bool isGeneralFull, isCausticFull;
		     i-- > 0;
		     pathCountGeneral += dword(!isGeneralFull), pathCountCaustic += dword(!isCausticFull) )
		{
			/// make a photon path
			photonPathBuffer_m.trace( sceneIndex, samplingControl.getEmitterControl1(), lightsourceHistogramStorage_m, 1.0f );

			const dword nodeCount = photonPathBuffer_m.getLength();

			/// if path wont fit in photon maps, end loop
			isGeneralFull = nodeCount > photonMap_m.getRemainingSpace();
			isCausticFull = nodeCount > photonMapCaustic_m.getRemainingSpace();
			if( isGeneralFull & isCausticFull )
			{
				break;
			}

			using perceptuum2::interaction::EmitterPathNode;
			const vector<EmitterPathNode>& photonPathNodes = photonPathBuffer_m.getNodes();

			/// add path nodes to photon maps, discluding node on emitter
			for( dword nodeIndex = 1 + dword(isDiscludeDirectLighting);  nodeIndex < nodeCount;  ++nodeIndex )
			{
				const EmitterPathNode& node = photonPathNodes[nodeIndex];

				const LightStream& inward    = node.getInwardLightstream();
				const dword        id        = 0;
				const bool         isCaustic = !node.isInwardDiffuse();

				if( !isGeneralFull )// & !isCaustic )
				{
					photonMap_m.addElement( inward.light_m, node.getPosition(), inward.direction_m, id );

					/// track overall stats
					++totalGeneralNodeCount_m;
				}

				if( !isCausticFull & isCaustic )
				{
					photonMapCaustic_m.addElement( inward.light_m, node.getPosition(), inward.direction_m, id );

					/// track overall stats
					++totalCausticNodeCount_m;
				}
			}

			samplingControl.advanceIllumination1();
		}

		/// scale photon maps to 1 / added path count
		if( pathCountGeneral > 0 )
		{
			photonMap_m.scale( 1.0f / float(pathCountGeneral) );
		}
		if( pathCountCaustic > 0 )
		{
			photonMapCaustic_m.scale( 1.0f / float(pathCountCaustic) );
		}

		photonMap_m.switchToIlluminatingMode();
		photonMapCaustic_m.switchToIlluminatingMode();

		/// track overall stats
		totalGeneralPathCount_m += pathCountGeneral;
		totalCausticPathCount_m += pathCountCaustic;

		if(isLogging_m) hxa7241::utility::Logger() << "Illumination -  totalGeneralPathCount: " << totalGeneralPathCount_m << "  totalGeneralNodeCount: " << totalGeneralNodeCount_m;
		if(isLogging_m) hxa7241::utility::Logger() << "Illumination -  totalCausticPathCount: " << totalCausticPathCount_m << "  totalCausticNodeCount: " << totalCausticNodeCount_m;
	}
}


void Illumination::getPhotonMapIllumination
(
	const PhotonMap&    photonMap,
	const EyePathNode&  lastEyePathNode,
	const UnitVector3f& eyeDirection,
	const Float01c      shinynessPoint,
	Light&              eyewardLight
)
{
	const vector<LightStream>* pIllumination = photonMap.getIllumination( lastEyePathNode.getPosition() );
	if( pIllumination != 0 )
	{
		/// get bounced light from eye path node
		lastEyePathNode.getGatheredLightLocal( *pIllumination, eyeDirection, shinynessPoint, eyewardLight );
	}
	else
	{
		eyewardLight = Light::ZERO();
	}
}




	/// make histogram with incoming light
		/// make square grid of side 2 * sqrt(photons count)
		/// get nearby photons from photon map
			/// direction and filtered power
		/// loop through photons
			/// (map direction to grid coords using bssdf inverse sdf) [a]
			/// map direction to grid coords with uniform mapping [b]
			/// add power to grid cell at coords
		/// scale grid by bssdf grid [b]
		/// loop through grid cells
			/// add cell power to histogram
		/// unitize histogram ?
	/// send rays using histogram
		/// make ray
			/// choose cell using histogram
			/// make sub cell coords with stochastics and cell dimensions
			/// (map coords to direction with bssdf sdf mapping) [a]
			/// map coords to direction with uniform mapping [b]
			/// set light scaling to (1 / photon probability) * (cell area / whole grid area)
		/// intersect ray
		/// do local gather with photon map at intersection
		/// calculate light back through ray and bounce toward eye
		/// scale light by 1/ray count * scaling
		/// accumulate light
	/// return light

	/// ? media
	/// ? sampling control
	/// ? interface through SurfacePoint
		/// ? direction clamping
		/// ? two grids - for whole sphere : yes
		/// ? facet or bump normal : bump


/*
/// unreviewed, untested
/// pending further consideration

void Illumination::gather
(
	const SamplingControlDiffuse& samplingControl,
	const SpacetimeIndex&         sceneIndex,
	const EyePathNode&            localNode,
	const MediaStack&             localNodeMedia,
	const UnitVector3f&           eyeDirection,
	const Float01c                shinynessPoint,
	Light&                        light
) const
{
	/// get nearby photons illumination from photon map
	const vector<LightStream>* pIlluminations = photonMap_m.getIllumination( localNode.getPosition() );
	if( pIlluminations != 0 )
	{
		const UnitVector3f negEyeDirection( -eyeDirection );

		/// make importance histogram
		hxa7241::utility::Histogram importance( importanceStorage_m );
		localNode.makeImportanceHistogram( *pIlluminations, negEyeDirection, gridStorage_m, importance );

		/// trace gather rays
		Light gatheredLight;
		for( dword rayIndex = 0, end = gatherRayCount_m;  rayIndex < end;  ++rayIndex )
		{
			/// make ray
			perceptuum2::modelling::Ray ray;
			Light scaling;
			{
				Float01 stochastics[SamplingControlDiffuse::COORDS_COUNT];
				samplingControl.getSampleCoords( rayIndex, stochastics );

				UnitVector3f direction;
				const bool isTransmitted = localNode.getSamplingDirection( importance, stochastics, negEyeDirection, direction, scaling );
				localNode.getRay( direction, isTransmitted, ray );
			}

			if( !(ray.getDirection().isZero() | scaling.isZero()) )
			{
				/// update media stack
				MediaStack media( localNodeMedia );
				media.update( ray );

				/// intersect ray
				perceptuum2::modelling::Intersection intersection;
				if( sceneIndex.getNextIntersection( ray, intersection ) )
				{
					const UnitVector3f localNodeDirection( -ray.getDirection() );

					/// make path node for hit point
					const EyePathNode distalNode( intersection, media.getOutsideObject( ray.getDirection().getVector(), intersection.normal_m.getVector(), intersection.object_m ) );

					/// get local gather from photon map
					/// aah, but i want the diffuse proportion of the nodes scattering... - need to modify EyePathNode...
					Light distalLight;
					getPhotonMapIllumination( photonMap_m, distalNode, localNodeDirection, shinynessPoint, distalLight );

					/// add emitted
					{
						Light emittedLight;
						distalNode.getEmittedLight( localNodeDirection, emittedLight );
						distalLight += emittedLight;
					}

					/// scale light through medium
					{
						perceptuum2::interaction::Medium medium( media.getTop() );
						medium.participateLight( ray.getPosition(), distalNode.getPosition(), distalLight );
					}

					/// accumulate light
					gatheredLight += (distalLight *= scaling);
				}
			}
		}

		if( gatherRayCount_m > 0 )
		{
			gatheredLight *= (3.1415926535897932384626433832795f / float(gatherRayCount_m));
		}
		light = gatheredLight;
	}
	else
	{
		light = Light::ZERO();
	}
}*/
