/*--------------------------------------------------------------------

   Perceptuum renderer, version 2.1
   Copyright (c) 2002-2003,  Harrison Ainsworth.

   www.hxa7241.org

--------------------------------------------------------------------*/

/*--------------------------------------------------------------------

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston,
   MA 02111-1307, USA

--------------------------------------------------------------------*/


#include "Vector3f.h"

#include "OctreeVisitorPhoton.h"   /// own header is included last


using namespace perceptuum2::transport;




/// standard object services -----------------------------------------------------------------------
OctreeVisitorPhoton::OctreeVisitorPhoton()
 :	OctreeVisitorT<Photon>(),
	pSelectedItems_m( 0 ),
	center_m        (),
	radiusOuter_m   (),
	radiusInner_m   (),
	visitCount_m    ( 0 )
{
}


OctreeVisitorPhoton::OctreeVisitorPhoton
(
	vector<std::pair<float, const Photon*> >* pSelectedItemsBuffer,
	const Vector3f& center,
	const float     radiusOuter,
	const float     radiusInner
)
 :	OctreeVisitorT<Photon>(),
	pSelectedItems_m( pSelectedItemsBuffer ),
	center_m        ( center ),
	radiusOuter_m   ( radiusOuter ),
	radiusInner_m   ( radiusInner ),
	visitCount_m    ( 0 )
{
}


OctreeVisitorPhoton::~OctreeVisitorPhoton()
{
}




/// commands ---------------------------------------------------------------------------------------
void OctreeVisitorPhoton::visitItem
(
	const Photon& item,
	const bool    isInside
)
{
	/// only act if inclusion condition met
	if( isInside ||
		isPhotonOverlappingShell( item, center_m, radiusOuter_m, radiusInner_m ) )
	{
		if( pSelectedItems_m != 0 )
		{
			/// check selected buffer wont overflow
			if( pSelectedItems_m->size() < pSelectedItems_m->capacity() )
			{
				/// get distance of photon from shell center
				const float distance2 = getDistanceSquared( item, center_m );

				/// add item and distance to collection
				pSelectedItems_m->push_back( std::pair<float, const Photon*>( distance2, &item ) );
			}

			++visitCount_m;
		}
	}
}




/// queries ----------------------------------------------------------------------------------------
bool OctreeVisitorPhoton::isOverlappingCell
(
	const Photon&   item,
	const Vector3f& position,
	const float     size
) const
{
	const Vector3f& itemPosition = item.getPosition();

	bool isOverlapped;

	const float offsetX = itemPosition.getX() - position.getX();
	isOverlapped  = (offsetX >= 0.0f) & (offsetX < size);

	const float offsetY = itemPosition.getY() - position.getY();
	isOverlapped &= (offsetY >= 0.0f) & (offsetY < size);

	const float offsetZ = itemPosition.getZ() - position.getZ();
	isOverlapped &= (offsetZ >= 0.0f) & (offsetZ < size);


	return isOverlapped;
}


dword OctreeVisitorPhoton::getSubcellOverlaps
(
	const Photon&   item,
	const Vector3f& planePositions
) const
{
	const Vector3f& itemPosition = item.getPosition();

	const bool isHigherX = itemPosition.getX() >= planePositions.getX();
	const bool isHigherY = itemPosition.getY() >= planePositions.getY();
	const bool isHigherZ = itemPosition.getZ() >= planePositions.getZ();

	const dword xBits = (dword(isHigherX) - 1) ^ 0xAA;
	const dword yBits = (dword(isHigherY) - 1) ^ 0xCC;
	const dword zBits = (dword(isHigherZ) - 1) ^ 0xF0;

	return (xBits & yBits & zBits) & 0xFF;
}


OctreeVisitor::ESelection OctreeVisitorPhoton::isSelected
(
	const OctreeBound& octreeBound,
	const bool         isInside
) const
{
	ESelection s = INSIDE;

	if( !isInside )
	{
		s = isCellOverlappingShell( octreeBound, center_m, radiusOuter_m, radiusInner_m );
	}

	return s;
}


dword OctreeVisitorPhoton::getSubcellSelections
(
	const OctreeBound& octreeBound,
	const bool         isInside
) const
{
	dword s = ALL_INSIDE;

	if( !isInside )
	{
		s = getSubcellShellOverlaps( octreeBound, center_m, radiusOuter_m, radiusInner_m );
	}

	return s;
}


dword OctreeVisitorPhoton::getVisitCount() const
{
	return visitCount_m;
}




/// implementation ---------------------------------------------------------------------------------
bool OctreeVisitorPhoton::isPhotonOverlappingShell
(
	const Photon&   item,
	const Vector3f& center,
	const float     radiusOuter,
	const float     radiusInner
)
{
	const float d2 = getDistanceSquared( item, center );

	return (d2 <  (radiusOuter * radiusOuter)) &
	       (d2 >= (radiusInner * radiusInner));
}


OctreeVisitor::ESelection OctreeVisitorPhoton::isCellOverlappingShell
(
	const OctreeBound& cellBound,
	const Vector3f&    center,
	const float        outerRadius,
	const float        innerRadius
)
{
	/// distance = sqrt( (boundCenter - center).dotSelf() )
	/// out =  distance >= (outerRadius + boundRadius) |
	///        distance <  (innerRadius - boundRadius)
	/// in  =  distance <  (outerRadius - boundRadius) &
	///       (distance >= (innerRadius + boundRadius) | innerRadius == 0 )


	const Vector3f& cellCenter      = cellBound.getCenter();
	const float     cellBoundRadius = cellBound.getRadius();

	const float dx = cellCenter.getX() - center.getX();
	const float dy = cellCenter.getY() - center.getY();
	const float dz = cellCenter.getZ() - center.getZ();

	const float d2 = (dx * dx) + (dy * dy) + (dz * dz);

	const float oPb  = outerRadius + cellBoundRadius;
	const float oPb2 = oPb * oPb;
	const float iMb  = innerRadius - cellBoundRadius;
	const float iMb2 = iMb * iMb;

	const bool isOut = (d2 >= oPb2) | ((d2 < iMb2) & (iMb > 0.0f));

	bool isIn = false;
	if( !isOut )
	{
		const float oMb  = outerRadius - cellBoundRadius;
		const float oMb2 = oMb * oMb;
		const float iPb  = innerRadius + cellBoundRadius;
		const float iPb2 = iPb * iPb;

		isIn = ((d2 < oMb2) & (oMb > 0.0f)) & ((d2 >= iPb2) | (innerRadius <= 0.0f));
	}

	///			in	!out
	/// out		0	0
	/// partial	0	1
	/// in		1	1
	return ESelection( (dword(isIn) << 1) | dword(!isOut) );
}


dword OctreeVisitorPhoton::getSubcellShellOverlaps
(
	const OctreeBound& cellBound,
	const Vector3f&    center,
	const float        outerRadius,
	const float        innerRadius
)
{
	Vector3f dxyzs[2];
	getComponentDistancesSquared( cellBound.getCenter(), cellBound.getSize(), center, dxyzs );

	const float subcellRadius = cellBound.getRadius() * 0.5f;

	const float opb    = outerRadius + subcellRadius;
	const float opbs   = opb * opb;
	const float imb    = innerRadius - subcellRadius;
	const float imbs   = imb * imb;
	const bool  imblez = (imb <= 0.0f);

	const float omb    = outerRadius - subcellRadius;
	const float ombs   = omb * omb;
	const float ipb    = innerRadius + subcellRadius;
	const float ipbs   = ipb * ipb;
	const bool  ombgz  = (omb > 0.0f);

	dword overlaps = 0;
	for( int i = 8;  i-- > 0; )
	{
		const float ds = dxyzs[i        & 1].getX() +
						 dxyzs[(i >> 1) & 1].getY() +
						 dxyzs[(i >> 2) & 1].getZ();

		const bool isNotOut = ((ds >= imbs) | imblez) &  (ds <  opbs);
		const bool isIn     = ((ds <  ombs) & ombgz ) & ((ds >= ipbs) | (innerRadius <= 0.0f));
		overlaps = (overlaps << 2) | ((dword(isIn) << 1) | dword(isNotOut));
	}

	return overlaps & 0xFFFF;
}


void OctreeVisitorPhoton::getComponentDistancesSquared
(
	const Vector3f& cellCenter,
	const float     cellSize,
	const Vector3f& position,
	Vector3f        dxyzs[2]
)
{
	const float dcx = cellCenter.getX() - position.getX();
	const float dcy = cellCenter.getY() - position.getY();
	const float dcz = cellCenter.getZ() - position.getZ();

	const float subcellHalfSize = cellSize * 0.25f;

	const float dx[2] = { (dcx - subcellHalfSize), (dcx + subcellHalfSize) };
	const float dy[2] = { (dcy - subcellHalfSize), (dcy + subcellHalfSize) };
	const float dz[2] = { (dcz - subcellHalfSize), (dcz + subcellHalfSize) };

	dxyzs[0].setXYZ( (dx[0] * dx[0]), (dy[0] * dy[0]), (dz[0] * dz[0]) );
	dxyzs[1].setXYZ( (dx[1] * dx[1]), (dy[1] * dy[1]), (dz[1] * dz[1]) );
}
