/*--------------------------------------------------------------------

   Perceptuum renderer, version 2.1
   Copyright (c) 2002-2003,  Harrison Ainsworth.

   www.hxa7241.org

--------------------------------------------------------------------*/

/*--------------------------------------------------------------------

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston,
   MA 02111-1307, USA

--------------------------------------------------------------------*/


#include "CieXyz.h"
#include "UnitVector3f.h"
#include "Polar.h"
#include "FpToInt.h"
#include "hxa7241math.h"

#include "Photon.h"   /// own header is included last


using namespace perceptuum2::transport;




/// check byte size
#define DESIRED_BYTE_SIZE 24
#if sizeof(Photon) > DESIRED_BYTE_SIZE
	#error size of perceptuum2::transport::Photon is larger than DESIRED_BYTE_SIZE
#endif




/// statics ----------------------------------------------------------------------------------------
float Photon::COS_THETA_STORE[256];
float Photon::SIN_THETA_STORE[256];
float Photon::COS_PHI_STORE[256];
float Photon::SIN_PHI_STORE[256];

const float*const Photon::COS_THETA = (initAngleTables( COS_THETA_STORE, SIN_THETA_STORE, COS_PHI_STORE, SIN_PHI_STORE ), COS_THETA_STORE);
const float*const Photon::SIN_THETA = SIN_THETA_STORE;
const float*const Photon::COS_PHI   = COS_PHI_STORE;
const float*const Photon::SIN_PHI   = SIN_PHI_STORE;




/// standard object services -----------------------------------------------------------------------
Photon::Photon()
{
//	Photon::set( CieXyz::ZERO(), Vector3f::ZERO(), UnitVector3f::ZERO(), 0 );

	power_m.set( 0 );
	position_m   = Vector3f::ZERO();
	theta_m      = 128;
	phi_m        = 0;
	previousId_m = 0;
}


//Photon::Photon
//(
//	const CieXyz&       power,
//	const Vector3f&     position,
//	const UnitVector3f& direction,
//	const dword         id
//)
//{
//	Photon::set( power, position, direction, id );
//}


//Photon::~Photon()
//{
//}


//Photon::Photon( const Photon& other )
//{
//	Photon::operator=( other );
//}


Photon& Photon::operator=( const Photon& other )
{
	power_m      = other.power_m;
	position_m   = other.position_m;
	theta_m      = other.theta_m;
	phi_m        = other.phi_m;
	previousId_m = other.previousId_m;

	return *this;
}




/// commands ---------------------------------------------------------------------------------------
void Photon::set
(
	const CieXyz&       power,
	const Vector3f&     position,
	const UnitVector3f& direction,
	const dword         id
)
{
	power_m    = power.toXyzE();
	position_m = position;

	{
		hxa7241::primitives::byte theta = 0;
		hxa7241::primitives::byte phi   = 0;
		compressDirection( direction, theta, phi );
		theta_m = ubyte(theta);
		phi_m   = ubyte(phi);
	}

	previousId_m = id;
}


void Photon::scalePower
(
	const float scaling
)
{
	CieXyz power( power_m );
	power *= scaling;

	power_m = power.toXyzE();
}




/// queries ----------------------------------------------------------------------------------------
void Photon::get
(
	CieXyz&       power,
	Vector3f&     position,
	UnitVector3f& direction,
	dword&        id
) const
{
	power.fromXyzE( power_m );
	position = position_m;
	decompressDirection( ubyte(theta_m), ubyte(phi_m), direction );
	id = previousId_m;
}


const Vector3f& Photon::getPosition() const
{
	return position_m;
}




/// implementation ---------------------------------------------------------------------------------
void Photon::initAngleTables
(
	float cosTheta[256],
	float sinTheta[256],
	float cosPhi[256],
	float sinPhi[256]
)
{
	static const float PI_OVER_256 = float(3.1415926535897932384626433832795 / 256.0);
	for( dword i = 256;  i-- > 0; )
	{
		const float theta = float(i) * PI_OVER_256;
		const float phi   = theta * 2.0f;

		hxa7241::graphics::sinAndcos( theta, sinTheta[i], cosTheta[i] );
		hxa7241::graphics::sinAndcos( phi,   sinPhi[i],   cosPhi[i]   );
	}
}


void Photon::compressDirection
(
	const UnitVector3f&        direction,
	hxa7241::primitives::byte& theta,
	hxa7241::primitives::byte& phi
)
{
	static const float _256_OVER_PI = float(256.0 / 3.1415926535897932384626433832795);
	using hxa7241::utility::fpToIntTrunc;

	{
		dword t = fpToIntTrunc( ::acos( direction.getZ() ) * _256_OVER_PI );
		if( t > 255 )
		{
			t = 255;
		}
		theta = hxa7241::primitives::byte(t);
	}

	{
		dword p = fpToIntTrunc( ::atan2( direction.getY(), direction.getX() ) * _256_OVER_PI * 0.5f );
		if( p > 255 )
		{
			p = 255;
		}
		else if( p < 0 )
		{
			p += 256;
		}
		phi = hxa7241::primitives::byte(p);
	}
}


void Photon::decompressDirection
(
	const hxa7241::primitives::byte theta,
	const hxa7241::primitives::byte phi,
	UnitVector3f&                   direction
)
{
	direction.setXYZ( SIN_THETA[ubyte(theta)] * COS_PHI[ubyte(phi)],
	                  SIN_THETA[ubyte(theta)] * SIN_PHI[ubyte(phi)],
	                  COS_THETA[ubyte(theta)] );
}
