/*--------------------------------------------------------------------

   Perceptuum renderer, version 2.1
   Copyright (c) 2002-2003,  Harrison Ainsworth.

   www.hxa7241.org

--------------------------------------------------------------------*/

/*--------------------------------------------------------------------

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston,
   MA 02111-1307, USA

--------------------------------------------------------------------*/


#ifndef PhotonMap_h
#define PhotonMap_h


#include "Primitives.h"
#include "Photon.h"
#include "Octree.h"
#include "LightStream.h"
#include <utility>
#include <vector>
#include <memory>
#include <functional>

#include "hxa7241_graphics.h"




#include "perceptuum2_transport.h"
namespace perceptuum2 { namespace transport {
	using namespace hxa7241::primitives;
	using perceptuum2::interaction::Light;
	using perceptuum2::interaction::LightStream;
	using hxa7241::graphics::Octree;
	using hxa7241::graphics::Vector3f;
	using hxa7241::graphics::UnitVector3f;
	using hxa7241::utility::Float01;
	using std::auto_ptr;
	using std::vector;
	class PhotonMapBuffers;


/**
 * store for photon interactions, providing range querys.<br/><br/>
 *
 * reference:
 * <cite>'realistic image synthesis using photon mapping' - jensen
 * (AKPeters 2001)</cite>
 *
 * @see PhotonMapBuffers
 *
 * @implementation
 * uses an octree rather than a kd-tree, because storage size is not so much
 * a concern, and an octree class is reusable for other graphics tasks.
 * <br/><br/>
 *
 * should refactor with the State pattern.
 */
class PhotonMap
{
/// standard object services -----------------------------------------------------------------------
public:
	explicit PhotonMap( PhotonMapBuffers* pBuffers,
	                    dword             capacity,
	                    bool              isLogging );
	         PhotonMap( PhotonMapBuffers* pBuffers,
	                    dword             capacity,
	                    dword             maxTreeLevels,
	                    dword             maxTreeCellItemCount,
	                    bool              isLogging );

	virtual ~PhotonMap();
private:
	         PhotonMap( const PhotonMap& );
	PhotonMap& operator=( const PhotonMap& );


/// commands ---------------------------------------------------------------------------------------
public:
	virtual void  switchToBuildingMode();
	virtual void  switchToIlluminatingMode();

	virtual void  addElement( const Light&,
	                          const Vector3f&     position,
	                          const UnitVector3f& direction,
	                          dword               id );
	virtual void  scale( float );


/// queries ----------------------------------------------------------------------------------------
public:
	virtual dword getCapacity()                                                               const;
	virtual dword getRemainingSpace()                                                         const;

	virtual const vector<LightStream>* getIllumination( const Vector3f& position )            const;

	static  bool  test( bool isLogging );


/// implementation ---------------------------------------------------------------------------------
protected:
	/// statics
	static  bool  getEstimateIlluminationsTest( const Vector3f&       center,
	                                            const vector<Photon>& photons,
	                                            PhotonMapBuffers*     pBuffers );
	static  bool  getEstimateIlluminations( const Vector3f&       center,
	                                        const Octree<Photon>* pOctree,
	                                        dword                 octreeMaxDepth,
	                                        dword                 octreeElementCount,
	                                        PhotonMapBuffers*     pBuffers,
	                                        dword&                visitCount,
	                                        dword&                shrinkCount );
	static  void  addSelectedPhotonsToIlluminations( PhotonMapBuffers* pBuffers );
	static  void  filterIlluminations( const vector<float>& distance2Buffer,
	                                   vector<LightStream>& illuminationBuffer );

	typedef std::pair<float, const Photon*> SelectedItem;

	/// for sorting in addSelectedPhotonsToIlluminations
	class SelectedComparer : public std::binary_function<SelectedItem, SelectedItem, bool>
	{
	public:
		SelectedComparer() {};

		bool operator()( const SelectedItem& p0, const SelectedItem& p1 ) const
		{
			return p0.first < p1.first;
		}
	};


/// fields -----------------------------------------------------------------------------------------
private:
	const dword maxOctreeLevels_m;
	const dword maxItemCountPerOctreeCell_m;

	/// the photons and stats
	vector<Photon> photons_m;
	Vector3f       minPosition_m;
	Vector3f       maxPosition_m;

	/// the spatial index and stats
	auto_ptr< Octree<Photon> > pOctree_m;
	dword                      octreeMaxDepth_m;
	dword                      octreeElementCount_m;

	/// external shared illumination estimating buffers
	PhotonMapBuffers* pBuffers_m;

	/// measurements
	mutable dword totalEstimationCount_m;
	mutable dword totalVisitCount_m;
	mutable dword totalShrinkCount_m;
	mutable dword maxVisitCount_m;
	mutable dword maxShrinkCount_m;

	bool isLogging_m;

	/// constants
	static const dword MAX_TREE_LEVELS;
	static const dword MAX_TREE_CELL_ITEM_COUNT;
	static const dword ESTIMATION_PROPORTION;
	static const dword ESTIMATION_SIZE;
};








class PhotonMapBuffers
{
/// standard object services -----------------------------------------------------------------------
public:
	         PhotonMapBuffers();

	        ~PhotonMapBuffers();
private:
	         PhotonMapBuffers( const PhotonMapBuffers& );
	PhotonMapBuffers& operator=( const PhotonMapBuffers& );


/// commands ---------------------------------------------------------------------------------------
public:
	typedef std::pair<float, const Photon*> SelectedItem;

	        void  reserve( dword );
	        void  resize( dword );

	        void  pushBack( const LightStream&,
	                        float distance2,
	                        dword id );
	        void  setElement( dword index,
	                          const LightStream&,
	                          float distance2,
	                          dword id );

	        vector<LightStream>&  getIlluminationBuffer();
	        vector<float>&        getDistance2Buffer();
	        vector<dword>&        getIdBuffer();

	        vector<SelectedItem>& getSelectedBuffer();


/// queries ----------------------------------------------------------------------------------------
public:
	        dword  getCapacity()                                                              const;
	        dword  getSize()                                                                  const;
	        dword  getSpare()                                                                 const;
	        bool   isFull()                                                                   const;

	        const vector<LightStream>&  getIlluminationBuffer()                               const;
	        const vector<float>&        getDistance2Buffer()                                  const;
	        const vector<dword>&        getIdBuffer()                                         const;

	        const vector<SelectedItem>& getSelectedBuffer()                                   const;


/// fields -----------------------------------------------------------------------------------------
private:
	// hmmm, use hxa7241::utility::Vector instead of std::vector...

	vector<LightStream>  illuminationBuffer_m;
	vector<float>        distance2Buffer_m;
	vector<dword>        idBuffer_m;

	vector<SelectedItem> selectedBuffer_m;
};




/// commands ---------------------------------------------------------------------------------------
inline
vector<LightStream>& PhotonMapBuffers::getIlluminationBuffer()
{
	return illuminationBuffer_m;
}


inline
vector<float>& PhotonMapBuffers::getDistance2Buffer()
{
	return distance2Buffer_m;
}


inline
vector<dword>& PhotonMapBuffers::getIdBuffer()
{
	return idBuffer_m;
}


inline
vector<PhotonMapBuffers::SelectedItem>& PhotonMapBuffers::getSelectedBuffer()
{
	return selectedBuffer_m;
}




/// queries ----------------------------------------------------------------------------------------
inline
dword PhotonMapBuffers::getCapacity() const
{
	return illuminationBuffer_m.capacity();
}


inline
dword PhotonMapBuffers::getSize() const
{
	return illuminationBuffer_m.size();
}


inline
bool PhotonMapBuffers::isFull() const
{
	return getSpare() == 0;
}


inline
const vector<LightStream>& PhotonMapBuffers::getIlluminationBuffer() const
{
	return illuminationBuffer_m;
}


inline
const vector<float>& PhotonMapBuffers::getDistance2Buffer() const
{
	return distance2Buffer_m;
}


inline
const vector<dword>& PhotonMapBuffers::getIdBuffer() const
{
	return idBuffer_m;
}


inline
const vector<PhotonMapBuffers::SelectedItem>& PhotonMapBuffers::getSelectedBuffer() const
{
	return selectedBuffer_m;
}


}}




#endif//PhotonMap_h
