/*--------------------------------------------------------------------

   Perceptuum renderer, version 2.1
   Copyright (c) 2002-2003,  Harrison Ainsworth.

   www.hxa7241.org

--------------------------------------------------------------------*/

/*--------------------------------------------------------------------

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston,
   MA 02111-1307, USA

--------------------------------------------------------------------*/


#include "Serialization.h"
#include "SamplingControl.h"
#include "Polar.h"
#include "Matrix4f.h"
#include "Vector4f.h"
#include "Float01.h"
#include "hxa7241math.h"

#include "Projection.h"   /// own header is included last


using namespace perceptuum2::transport;




/// statics
const UnitVector3f Projection::LOOKDIRECTION_DEFAULT( UnitVector3f::Z() );
const UnitVector3f Projection::UPDIRECTION_DEFAULT  ( UnitVector3f::Y() );
const float        Projection::WIDTH_DEFAULT          = 0.285f;
const float        Projection::SCREENDISTANCE_DEFAULT = 0.5f;
const float        Projection::LENSDIAMETER_DEFAULT   = 0.010f;
const float        Projection::FOCALLENGTH_DEFAULT    = 4.000f;
const float        Projection::Z_NEAR_CLIP_PLANE      = 0.01f;




/// standard object services -----------------------------------------------------------------------
Projection::Projection()
 :	eyePosition_m(),
	lookDirection_m(),
	upDirection_m(),
	width_m(),
	screenDistance_m(),
	lensDiameter_m(),
	focalLength_m(),
	screenCenterVector_m(),
	sideDirection_m()
{
	setToDefault();
}


Projection::~Projection()
{
}


Projection::Projection( const Projection& other )
{
	Projection::operator=( other );
}


Projection& Projection::operator=( const Projection& other )
{
	if( &other != this )
	{
		eyePosition_m        = other.eyePosition_m;
		lookDirection_m      = other.lookDirection_m;
		upDirection_m        = other.upDirection_m;
		width_m              = other.width_m;
		screenDistance_m     = other.screenDistance_m;
		lensDiameter_m       = other.lensDiameter_m;
		focalLength_m        = other.focalLength_m;
		screenCenterVector_m = other.screenCenterVector_m;
		sideDirection_m      = other.sideDirection_m;
	}

	return *this;
}




/// commands ---------------------------------------------------------------------------------------
void Projection::setToDefault()
{
	eyePosition_m    = Vector3f::ZERO();
	lookDirection_m  = LOOKDIRECTION_DEFAULT;
	upDirection_m    = UPDIRECTION_DEFAULT;
	width_m          = WIDTH_DEFAULT;
	screenDistance_m = SCREENDISTANCE_DEFAULT;
	lensDiameter_m   = LENSDIAMETER_DEFAULT;
	focalLength_m    = FOCALLENGTH_DEFAULT;

	setPreCalculations();
}




/// queries ----------------------------------------------------------------------------------------
	        /*Vector3f&     getEyePosition( Vector3f& )       const;
	        UnitVector3f& getLookDirection( UnitVector3f& ) const;
	        UnitVector3f& getUpDirection( UnitVector3f& )   const;
	        UnitVector3f& getSideDirection( UnitVector3f& ) const;
	        float         getWidth()                        const;
	        float         getScreenDistance()               const;
	        float         getLensDiameter()                 const;
	        float         getFocalLength()                  const;*/

/*Vector3f& Projection::getEyePosition( Vector3f& eyePosition ) const
{
	return eyePosition = eyePosition_m;
}


UnitVector3f& Projection::getLookDirection( UnitVector3f& lookDirection ) const
{
	return lookDirection = lookDirection_m;
}


UnitVector3f& Projection::getUpDirection( UnitVector3f& upDirection ) const
{
	return upDirection = upDirection_m;
}


UnitVector3f& Projection::getSideDirection( UnitVector3f& sideDirection ) const
{
	return sideDirection = sideDirection_m;
}


float Projection::getWidth() const
{
	return width_m;
}


float Projection::getScreenDistance() const
{
	return screenDistance_m;
}


float Projection::getLensDiameter() const
{
	return lensDiameter_m;
}


float Projection::getFocalLength() const
{
	return focalLength_m;
}*/


void Projection::getRay
(
	const float            x,
	const float            y,
	const SamplingControl& samplingControl,
	Vector3f&              start,
	UnitVector3f&          direction
) const
{
	/// get lens displacement
	Vector3f lensDisplacementVector;
	{
		/// get 2d lens position
		float lensX;
		float lensY;
		{
			/// get 2d sample point 0-1, 0-1
			using hxa7241::utility::Float01;
			const Vector2<Float01> samplePoint( samplingControl.getLensPoint() );

			hxa7241::graphics::squareToDisk( samplePoint.getX(), samplePoint.getY(), lensX, lensY );
			lensX *= (lensDiameter_m * 0.5f);
			lensY *= (lensDiameter_m * 0.5f);
		}

		/// convert lens position to position in space
		lensDisplacementVector  = sideDirection_m.getVector() * lensX;
		lensDisplacementVector += upDirection_m.getVector()   * lensY;
	}

	/// get focal point direction
	Vector3f focalVector;
	{
		/// point on screen
		focalVector = screenCenterVector_m;
		focalVector += (sideDirection_m.getVector() * (x * width_m));
		focalVector += (upDirection_m.getVector()   * (y * width_m));

		/// focal point
		focalVector.unitizeEq();
		focalVector *= focalLength_m;
	}


	/// set ray start and direction
	{
		start  = eyePosition_m;
		start += lensDisplacementVector;

		/// lens point to focal point direction
		direction = focalVector - lensDisplacementVector;
	}
}


void Projection::getWorldToViewTransform
(
	const SamplingControl& samplingControl,
	const Vector2f&        subPixelPosition,
	Matrix4f&              matrix
) const
{
	/// get defining vectors
	Vector3f     eyePosition;
	UnitVector3f lookDirection;
	UnitVector3f upDirection( upDirection_m );
	getRay( 0.0f, 0.0f, samplingControl, eyePosition, lookDirection );

	/// adjust eye position with sub pixel position
	eyePosition += Vector3f( subPixelPosition.getX(), subPixelPosition.getY(), 0.0f );

	/// make transformation matrix
	{
		/// see: ch 8, 'jim blinns corner, a trip down the graphics pipeline' - blinn (morgan kaufmann 1996)

		const float lookDotUp = lookDirection.dot( upDirection );
		float oneOverVy;// = 0.0f;
		{
			float oneMinusDot2 = 1.0f - (lookDotUp * lookDotUp);
			if( oneMinusDot2 < 0.0f )
			{
				oneMinusDot2 = 0.0f;
			}
			const float vy = float(::sqrt( oneMinusDot2 ));
			oneOverVy = 1.0f / vy;
		}

		/// define rotation 3x3 matrix
		const Vector3f  col0( upDirection.getVector().cross( lookDirection.getVector() ) *= oneOverVy );
		const Vector3f  col1( ((lookDirection.getVector() * -lookDotUp) += upDirection.getVector()) *= oneOverVy );
		const Vector3f& col2( lookDirection.getVector() );

		/// extend with translation for making a 4x4 matrix
		using hxa7241::graphics::Vector4f;
		const Vector4f  columns[] = { Vector4f( col0, -(eyePosition.getX()) ),
									  Vector4f( col1, -(eyePosition.getY()) ),
									  Vector4f( col2, -(eyePosition.getZ()) ),
									  Vector4f::W() };

		matrix.setColumns( columns );
	}
}


void Projection::getViewToClipTransform
(
	const SamplingControl& ,//samplingControl,
	const Vector2f&        oneOverWidthHeight,
	Matrix4f&              matrix
) const
{
	/// perspective
	const float viewAngleDegrees = hxa7241::graphics::radiansToDegrees( float(::atan( (width_m * 0.5f) / screenDistance_m )) );
	const float zNear            = Z_NEAR_CLIP_PLANE;
	const float zFar             = FLOAT_MAX;
	matrix.setToPerspective( viewAngleDegrees, zNear, zFar );


	/// scale to square screen
	const Matrix4f scaling( Matrix4f::SCALE, Vector3f( oneOverWidthHeight.getX(), oneOverWidthHeight.getY(), 1.0f ) );
	matrix.multiplyEqPost( scaling );


	/// translate into 0-1 cube
	const Matrix4f translation( Matrix4f::TRANSLATE, Vector3f( 0.5f, 0.5f, 0.0f ) );
	matrix.multiplyEqPost( translation );
}




/// implementation ---------------------------------------------------------------------------------
void Projection::enforceConstraints()
{
	/// make direction vectors default if zero
	if( lookDirection_m == UnitVector3f::ZERO() )
	{
		lookDirection_m = LOOKDIRECTION_DEFAULT;
	}

	if( upDirection_m == UnitVector3f::ZERO() )
	{
		upDirection_m = UPDIRECTION_DEFAULT;
	}


	/// make direction vectors orthogonal
	upDirection_m.orthogonalizeEq( lookDirection_m );


	width_m          = width_m          > 0.0f ? width_m          : WIDTH_DEFAULT;
	screenDistance_m = screenDistance_m > 0.0f ? screenDistance_m : SCREENDISTANCE_DEFAULT;
	lensDiameter_m   = lensDiameter_m   > 0.0f ? lensDiameter_m   : 0.0f;
	focalLength_m    = focalLength_m    > 0.0f ? focalLength_m    : FOCALLENGTH_DEFAULT;
}


void Projection::setPreCalculations()
{
	screenCenterVector_m  = lookDirection_m.getVector();
	screenCenterVector_m *= screenDistance_m;

	sideDirection_m = upDirection_m;
	sideDirection_m.crossEq( lookDirection_m );
}




/// serialization ----------------------------------------------------------------------------------
SerializerOut& perceptuum2::transport::operator<<( SerializerOut& out, const Projection& obj )
{
	out.write( "eyePosition"   , obj.eyePosition_m   , "point in space where the eye is" );
	out.write( "lookDirection" , obj.lookDirection_m , "the direction vector along which the eye is looking" );
	out.write( "upDirection"   , obj.upDirection_m   , "the direction that defines upright" );
	out.write( "width"         , obj.width_m         , "width of the screen in scene units" );
	out.write( "screenDistance", obj.screenDistance_m, "distance of the screen from the lens in scene units" );
	out.write( "lensDiameter"  , obj.lensDiameter_m  , "diameter of the lens in scene units" );
	out.write( "focalLength"   , obj.focalLength_m   , "focal length of the lens in scene units" );

	return out;
}


SerializerIn& perceptuum2::transport::operator>>( SerializerIn& in, Projection& obj )
{
	in.read( "eyePosition"   , obj.eyePosition_m    );
	in.read( "lookDirection" , obj.lookDirection_m  );
	in.read( "upDirection"   , obj.upDirection_m    );
	in.read( "width"         , obj.width_m          );
	in.read( "screenDistance", obj.screenDistance_m );
	in.read( "lensDiameter"  , obj.lensDiameter_m   );
	in.read( "focalLength"   , obj.focalLength_m    );

	obj.enforceConstraints();
	obj.setPreCalculations();

	return in;
}


const char* Projection::CLASS_NAME()
{
	static const char CLASS_NAME[] = "perceptuum2::transport::Projection";
	return CLASS_NAME;
}


const char* Projection::CLASS_DESCRIPTION()
{
	static const char CLASS_DESCRIPTION[] = "defines the point of view for a rendering";
	return CLASS_DESCRIPTION;
}
