/*--------------------------------------------------------------------

   Perceptuum renderer, version 2.1
   Copyright (c) 2002-2003,  Harrison Ainsworth.

   www.hxa7241.org

--------------------------------------------------------------------*/

/*--------------------------------------------------------------------

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston,
   MA 02111-1307, USA

--------------------------------------------------------------------*/


#include "Vector2.h"
#include "Float01.h"
#include "Serialization.h"
#include "Logger.h"

#include "SamplingControl.h"   /// own header is included last


using namespace perceptuum2::transport;




//#define ENABLE_TEST_CODE


/// statics
/// (values chosen because their plots look ok)
const dword SamplingControl::HALTON_BASES[]             = { 0, 1, 4, 2, 3 };

const dword SamplingControl::SOBOL_DIMENSIONS_EMIT_00[] = {  2,  3,  4,  5,  6 };
const dword SamplingControl::SOBOL_DIMENSIONS_EMIT_01[] = { 22, 23, 25, 35 };
const dword SamplingControl::RANDOM_SEEDS_EMIT_0[]      = { 29, 41, 53, 89 };
const dword SamplingControl::SPECIAL_LEVELS_EMIT_0      = 1;

const dword SamplingControl::SOBOL_DIMENSIONS_EMIT_10[] = {  9, 10,  7, 11, 12 };
const dword SamplingControl::SOBOL_DIMENSIONS_EMIT_11[] = { 26, 27, 28, 36, 33, 34, 31, 37 };
const dword SamplingControl::RANDOM_SEEDS_EMIT_1[]      = {  3, 11, 19, 71, 31, 43, 59, 79 };
const dword SamplingControl::SPECIAL_LEVELS_EMIT_1      = 2;

const dword SamplingControl::SOBOL_DIMENSIONS_EYE[]     = { 14, 15, 13, 38, 20, 21, 19, 39 };
const dword SamplingControl::RANDOM_SEEDS_EYE[]         = {  5, 13, 23, 73, 37, 47, 61, 83 };
const dword SamplingControl::SPECIAL_LEVELS_EYE         = 2;

const dword SamplingControl::SOBOL_DIMENSIONS_DIFFUSE[] = { 17, 18, 16, 32 };
const dword SamplingControl::RANDOM_SEEDS_DIFFUSE[]     = {  2,  7, 17, 67 };
const dword SamplingControl::SPECIAL_LEVELS_DIFFUSE     = 1;

const dword SamplingControl::SEEDS[]                    = { 0x00701C07, 0x0380E038, 0x1C0701C0, 0xE0380E00 };




/// standard object services -----------------------------------------------------------------------
SamplingControl::SamplingControl()
 :	emitterControl0_m( SOBOL_DIMENSIONS_EMIT_00, SOBOL_DIMENSIONS_EMIT_01, RANDOM_SEEDS_EMIT_0, SPECIAL_LEVELS_EMIT_0, SEEDS[0] ),
	emitterControl1_m( SOBOL_DIMENSIONS_EMIT_10, SOBOL_DIMENSIONS_EMIT_11, RANDOM_SEEDS_EMIT_1, SPECIAL_LEVELS_EMIT_1, SEEDS[1] ),
	eyePathControl_m ( SOBOL_DIMENSIONS_EYE,     RANDOM_SEEDS_EYE,     SPECIAL_LEVELS_EYE,     SEEDS[2] ),
	diffuseControl_m ( SOBOL_DIMENSIONS_DIFFUSE )
{
	for( dword i = sizeof(haltons_m)/sizeof(*haltons_m);  i-- > 0; )
	{
		haltons_m[i].setBase( HALTON_BASES[i] );
	}
}


SamplingControl::~SamplingControl()
{
}


SamplingControl::SamplingControl( const SamplingControl& other )
 :	emitterControl0_m( other.emitterControl0_m ),
	emitterControl1_m( other.emitterControl1_m ),
	eyePathControl_m ( other.eyePathControl_m  ),
	diffuseControl_m ( other.diffuseControl_m )
{
	for( dword i = sizeof(haltons_m)/sizeof(*haltons_m);  i-- > 0; )
	{
		haltons_m[i] = other.haltons_m[i];
	}
}


SamplingControl& SamplingControl::operator=( const SamplingControl& other )
{
	assign( other );

	return *this;
}




/// commands ---------------------------------------------------------------------------------------
void SamplingControl::advanceFrame()
{
	for( dword i = sizeof(haltons_m)/sizeof(*haltons_m);  i-- > 0; )
	{
		haltons_m[i].next();
	}

	eyePathControl_m.nextPath();
	diffuseControl_m.advance();
}


void SamplingControl::advanceIllumination0()
{
	emitterControl0_m.advance();
}


void SamplingControl::advanceIllumination1()
{
	emitterControl1_m.advance();
}




/// queries ----------------------------------------------------------------------------------------
Float01 SamplingControl::getTimePoint() const
{
	return Float01( haltons_m[TIME].getItem() );
}


Vector2<Float01> SamplingControl::getPixelPoint() const
{
	return Vector2<Float01>( haltons_m[PIXEL_X].getItem(),
	                         haltons_m[PIXEL_Y].getItem() );
}


Vector2<Float01> SamplingControl::getLensPoint() const
{
	return Vector2<Float01>( haltons_m[LENS_X].getItem(),
	                         haltons_m[LENS_Y].getItem() );
}


const SamplingControlEmitter& SamplingControl::getEmitterControl0() const
{
	return emitterControl0_m;
}


const SamplingControlEmitter& SamplingControl::getEmitterControl1() const
{
	return emitterControl1_m;
}


const SamplingControlPath& SamplingControl::getEyePathControl() const
{
	return eyePathControl_m;
}


const SamplingControlDiffuse& SamplingControl::getDiffuseControl() const
{
	return diffuseControl_m;
}


dword SamplingControl::getFrameIndex() const
{
	return haltons_m[TIME].getIndex();
}




/// implementation ---------------------------------------------------------------------------------
void SamplingControl::assign
(
	const SamplingControl& other
)
{
	if( &other != this )
	{
		for( dword i = sizeof(haltons_m)/sizeof(*haltons_m);  i-- > 0; )
		{
			haltons_m[i] = other.haltons_m[i];
		}

		emitterControl0_m = other.emitterControl0_m;
		emitterControl1_m = other.emitterControl1_m;
		eyePathControl_m  = other.eyePathControl_m;
		diffuseControl_m  = other.diffuseControl_m;
	}
}




/// serialization ----------------------------------------------------------------------------------
SerializerOut& perceptuum2::transport::operator<<( SerializerOut& out, const SamplingControl& obj )
{
	for( dword i = 0, end = sizeof(obj.haltons_m)/sizeof(*obj.haltons_m);  i < end;  ++i )
	{
		out.write( "halton", obj.haltons_m[i].getIndex() );
	}

	out.write( "emitterControl0", obj.emitterControl0_m );
	out.write( "emitterControl1", obj.emitterControl1_m );
	out.write( "eyePathControl",  obj.eyePathControl_m  );
	out.write( "diffuseControl",  obj.diffuseControl_m );

	return out;
}


SerializerIn& perceptuum2::transport::operator>>( SerializerIn& in, SamplingControl& obj )
{
	for( dword i = 0, end = sizeof(obj.haltons_m)/sizeof(*obj.haltons_m);  i < end;  ++i )
	{
		dword index = 0;
		in.read( "halton", index );

		obj.haltons_m[i].setBase( obj.HALTON_BASES[i] );
		obj.haltons_m[i].setIndex( index );
	}

	in.read( "emitterControl0", obj.emitterControl0_m );
	in.read( "emitterControl1", obj.emitterControl1_m );
	in.read( "eyePathControl",  obj.eyePathControl_m  );
	in.read( "diffuseControl",  obj.diffuseControl_m );

	return in;
}


const char* SamplingControl::CLASS_NAME()
{
	static const char CLASS_NAME[] = "perceptuum2::transport::SamplingControl";
	return CLASS_NAME;
}


const char* SamplingControl::CLASS_DESCRIPTION()
{
	static const char CLASS_DESCRIPTION[] = "grouping of all stochastic sequences used in sampling";
	return CLASS_DESCRIPTION;
}








/// testing ----------------------------------------------------------------------------------------
#ifndef ENABLE_TEST_CODE
bool SamplingControl::test
(
	const bool isLogging
)
{
	if(isLogging) hxa7241::utility::Logger() << "SamplingControl::test  [disabled]";

	return false;
}


#else
#include "Vector3f.h"
#include "ByteSerializerOut.h"
#include "ByteSerializerIn.h"
#include <set>

dword getCrc( const hxa7241::primitives::byte* pBytes,
              const dword                      bytesLength );

bool SamplingControl::test
(
	const bool isLogging
)
{
	bool isSucceeded = true;

	static const char PREFIX[] = "SamplingControl::test  ";
	if(isLogging) hxa7241::utility::Logger() << PREFIX << "=== general test ===";


	/// test serialization and copying
	{
		bool isSerializationCopyingOk = true;


		SamplingControl sc0;

		dword i = 0;
		for( ;  i < 10;  ++i )
		{
			sc0.advanceFrame();
			sc0.advanceIllumination0();
			sc0.advanceIllumination0();
			sc0.advanceIllumination1();
			sc0.advanceIllumination1();
		}

		static const char fileName[] = "SamplingControlTestFile.tmp";
		hxa7241::utility::ByteSerializerOut::serializeOut( sc0, fileName );

		SamplingControl sc1;
		hxa7241::utility::ByteSerializerIn::serializeIn( sc1, fileName );

		SamplingControl sc2( sc1 );

		for( ;  i < 20;  ++i )
		{
			isSerializationCopyingOk &= (sc0.getTimePoint() == sc2.getTimePoint());
			isSerializationCopyingOk &= (sc0.getPixelPoint() == sc2.getPixelPoint());
			isSerializationCopyingOk &= (sc0.getLensPoint() == sc2.getLensPoint());

			for( dword j = 0;  j < 2;  ++j )
			{
				const SamplingControlEmitter& sce0 = j == 0 ? sc0.getEmitterControl0() : sc0.getEmitterControl1();
				const SamplingControlEmitter& sce2 = j == 0 ? sc2.getEmitterControl0() : sc2.getEmitterControl1();

				isSerializationCopyingOk &= (sce0.getPositionCoords() == sce2.getPositionCoords());
				isSerializationCopyingOk &= (sce0.getDirectionCoords() == sce2.getDirectionCoords());

				for( dword pathNodeIndex = 0;  pathNodeIndex < 4;  ++pathNodeIndex )
				{
					Float01 coords0[SamplingControlPath::COORDS_COUNT];
					Float01 coords2[SamplingControlPath::COORDS_COUNT];

					sce0.getInteractionCoords( pathNodeIndex, coords0 );
					sce2.getInteractionCoords( pathNodeIndex, coords2 );

					for( dword k = SamplingControlPath::COORDS_COUNT;  k-- > 0; )
					{
						isSerializationCopyingOk &= (coords0[k] == coords2[k]);
					}
				}
			}

			const SamplingControlPath& scp0 = sc0.getEyePathControl();
			const SamplingControlPath& scp2 = sc2.getEyePathControl();
			for( dword pathNodeIndex = 0;  pathNodeIndex < 4;  ++pathNodeIndex )
			{
				Float01 coords0[SamplingControlPath::COORDS_COUNT];
				Float01 coords2[SamplingControlPath::COORDS_COUNT];

				scp0.getSampleCoords( pathNodeIndex, coords0 );
				scp2.getSampleCoords( pathNodeIndex, coords2 );

				for( dword k = SamplingControlPath::COORDS_COUNT;  k-- > 0; )
				{
					isSerializationCopyingOk &= (coords0[k] == coords2[k]);
				}
			}

			const SamplingControlDiffuse& scdp0 = sc0.getDiffuseControl();
			const SamplingControlDiffuse& scdp2 = sc2.getDiffuseControl();
			for( dword index = 0;  index < 4;  ++index )
			{
				Float01 coords0[SamplingControlDiffuse::COORDS_COUNT];
				Float01 coords2[SamplingControlDiffuse::COORDS_COUNT];

				scdp0.getSampleCoords( index, coords0 );
				scdp2.getSampleCoords( index, coords2 );

				for( dword k = SamplingControlDiffuse::COORDS_COUNT;  k-- > 0; )
				{
					isSerializationCopyingOk &= (coords0[k] == coords2[k]);
				}
			}


			sc0.advanceFrame();
			sc0.advanceIllumination0();
			sc0.advanceIllumination0();
			sc0.advanceIllumination1();
			sc0.advanceIllumination1();

			sc2.advanceFrame();
			sc2.advanceIllumination0();
			sc2.advanceIllumination0();
			sc2.advanceIllumination1();
			sc2.advanceIllumination1();
		}


		if(isLogging) hxa7241::utility::Logger() << PREFIX << (isSerializationCopyingOk ? "--- serialization/copy test succeeded" : "*** serialization/copy test failed");

		isSucceeded &= isSerializationCopyingOk;
	}


	/// test distinctness
	{
		bool isDistinctnessOk = true;


		static const dword LEN = 32;

		vector<float> time( LEN );
		vector<float> pixelx( LEN );
		vector<float> pixely( LEN );
		vector<float> lensx( LEN );
		vector<float> lensy( LEN );
		vector<float> emitposx[2];
		vector<float> emitposy[2];
		vector<float> emitposz[2];
		vector<float> emitdirx[2];
		vector<float> emitdiry[2];
		vector<float> emit0path[32];
		vector<float> emit1path[32];
		vector<float>* emitpath[2] = { emit0path, emit1path };
		for( dword i = 2;  i-- > 0; )
		{
			emitposx[i].resize( LEN );  emitposy[i].resize( LEN );  emitposz[i].resize( LEN );  emitdirx[i].resize( LEN );  emitdiry[i].resize( LEN );
			for( dword j = 32;  j-- > 0; )
			{
				emitpath[i][j].resize( LEN );
			}
		}
		vector<float> eyepath[32];
		for( dword j = 32;  j-- > 0; )
		{
			eyepath[j].resize( LEN );
		}
		vector<float> diffuse[SamplingControlDiffuse::COORDS_COUNT];
		for( dword j = SamplingControlDiffuse::COORDS_COUNT;  j-- > 0; )
		{
			diffuse[j].resize( LEN );
		}

		SamplingControl sc;


		dword i = 0;
		for( ;  i < LEN;  ++i )
		{
			time[i]   = sc.getTimePoint();
			pixelx[i] = sc.getPixelPoint().getX();
			pixely[i] = sc.getPixelPoint().getY();
			lensx[i]  = sc.getLensPoint().getX();
			lensy[i]  = sc.getLensPoint().getY();

			for( dword j = 0;  j < 2;  ++j )
			{
				const SamplingControlEmitter& sce = j == 0 ? sc.getEmitterControl0() : sc.getEmitterControl1();

				emitposx[j][i] = sce.getPositionCoords().getX();
				emitposy[j][i] = sce.getPositionCoords().getY();
				emitposz[j][i] = sce.getPositionCoords().getZ();
				emitdirx[j][i] = sce.getDirectionCoords().getX();
				emitdiry[j][i] = sce.getDirectionCoords().getY();

				for( dword pathNodeIndex = 0;  pathNodeIndex < 4;  ++pathNodeIndex )
				{
					Float01 coords[SamplingControlPath::COORDS_COUNT];
					sce.getInteractionCoords( pathNodeIndex, coords );

					for( dword c = 0;  c < SamplingControlPath::COORDS_COUNT;  ++c )
					{
						emitpath[j][(pathNodeIndex * SamplingControlPath::COORDS_COUNT) + c][i] = coords[c];
					}
				}
			}

			const SamplingControlPath& scp = sc.getEyePathControl();
			for( dword pathNodeIndex = 0;  pathNodeIndex < 4;  ++pathNodeIndex )
			{
				Float01 coords[SamplingControlPath::COORDS_COUNT];
				scp.getSampleCoords( pathNodeIndex, coords );

				for( dword c = 0;  c < SamplingControlPath::COORDS_COUNT;  ++c )
				{
					eyepath[(pathNodeIndex * SamplingControlPath::COORDS_COUNT) + c][i] = coords[c];
				}
			}

			if( (i & 0x03) == 0 )
			{
				const SamplingControlDiffuse& scdp = sc.getDiffuseControl();
				for( dword index = 0;  index < 4;  ++index )
				{
					Float01 coords[SamplingControlDiffuse::COORDS_COUNT];
					scdp.getSampleCoords( index, coords );

					for( dword c = 0;  c < SamplingControlDiffuse::COORDS_COUNT;  ++c )
					{
						diffuse[c][i + index] = coords[c];
					}
				}
			}

			sc.advanceFrame();
			sc.advanceIllumination0();
			sc.advanceIllumination1();
		}

		{
			/// make vector of all sequence start pointers
			vector<float*> vp;
			vp.push_back( &(time[0]) );
			vp.push_back( &(pixelx[0]) );
			vp.push_back( &(pixely[0]) );
			vp.push_back( &(lensx[0]) );
			vp.push_back( &(lensy[0]) );
			for( dword i = 2;  i-- > 0; )
			{
				vp.push_back( &(emitposx[i][0]) );
				vp.push_back( &(emitposy[i][0]) );
				vp.push_back( &(emitposz[i][0]) );
				vp.push_back( &(emitdirx[i][0]) );
				vp.push_back( &(emitdiry[i][0]) );
				for( dword j = 32;  j-- > 0; )
				{
					vp.push_back( &(emitpath[i][j][0]) );
				}
			}
			for( dword j = 32;  j-- > 0; )
			{
				vp.push_back( &(eyepath[j][0]) );
			}
			for( dword j = SamplingControlDiffuse::COORDS_COUNT;  j-- > 0; )
			{
				vp.push_back( &(diffuse[j][0]) );
			}

			/// calculate crc for each sequence
			vector<dword> crcs( vp.size() );
			for( dword i = 0;  i < dword(vp.size());  ++i )
			{
				crcs[i] = ::getCrc( reinterpret_cast<const byte*>( vp[i] ), (LEN * sizeof(float)) );
			}

			/// display sequences
			if( isLogging )
			{
				for( dword i = 0;  i < LEN;  ++i )
				{
					hxa7241::utility::Logger lll;
					lll << PREFIX << "points: " << i << "  ";
					for( dword j = 0;  j < dword(vp.size());  ++j )
					{
						lll << vp[j][i] << " ";
					}
				}

				{
					hxa7241::utility::Logger lll;
					lll << PREFIX << "crcs:       ";
					for( dword i = 0;  i < dword(vp.size());  ++i )
					{
						lll << (void*)(crcs[i]) << " ";
					}
				}

				hxa7241::utility::Logger() << PREFIX << vp.size() << " sequences";
			}

			/// test each sequences own members are distinct
			for( dword i = 0;  i < dword(vp.size());  ++i )
			{
				bool isItemsDistinct = true;

				std::set<dword> itemsSet;
				const float* pItems = vp[i];

				for( dword j = 0;  j < LEN;  ++j )
				{
					const bool isInserted = itemsSet.insert( *(reinterpret_cast<const dword*>( &(pItems[j]) )) ).second;
					isItemsDistinct &= isInserted;
				}

				if(isLogging & !isItemsDistinct) hxa7241::utility::Logger() << PREFIX << "nondistinct items in sequence: " << i;

				isDistinctnessOk &= isItemsDistinct;
			}

			/// test all sequences are distinct
			{
				bool isSequenceMatch = false;
				for( dword a = 0;  a < dword(crcs.size());  ++a )
				{
					for( dword b = 0;  b < dword(crcs.size());  ++b )
					{
						const bool isMatch = (a < b) & (crcs[a] == crcs[b]);
						if(isLogging & isMatch) hxa7241::utility::Logger() << PREFIX << "nondistinct sequences: " << a << " " << b;
						isSequenceMatch |= isMatch;
					}
				}

				isDistinctnessOk &= !isSequenceMatch;
			}
		}


		if(isLogging) hxa7241::utility::Logger() << PREFIX << (isDistinctnessOk ? "--- distinctness test succeeded" : "*** distinctness test failed");

		isSucceeded &= isDistinctnessOk;
	}


	if(isLogging) hxa7241::utility::Logger() << PREFIX  << (isSucceeded ? "--- general test succeeded" : "*** general test failed");

	return isSucceeded;
}




/// CRC stuff from: 'numerical recipes in c' p900, www.nr.com
unsigned short icrc1( unsigned short crc, unsigned char onech )
{
	int i;
	unsigned short ans = (unsigned short)(crc ^ onech << 8);

	for( i = 0;  i < 8;  ++i )
	{
		if( ans & 0x8000 )
		{
			ans = (unsigned short)((ans <<= 1) ^ 4129);
		}
		else
		{
			ans <<= 1;
		}
	}

	return ans;
}

typedef unsigned char uchar;
#define LOBYTE_crc(x) ((uchar)((x) & 0xFF))
#define HIBYTE_crc(x) ((uchar)((x) >> 8))

unsigned short icrc( unsigned short crc, const unsigned char* bufptr, unsigned long len, short jinit, int jrev )
{
	unsigned short icrc1( unsigned short crc, unsigned char onech );
	static unsigned short icrctb[256], init = 0;
	static uchar rchr[256];
	unsigned short cword = crc;
	unsigned long j;
	static uchar it[16] = { 0, 8, 4, 12, 2, 10, 6, 14, 1, 9, 5, 13, 3, 11, 7, 15 };

	if( !init )
	{
		init = 1;
		for( j = 0; j <= 255; ++j )
		{
			icrctb[j] = icrc1( (unsigned short)(j << 8), (uchar)0 );
			rchr[j] = (uchar)(it[j & 0x0F] << 4 | it[j >> 4]);
		}
	}

	if( jinit >= 0 )
	{
		cword = (unsigned short)(((uchar)jinit) | (((uchar)jinit) << 8));
	}
	else if( jrev < 0 )
	{
		cword = (unsigned short)(rchr[HIBYTE_crc(cword)] | rchr[LOBYTE_crc(cword)] << 8);
	}

	for( j = 0; j < len; ++j )
	{
		cword = (unsigned short)(icrctb[(jrev < 0 ? rchr[bufptr[j]] : bufptr[j]) ^ HIBYTE_crc(cword)] ^ LOBYTE_crc(cword) << 8);
	}

	return jrev >= 0 ? cword : (unsigned short)(rchr[HIBYTE_crc(cword)] | rchr[LOBYTE_crc(cword)] << 8);
}


dword getCrc
(
	const hxa7241::primitives::byte* pBytes,
	const dword                      bytesLength
)
{
	dword crc = 0;

	if( bytesLength >= 0 )
	{
		crc = dword( udword( ::icrc( 0, reinterpret_cast<const unsigned char*>(pBytes), bytesLength, 0xFF, 0 ) ) );
	}

	return crc;
}

#endif
