/*--------------------------------------------------------------------

   Perceptuum renderer, version 2.1
   Copyright (c) 2002-2003,  Harrison Ainsworth.

   www.hxa7241.org

--------------------------------------------------------------------*/

/*--------------------------------------------------------------------

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston,
   MA 02111-1307, USA

--------------------------------------------------------------------*/


#include "perceptuum2.rc.h"
#include "Perceptuum2.h"
#include <commctrl.h>
#include <math>

#include "ToneDialog.h"   /// own header is included last


using namespace perceptuum2::platform_ui;




/// statics
const int   ToneDialog::VALUE_DIGITS  = 6;
const float ToneDialog::VALUE_MIN     =  0.0001f;
const float ToneDialog::VALUE_MAX     = 10000.0f;

const word  ToneDialog::SLIDER_MIN    = 0;
const word  ToneDialog::SLIDER_MAX    = 10000;
const word  ToneDialog::SLIDER_PAGE   = 8;
const word  ToneDialog::SLIDER_DECADE = word( (SLIDER_MAX - SLIDER_MIN) / SLIDER_PAGE );




/// standard object services -----------------------------------------------------------------------
ToneDialog::ToneDialog
(
	HINSTANCE    hInstance,
	HWND         hWndParent,
	Perceptuum2* pP2
)
 :	Win32Dialog( hInstance, hWndParent ),
	pP2_m( pP2 ),
	hWndTrackBar_m( NULL ),
	value_m( 0.0f ),
	controlFlag_m( true )
{
}


ToneDialog::~ToneDialog()
{
}




/// commands ---------------------------------------------------------------------------------------
int ToneDialog::run
(
	float& value
)
{
	value_m = value;

	const int dlgReturn = Win32Dialog::run();
	if( dlgReturn != FALSE )
	{
		value = value_m;
	}
	else
	{
		if( pP2_m != 0 )
		{
			pP2_m->setTonemapAdaptationLuminance( value );
		}
	}

	return dlgReturn;
}


int ToneDialog::run()
{
	return Win32Dialog::run();
}




/// overrides --------------------------------------------------------------------------------------
int ToneDialog::getDialogResourceId() const
{
	return IDD_TONEMAP;
}


bool ToneDialog::dialogProc
(
	HWND   hWndDlg,
	UINT   message,
	WPARAM wParam,
	LPARAM //lParam
)
{
	bool didHandleMessage = true;

	switch( message )
	{
	case WM_INITDIALOG :
		{
			/// set max digits for edit
			const int EDIT_BYTE_MAX = VALUE_DIGITS * sizeof(TCHAR);
			HWND hWndEdit = ::GetDlgItem( hWndDlg, IDE_TONE_DIGITS );
			::SendMessage( hWndEdit, EM_SETLIMITTEXT, (WPARAM)EDIT_BYTE_MAX, 0 );

			/// create trackbar
			hWndTrackBar_m = createTrackBar( hWndDlg, hInstance_m, SLIDER_MIN, SLIDER_MAX, SLIDER_PAGE, word((SLIDER_MIN + SLIDER_MAX) / 2) );

			/// clamp value to range
			value_m = clampBetween( value_m, VALUE_MIN, VALUE_MAX );

			/// init edit (and bar pos)
			setEditFloat( hWndDlg, IDE_TONE_DIGITS, value_m, VALUE_DIGITS-1 );
			//setSliderValue( hWndTrackBar_m, value_m );   /// done by the message handlers

			break;
		}
	case WM_VSCROLL :
		{
			if( controlFlag_m )
			{
				controlFlag_m = false;

				float value;
				const bool ok = getSliderValue( hWndTrackBar_m, value );
				if( ok )
				{
					value_m = value;
					setEditFloat( hWndDlg, IDE_TONE_DIGITS, value_m, VALUE_DIGITS-1 );

					if( pP2_m != 0 )
					{
						pP2_m->setTonemapAdaptationLuminance( value_m );
					}
				}

				controlFlag_m = true;
			}

			break;
		}
	case WM_COMMAND :
		{
			if( HIWORD(wParam) == EN_UPDATE )
			{
				if( controlFlag_m )
				{
					controlFlag_m = false;

					float value;
					const bool ok = getEditFloat( hWndDlg, IDE_TONE_DIGITS, value );
					if( ok )
					{
						value_m = clampBetween( value, VALUE_MIN, VALUE_MAX );
						setSliderValue( hWndTrackBar_m, value_m );

						if( pP2_m != 0 )
						{
							pP2_m->setTonemapAdaptationLuminance( value_m );
						}
					}

					controlFlag_m = true;
				}
			}
			else
			{
				switch( wParam )
				{
				case IDB_TONE_AUTO :
					{
						if( pP2_m != 0 )
						{
							const float luminance = pP2_m->getTonemapAveLuminance();

							value_m = clampBetween( luminance, VALUE_MIN, VALUE_MAX );
							setEditFloat( hWndDlg, IDE_TONE_DIGITS, value_m, VALUE_DIGITS-1 );

							pP2_m->setTonemapAdaptationLuminance( value_m );
						}

						break;
					}
				case IDOK :
					{
						float value;
						const bool ok = getEditFloat( hWndDlg, IDE_TONE_DIGITS, value );
						if( ok )
						{
							/// save value, and end ok
							value_m = clampBetween( value, VALUE_MIN, VALUE_MAX );
							::EndDialog( hWndDlg, TRUE );
						}
						else
						{
							/// set focus to edit
							HWND hWndEdit = ::GetDlgItem( hWndDlg, IDE_TONE_DIGITS );
							::SetFocus( hWndEdit );
						}

						break;
					}
				case IDCANCEL :
					{
						::EndDialog( hWndDlg, FALSE );
						break;
					}
				default :
					{
						didHandleMessage = false;
						break;
					}
				}
			}
			break;
		}
	default:
		didHandleMessage = false;
	}

	return didHandleMessage;
}




/// implementation ---------------------------------------------------------------------------------
HWND ToneDialog::createTrackBar
(
	HWND       hWndDlg,
	HINSTANCE  hInstance,
	const word min,
	const word max,
	const word page,
	const word pos
)
{
	/// get position and size
	RECT trackBarRect;
	bool ok = getControlRect( hWndDlg, IDC_TONE_BAR_POS, trackBarRect );

	/// create the window
	HWND hWndTrackBar = NULL;
	if( ok )
	{
		hWndTrackBar = ::CreateWindowEx( 0, TRACKBAR_CLASS, NULL,
		                                 WS_CHILD | WS_VISIBLE | WS_TABSTOP | TBS_VERT | TBS_LEFT | TBS_FIXEDLENGTH | TBS_AUTOTICKS,
		                                 trackBarRect.left, trackBarRect.top,
		                                 trackBarRect.right - trackBarRect.left, trackBarRect.bottom - trackBarRect.top,
		                                 hWndDlg, (HMENU)IDC_TONE_BAR, hInstance, NULL );
		if( hWndTrackBar != NULL )
		{
			::SendMessage( hWndTrackBar, TBM_SETRANGE,    (WPARAM)FALSE,  (LPARAM)((LONG)(((WORD)(min)) | (((DWORD)((WORD)(max))) << 16))) );
			::SendMessage( hWndTrackBar, TBM_SETPAGESIZE, 0,              (LPARAM)((max-min)/page) );
			::SendMessage( hWndTrackBar, TBM_SETTICFREQ,  (max-min)/page, (LPARAM)0 );
			::SendMessage( hWndTrackBar, TBM_SETSEL,      (WPARAM)FALSE,  (LPARAM)((LONG)(((WORD)(pos)) | (((DWORD)((WORD)(pos))) << 16))) );
			::SendMessage( hWndTrackBar, TBM_SETPOS,      (WPARAM)TRUE,   (LPARAM)pos );
		}
	}

	return hWndTrackBar;
}




bool ToneDialog::getSliderValue
(
	HWND   hWndTrackBar,
	float& value
)
{
	if( hWndTrackBar != 0 )
	{
		/// get position
		const word pos = word( ::SendMessage( hWndTrackBar, TBM_GETPOS, 0, 0 ) );

		/// calc value (logarithmic scale)
		value = float( std::pow( 10.0, (float(SLIDER_MAX - pos) / float(SLIDER_DECADE)) - float(SLIDER_PAGE / 2) ) );
	}

	return hWndTrackBar != 0;
}


void ToneDialog::setSliderValue
(
	HWND   hWndTrackBar,
	float  value
)
{
	/// calc the position (logarithmic scale)
	const dword pos = dword( ( -float(std::log10( value )) + float(SLIDER_PAGE / 2) ) * float(SLIDER_DECADE) );

	/// set the slider
	if( hWndTrackBar != 0 )
	{
		::SendMessage( hWndTrackBar, TBM_SETPOS, (WPARAM)TRUE, (LPARAM)pos );
	}
}


float ToneDialog::clampBetween
(
	float       val,
	const float min,
	const float max
)
{
	if( val < min )
	{
		val = min;
	}
	else
	if( val > max )
	{
		val = max;
	}

	return val;
}
