/*--------------------------------------------------------------------

   Perceptuum renderer, version 2.1
   Copyright (c) 2002-2003,  Harrison Ainsworth.

   www.hxa7241.org

--------------------------------------------------------------------*/

/*--------------------------------------------------------------------

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston,
   MA 02111-1307, USA

--------------------------------------------------------------------*/


#ifndef FpToInt_h
#define FpToInt_h


#include "Primitives.h"




#include "hxa7241_utility.h"
namespace hxa7241 { namespace utility {
	using namespace hxa7241::primitives;


/**
 * faster conversion from float to 32bit int, truncating fractional part.
 */
dword fpToIntTrunc( const float f );


/**
 * faster conversion from float to 32bit int, rounding fractional part.
 */
dword fpToIntRound( const float f );




/// INLINES ///


/**
 * faster conversion from float to 32bit int, rounding according to fp mode.
 * <br/><br/>
 *
 * doesnt clamp to dword range.
 *
 * @implementation
 * uses the alignment matching behaviour of an fp add to reposition the floating
 * point right of bit 0.
 */
inline
dword fpToInt( const float f )
{
	/// add 2^52 * 1.5
	const double d = double(f) + 6755399441055744.0;
	return *(reinterpret_cast<const dword*>( &d ));

//	double dtemp = ((((65536.0 * 65536.0 * 16) + (65536.0 * 0.5)) * 65536.0)) + double(f);
//	return *(reinterpret_cast<dword*>( &dtemp )) - dword(0x80000000);
}



/**
 * faster conversion from float to 16.16bit fixed point int, rounding according to
 * fp mode.<br/><br/>
 *
 * doesnt clamp to dword range.
 *
 * @implementation
 * uses the alignment matching behaviour of an fp add to reposition the floating
 * point right of bit 16.
 */
inline
dword fpToInt1616( const float f )
{
	/// add (2^52)/65536 * 1.5
	const double d = double(f) + 103079215104.0;
	return *(reinterpret_cast<const dword*>( &d ));

//	double dtemp = ((((65536.0 * 16) + (0.5)) * 65536.0)) + double(f);
//	return *(reinterpret_cast<dword*>( &dtemp )) - dword(0x80000000);
}



/**
 * faster conversion from 0-1 float to 32bit int, rounding according to fp mode.
 * <br/><br/>
 *
 * doesnt clamp to dword range.
 *
 * @implementation
 * uses the alignment matching behaviour of an fp add to reposition the floating
 * point right of bit 32.
 */
inline
dword fp01ToDword( const float f )
{
	/// add (2^52)/(2^32) * 1.5
	const double d = double(f) + 1572864.0;
	return *(reinterpret_cast<const dword*>( &d ));

//	double dtemp = ((((16) + (0.5 / 65536.0)) * 65536.0)) + double(f);
//	return *(reinterpret_cast<dword*>( &dtemp )) - dword(0x80000000);
}




struct FRACTIONALIZER
{
	static const float  _4BIT;
	static const float  _8BIT;
	static const float _16BIT;
};

/**
 * converts from 4bit int to float fraction.<br/><br/>
 *
 * repeats int bit pattern along fraction bits.
 */
inline
float unibbleToFp( const dword n )
{
	return float(n) * FRACTIONALIZER::_4BIT;
}

/**
 * converts from 8bit int to float fraction.<br/><br/>
 *
 * repeats int bit pattern along fraction bits.
 */
inline
float ubyteToFp( const dword b )
{
	return float(b) * FRACTIONALIZER::_8BIT;
}

/**
 * converts from 16bit int to float fraction.<br/><br/>
 *
 * repeats int bit pattern along fraction bits.
 */
inline
float uwordToFp( const dword w )
{
	return float((w << 8) | (w >> 8)) * FRACTIONALIZER::_16BIT;
}


}}




#endif//FpToInt_h
