/*--------------------------------------------------------------------

   Perceptuum renderer, version 2.1
   Copyright (c) 2002-2003,  Harrison Ainsworth.

   www.hxa7241.org

--------------------------------------------------------------------*/

/*--------------------------------------------------------------------

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston,
   MA 02111-1307, USA

--------------------------------------------------------------------*/


#include "HaltonSequence.h"   /// own header is included last


using namespace hxa7241::utility;




/// statics
const dword HaltonSequence::PRIMES[] = { 2, 3, 5, 7, 11, 13, 17, 19, 23, 29, 31, 37, 41, 43, 47 };




/// standard object services -----------------------------------------------------------------------
HaltonSequence::HaltonSequence()
{
	HaltonSequence::setBase( 0 );
}


HaltonSequence::HaltonSequence
(
	const dword base
)
{
	HaltonSequence::setBase( base );
}


HaltonSequence::~HaltonSequence()
{
}


HaltonSequence::HaltonSequence( const HaltonSequence& other )
{
	HaltonSequence::operator=( other );
}


HaltonSequence& HaltonSequence::operator=( const HaltonSequence& other )
{
	if( &other != this )
	{
		Sequence::operator=( other );

		prime_m = other.prime_m;
		index_m = other.index_m;
		point_m = other.point_m;
	}

	return *this;
}




/// commands ---------------------------------------------------------------------------------------
void HaltonSequence::setBase
(
	const dword base
)
{
	prime_m = PRIMES[ (base >= 0 ? base : -base) % (sizeof(PRIMES)/sizeof(*PRIMES)) ];

	setIndex( 0 );
}


void HaltonSequence::setIndex
(
	const dword index
)
{
	index_m = index >= 0 ? index : 0;

	point_m = getItem( prime_m, index_m );
}


void HaltonSequence::next()
{
	setIndex( index_m + 1 );
}




/// queries ----------------------------------------------------------------------------------------
///// static
//float HaltonSequence::getItem2
//(
//	const dword prime,
//	float       item
//)
//{
//	/// mostly works, but goes badly awry in patches
//
//	const float r = 1.0 - item - 0.000001f;
//	const float oneOverPrime = 1.0f / float(prime);
//
//	if( oneOverPrime < r )
//	{
//		item += oneOverPrime;
//	}
//	else
//	{
//		float h = oneOverPrime;
//		float hh;
//		do
//		{
//			hh = h;
//			h *= oneOverPrime;
//		} while( h >= r );
//
//		item += (hh + h - 1.0f);
//	}
//
//	return item;
//}


/// static
float HaltonSequence::getItem
(
	const dword prime,
	dword       index
)
{
	if( index < 0 )
	{
		index = 0;
	}

	const float oneOverPrime = 1.0f / float(prime);
	float p      = oneOverPrime;
	float result = 0.0f;

	for( dword k = index + 1;  k > 0;  p *= oneOverPrime )
	{
		/// assuming * and - is faster than %
		const dword kDiv = k / prime;
		const dword kMod = k - (kDiv * prime);

		result += float(kMod) * p;

		k = kDiv;
	}
//	for( dword k = index + 1;  k > 0;  p *= oneOverPrime, k /= prime )
//	{
//		/// neater, but does two divides
//		result += float(k % prime) * p;
//	}

	return result;
}


/// static
float HaltonSequence::getItemJittered
(
	const dword prime,
	const dword index,
	const float rand01
)
{
	return jitterPoint( getItem( prime, index ), index, rand01 );
}


/// static
float HaltonSequence::jitterPoint
(
	const float point,
	dword       index,
	const float rand01
)
{
	if( index < 0 )
	{
		index = 0;
	}

	/// jitter
	const float jitter = (rand01 - 0.5f) / float(index + 1);
	float jittered = point + jitter;

	/// wrap around
	if( jittered < 0.0f )
	{
		jittered += 1.0f;
	}
	else if( jittered >= 1.0f )
	{
		jittered -= 1.0f;
	}

	return jittered;
}










/// 'Sampling with Hammersley and Halton Points'
/// - Wong, Luk, Heng
/// Journal of Graphics Tools, vol2, no2, 1997

///*
// * Halton point set generation
// * two p-adic Van der Corport sequences
// * Useful for incremental approach.
// * p1 = 2, p2 = 3(default)
// */
//void PlaneHalton(float *result, int n, int p2)
//{
//  float p, u, v, ip;
//  int k, kk, pos, a;
//
//  for (k=0, pos=0 ; k<n ; k++)
//  {
//    u = 0;
//    for (p=0.5, kk=k ; kk ; p*=0.5, kk>>=1)
//      if (kk & 1)                          // kk mod 2 == 1
//	u += p;
//
//    v = 0;
//    ip = 1.0/p2;                           // recipical of p2
//    for (p=ip, kk=k ; kk ; p*=ip, kk/=p2)  // kk = (int)(kk/p2)
//      if ((a = kk % p2))
//	v += a * p;
//
//    result[pos++] = u;
//    result[pos++] = v;
//  }
//}
//
//
///*
// * Halton point set generation
// * two p-adic Van der Corport sequences
// * Useful for incremental approach.
// * p1 = 2(default), p2 = 3(default)
// */
/*void HaltonSequence::PlaneHalton2( float* result, int n, int p1, int p2 )
{
	float p, u, v, ip;
	int k, kk, pos, a;

	for( k = 0, pos = 0;  k < n;  k++ )
	{
		u = 0;
		ip = 1.0f / p1;
		for( p = ip, kk = k;  kk;  p *= ip, kk /= p1 )
		{
			if( (a = kk % p1) != 0 )
			{
				u += a * p;
			}
		}

		v = 0;
		ip = 1.0f / p2;
		for( p = ip, kk = k;  kk;  p *= ip, kk /= p2 )
		{
			if( (a = kk % p2) != 0 )
			{
				v += a * p;
			}
		}

		result[ pos++ ] = u;
		result[ pos++ ] = v;
	}
}*/








/*class Halton
{
/// standard object services -----------------------------------------------------------------------
public:
	         Halton( dword base,
	                 dword index );
	        ~Halton();
	         Halton( const Halton& );
	Halton& operator=( const Halton& );


/// commands ---------------------------------------------------------------------------------------
	        void  next();


/// queries ----------------------------------------------------------------------------------------
	        float getItem() const;


/// fields -----------------------------------------------------------------------------------------
private:
	float invBase_m;
	float value_m;
};




Halton::Halton
(
	const dword base,
	const dword index
)
 :	invBase_m( 1.0f / float(base) ),
	value_m  ( 0.0f )
{
	dword i = index;
	float f = invBase_m;
	while( i > 0 )
	{
		value_m += f * float(i % base);
		i /= base;
		f *= invBase_m;
	}
}


inline
Halton::~Halton()
{
}


inline
Halton::Halton
(
	const Halton& other
)
{
	Halton::operator=( other );
}


Halton& Halton::operator=( const Halton& other )
{
	invBase_m = other.invBase_m;
	value_m   = other.value_m;

	return *this;
}


void Halton::next()
{
	float r = 1.0f - value_m - FLOAT_EPSILON;
	if( invBase_m < r )
	{
		value_m += invBase_m;
	}
	else
	{
		float h = invBase_m;
		float hh;
		do
		{
			hh = h;
			h *= invBase_m;
		} while( h >= r );

		value_m += hh + h - 1.0f;
	}
}


inline
float Halton::getItem() const
{
	return value_m;
}*/
