/*--------------------------------------------------------------------

   Perceptuum renderer, version 2.1
   Copyright (c) 2002-2003,  Harrison Ainsworth.

   www.hxa7241.org

--------------------------------------------------------------------*/

/*--------------------------------------------------------------------

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston,
   MA 02111-1307, USA

--------------------------------------------------------------------*/


#ifndef SobolSequence_h
#define SobolSequence_h


#include "Primitives.h"
#include "Sequence.h"
#include "Sheet.h"




#include "hxa7241_utility.h"
namespace hxa7241 { namespace utility {
	using namespace hxa7241::primitives;
	using hxa7241::utility::SerializerOut;
	using hxa7241::utility::SerializerIn;
	using hxa7241::utility::Sheet;


#define SOBOL_MAX_DIMENSION 40
#define SOBOL_BIT_COUNT 30




/**
 * one dimension of a sobol sequence.<br/><br/>
 *
 * only 40 dimensions implemented, sequence length ~ 1e+9.<br/><br/>
 *
 * @implementation
 * 2003-04-18<br/>
 * this is essentially a wrapper for an implementation by Gerard Jungman,
 * from 'GSL - GNU Scientific Library' v1.3, www.gnu.org/software/gsl/<br/>
 * modified original code at bottom of .c file.<br/>
 * - although, in the end, i used the gsl code to generate a lookup table,
 * and the code that executes at runtime is a rewrite following a small part
 * of the gsl code.<br/>
 * the original gsl code is retained for testing and reference.
 */
class SobolSequence
	: public Sequence
{
/// standard object services -----------------------------------------------------------------------
public:
	         SobolSequence();
	         /** @param dimension [0, SOBOL_MAX_DIMENSION-1] */
	explicit SobolSequence( dword dimension );

	        ~SobolSequence();
	         SobolSequence( const SobolSequence& );
	SobolSequence& operator=( const SobolSequence& );


/// commands ---------------------------------------------------------------------------------------
	        /** @param dimension [0, SOBOL_MAX_DIMENSION-1] */
	        void  setDimension( dword dimension );
	        /** @param index value >= 0 */
	        void  setIndex( dword index );

	        void  next();


/// queries ----------------------------------------------------------------------------------------
	        dword getDimension()                                                              const;
	        dword getIndex()                                                                  const;

	        float getItem()                                                                   const;

	/// testing
	static  bool test( bool isLogging );
	static  void testView( dword dim0,
	                       dword dim1,
	                       dword iterations,
	                       const Sheet<byte>& image );


/// serialization ----------------------------------------------------------------------------------
	friend  SerializerOut& operator<<( SerializerOut&, const SobolSequence& );
	friend  SerializerIn&  operator>>( SerializerIn&,        SobolSequence& );
	static const char* CLASS_NAME();
	static const char* CLASS_DESCRIPTION();


/// implementation ---------------------------------------------------------------------------------
protected:
	        /** @return new point */
	        float nextImplementation();


/// fields -----------------------------------------------------------------------------------------
private:
	dword  dimension_m;

	dword  index_m;
	float  lastDenominatorInv_m;
	dword  lastNumerator_m;
	dword  vDirection_m[SOBOL_BIT_COUNT];

	float  point_m;

	/// constant precalculations
	static const dword V_DIRECTION[SOBOL_BIT_COUNT][SOBOL_MAX_DIMENSION];
};




/// serialization ----------------------------------------------------------------------------------
SerializerOut& operator<<( SerializerOut&, const SobolSequence& );
SerializerIn&  operator>>( SerializerIn&,        SobolSequence& );








/// inlines ///

/// queries ----------------------------------------------------------------------------------------
inline
dword SobolSequence::getDimension() const
{
	return dimension_m;
}


inline
dword SobolSequence::getIndex() const
{
	return index_m;
}


inline
float SobolSequence::getItem() const
{
	return point_m;
}


}}




#endif//SobolSequence_h
