/*--------------------------------------------------------------------

   Perceptuum renderer, version 2.1
   Copyright (c) 2002-2003,  Harrison Ainsworth.

   www.hxa7241.org

--------------------------------------------------------------------*/

/*--------------------------------------------------------------------

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston,
   MA 02111-1307, USA

--------------------------------------------------------------------*/


#include "Logger.h"
#include "Exception.h"
#include <string>
#include <stdio>

#include "Win32Dialog.h"   /// own header is included last


using namespace hxa7241::win32;




/// statics
const char Win32Dialog::CLASS_NAME[] = "hxa7241::win32::Win32Dialog";




/// standard object services -----------------------------------------------------------------------
Win32Dialog::Win32Dialog
(
	HINSTANCE hInstance,
	HWND      hWndParent
)
 :	hInstance_m( hInstance ),
	hWndParent_m( hWndParent )
{
}


Win32Dialog::~Win32Dialog()
{
}




/// commands ---------------------------------------------------------------------------------------
BOOL CALLBACK Win32DialogProc( HWND hwndDlg, UINT message, WPARAM wParam, LPARAM lParam );


int Win32Dialog::run()
{
	return ::DialogBoxParam( hInstance_m, MAKEINTRESOURCE( getDialogResourceId() ), hWndParent_m,
	                         (DLGPROC)Win32DialogProc, reinterpret_cast<LPARAM>(this) );
}




/// implementation ---------------------------------------------------------------------------------
bool Win32Dialog::setPositionCentered
(
	HWND hWndDlg
)
{
	/// get centered position
	POINT centered;
	bool  ok = false;
	{
		/// get desktop size
		HWND dw = ::GetDesktopWindow();
		if( dw != NULL )
		{
			RECT screen;
			const BOOL oks = ::GetClientRect( dw, &screen );
			if( oks )
			{
				/// get window size
				RECT window;
				const BOOL okw = ::GetWindowRect( hWndDlg, &window );
				if( okw )
				{
					centered.x = (screen.right  - (window.right  - window.left)) / 2;
					centered.y = (screen.bottom - (window.bottom - window.top))  / 2;
					ok = true;
				}
			}
		}
	}

	/// set position
	if( ok )
	{
		::SetWindowPos( hWndDlg, NULL, centered.x, centered.y, 0, 0, SWP_NOOWNERZORDER | SWP_NOZORDER | SWP_NOSIZE );
	}

	return ok;
}


void Win32Dialog::onEnUpdateInt
(
	HWND        hWndDlg,
	const int   editId,
	const dword min,
	const dword max
)
{
	static bool antiRecurse = false;

	/// stop the modification from generating another modification
	if( !antiRecurse )
	{
		/// get edit text length, to check if its empty
		HWND hWndEdit = ::GetDlgItem( hWndDlg, editId );
		const int length = int( ::SendMessage( hWndEdit, EM_LINELENGTH, 0, 0 ) );

		/// leave empty edit alone, so its not clamped into range
		if( length > 0 )
		{
			const bool isSigned = min < 0;

			/// get edit value
			BOOL ok;
			UINT value = ::GetDlgItemInt( hWndDlg, editId, &ok, isSigned );

			/// clamp to range
			if( isSigned )
			{
				value = (dword(value) >= min) ? value : UINT(min);
				value = (dword(value) <= max) ? value : UINT(max);
			}
			else
			{
				value = (value >= UINT(min)) ? value : UINT(min);
				value = (value <= UINT(max)) ? value : UINT(max);
			}

			/// note selection point
			dword posStart;
			dword posEnd;
			HWND hWndEdit = ::GetDlgItem( hWndDlg, editId );
			::SendMessage( hWndEdit, EM_GETSEL, (WPARAM)(&posStart), (WPARAM)(&posEnd) );

			/// set edit with checked value
			antiRecurse = true;
			::SetDlgItemInt( hWndDlg, editId, value, isSigned );
			antiRecurse = false;

			/// set selection point to previous position
			::SendMessage( hWndEdit, EM_SETSEL, (WPARAM)posStart, (WPARAM)posEnd );
		}
	}
}


void Win32Dialog::onEnUpdateFloat
(
	HWND        hWndDlg,
	const int   editId,
	const float min,
	const float max
)
{
	static bool antiRecurse = false;

	/// stop the modification from generating another modification
	if( !antiRecurse )
	{
		/// get edit text length, to check if its empty
		HWND hWndEdit = ::GetDlgItem( hWndDlg, editId );
		const int length = int( ::SendMessage( hWndEdit, EM_LINELENGTH, 0, 0 ) );

		/// leave empty edit alone, so its not clamped into range
		if( length > 0 )
		{
			/// get edit value
			const int BUF_LEN = 8;
			TCHAR buf[BUF_LEN];
			::GetDlgItemText( hWndDlg, editId, buf, BUF_LEN );
			float value;
			std::sscanf( buf, "%g", &value ); //int sscanf(char *buffer, const char *format,...);  %g

			/// clamp to range
			value = (value >= min) ? value : min;
			value = (value <= max) ? value : max;

			/// note selection point
			dword posStart;
			dword posEnd;
			HWND hWndEdit = ::GetDlgItem( hWndDlg, editId );
			::SendMessage( hWndEdit, EM_GETSEL, (WPARAM)(&posStart), (WPARAM)(&posEnd) );

			/// set edit with checked value
			std::_snprintf( buf, BUF_LEN, "% .4g", double(value) );  //int _snprintf(char *buffer, size_t count, const char *format,...);
			antiRecurse = true;
			::SetDlgItemText( hWndDlg, editId, buf );
			antiRecurse = false;

			/// set selection point to previous position
			::SendMessage( hWndEdit, EM_SETSEL, (WPARAM)posStart, (WPARAM)posEnd );
		}
	}
}


bool Win32Dialog::getEditFloat
(
	HWND      hWndDlg,
	const int editId,
	float&    value
)
{
	const int BUF_LEN = 14;
	TCHAR buf[BUF_LEN];

	bool ok = false;
	/// get text
	const UINT len = ::GetDlgItemText( hWndDlg, editId, buf, BUF_LEN );
	if( len > 0 )
	{
		/// translate to value
		float f = 0.0f;
		const bool fail = !bool( std::sscanf( buf, "%g", &f ) ); //int sscanf( char* buffer, const char* format,... );  %g == float*
		if( !fail )
		{
			value = f;
			ok = true;
		}
	}

	return ok;
}


bool Win32Dialog::setEditFloat
(
	HWND         hWndDlg,
	const int    editId,
	const float  value,
	const udword digits
)
{
	const int BUF_LEN = 14;
	TCHAR buf[BUF_LEN];

	std::_snprintf( buf, BUF_LEN, "%.*g", int(digits), double(value) );  //int _snprintf( char* buffer, size_t count, const char* format,... );
	const bool ok = bool( ::SetDlgItemText( hWndDlg, editId, buf ) );

	return ok;
}


bool Win32Dialog::getControlRect
(
	HWND      hWndDlg,
	const int id,
	RECT&     rect
)
{
	bool ok = false;

	/// get position and size from dialog resource
	HWND hWndPos = ::GetDlgItem( hWndDlg, id );
	if( hWndPos != NULL )
	{
		RECT r;
		const BOOL okw = ::GetWindowRect( hWndPos, &r );
		if( okw )
		{
			BOOL okc = TRUE;
			POINT p1 = { r.left, r.top };
			okc &= ::ScreenToClient( hWndDlg, &p1 );
			POINT p2 = { r.right, r.bottom };
			okc &= ::ScreenToClient( hWndDlg, &p2 );

			if( okc )
			{
				rect.left   = p1.x;
				rect.top    = p1.y;
				rect.right  = p2.x;
				rect.bottom = p2.y;
				ok = true;
			}
		}
	}

	return ok;
}




/// win32 ------------------------------------------------------------------------------------------
BOOL CALLBACK Win32DialogProc
(
	HWND   hWndDlg,
	UINT   message,
	WPARAM wParam,
	LPARAM lParam
)
{
	/// set up association between hWnd and object
	if( message == WM_INITDIALOG )
	{
		::SetLastError( 0 );
		/// init hWnd with object ptr
		const LONG ok = ::SetWindowLong( hWndDlg, DWL_USER, static_cast<LONG>(lParam) );

		/// quit if fail
		if( (ok == 0) && (::GetLastError() != 0) )
		{
			::EndDialog( hWndDlg, FALSE );
		}
	}


	/// dispatch the message
	BOOL didHandleMessage = FALSE;
	{
		/// get the object ptr for this hWnd
		const LONG objectPtrAsLong = ::GetWindowLong( hWndDlg, DWL_USER );
		if( objectPtrAsLong != 0 )
		{
			Win32Dialog* pObject = reinterpret_cast<Win32Dialog*>( objectPtrAsLong );

			bool ok = false;
			try
			{
				/// call (virtually of course) the objects proc
				didHandleMessage = pObject->dialogProc( hWndDlg, message, wParam, lParam );
				ok = true;
			}
			catch( const std::exception& e )
			{
				using hxa7241::utility::Logger;
				Logger() << Logger::EXC2 << e.what();
			}
			catch( ... )
			{
				hxa7241::utility::Exception e;
				using hxa7241::utility::Logger;
				Logger() << Logger::EXC2 << e.what();
			}

			if( !ok )
			{
				::EndDialog( hWndDlg, FALSE );
			}
		}
		else
		{
			didHandleMessage = FALSE;
		}
	}

	return didHandleMessage;
}


void Win32Dialog::convertText
(
	const bool toWin32,
	string&    text
)
{
	for( int i = text.length();  i-- > 0; )
	{
		if( toWin32 && (text[i] == '\n') )
		{
			text.insert( i, 1, '\r' );
		}
		if( !toWin32 && (text[i] == '\r') )
		{
			text.erase( i );
		}
	}
}
